package Asterisk::API::Structures::Admins;

use strict;
use warnings;

use Asterisk::API::Mapper::Generic;
use Asterisk::API::Documentation::Generic;

our $VERSION = "1";

our %out_hash_map = (
  "id" => {
    "description" => "Admin ID",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 1,
    "transform" => "id",
    "type" => "integer"
  },
  "user" => {
    "description" => "User",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "login",
    "type" => "string"
  },
  "wpass" => {
    "description" => "Wpass",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "password",
    "type" => "string"
  }

);

my %in_hash_map = (
  "id" => {
    "description" => "Admin ID",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 1,
    "transform" => "id",
    "type" => "integer"
  },
  "login" => {
    "description" => "Login",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "user",
    "type" => "string"
  },
  "password" => {
    "description" => "Password",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "wpass",
    "type" => "string"
  }

);

our %requests = (
	types => [
	    {
		"type" => "admin",
		"structure" => \%out_hash_map
	    }
	],
	operations => [
	    {
		"path" => "/api/admin/{tenantName}/",
		"method" => "POST",
		"summary" => "Add admin",
		"notes" => "Add admin record",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "admin",
			"parameter" => "request",
			"sample" => q({ "login" : "admin-tl", "password" : "4a55w0rd" })
		    },
		    {
			"name" => "tenantName",
			"description" => "Tenant Name or Tenant ID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "thirdlane"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "admin"
			}
		]
	    },
	    {
		"path" => "/api/admin/{tenantName}/{adminId}",
		"method" => "DELETE",
		"summary" => "Delete admin",
		"notes" => "Delete admin record",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "tenantName",
			"description" => "Tenant Name or Tenant ID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "thirdlane"
		    },
		    {
			"name" => "adminId",
			"description" => "Admin ID",
			"required" => 1,
			"type" => "integer",
			"parameter" => "path"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "admin"
			}
		]
	    }

	]
);

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{version} = $VERSION;

    $self->{mapper} = new Asterisk::API::Mapper::Generic;

    my $out_mapping = $self->{mapper}->prepare( \%out_hash_map );
    $self->{defaults}->{out} = $out_mapping->{defaults};
    $self->{map}->{out} = $out_mapping->{map};
    $self->{type}->{out} = $out_mapping->{type};
    $self->{reverse}->{out} = $out_mapping->{reverse};

    my $in_mapping = $self->{mapper}->prepare( \%in_hash_map );
    $self->{defaults}->{in} = $in_mapping->{defaults};
    $self->{map}->{in} = $in_mapping->{map};
    $self->{type}->{in} = $in_mapping->{type};
    $self->{reverse}->{in} = $in_mapping->{reverse};

    return $self;
}


sub info {
    my ( $self ) = @_;

    my $doc = new Asterisk::API::Documentation::Generic;

    return $doc->prepare_types( \%requests );
}

sub map_out {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{out} } ) { # fill default values
	    $out_hash->{$self->{map}->{out}->{$key}} = $self->{defaults}->{out}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{out}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{out}->{$key} && exists $hash->{$key};
    }

    $out_hash = $self->{mapper}->fix_type($out_hash, $self->{type}->{out}, $self->{reverse}->{out});

    return $out_hash;
}


sub map_in {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{in} } ) { # fill default values
	    $out_hash->{$self->{map}->{in}->{$key}} = $self->{defaults}->{in}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{in}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{in}->{$key} && exists $hash->{$key};
    }

    return $out_hash;
}

1;
