package Asterisk::API::Structures::CallHistory;

use strict;
use warnings;

use Asterisk::API::Mapper::Generic;
use Asterisk::API::Documentation::Generic;

our $VERSION = "1";

use DateTime;
use DateTime::Format::Strptime;

our %out_hash_map = (
  "billsec" => {
    "description" => "Billsec",
    "format" => "int64",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "duration",
    "type" => "integer"
  },
  "direction" => {
    "description" => "Call Direction",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "incoming",
    "transform" => "direction",
    "type" => "string"
  },
  "disposition" => {
    "description" => "Disposition",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "ANSWERED",
    "transform" => "status",
    "type" => "string"
  },
  "dst" => {
    "description" => "Call Destination",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "901",
    "transform" => "destination",
    "type" => "string"
  },
  "lastapp" => {
    "description" => "Lastapp",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "last_app",
    "type" => "string"
  },
  "recording" => {
    "description" => "Recording",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "recording",
    "type" => "string"
  },
  "uniqueid" => {
    "description" => "Uniqueid",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "unique_id",
    "type" => "string"
  },
  "date" => {
    "description" => "Call Date",
    "format" => "timestamp",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Tue Nov 24 11:05:16 PM EET 2015",
    "transform" => "date",
    "type" => "string"
  },
  "date_epoch" => {
    "description" => "Call Date in UTC epoch value",
    "format" => "unixtime",
    "hidden" => 0,
    "required" => 0,
    "sample" => "1448399116",
    "transform" => "date_epoch",
    "type" => "integer"
  },
  "caller_id" => {
    "description" => "Caller Information",
    "format" => "caller_id",
    "hidden" => 0,
    "required" => 0,
    "sample" => qq({ "number" : "901", "name" : "John Snow" }),
    "transform" => "caller_id",
    "type" => "caller_id"
  },
  "user_src" => {
    "description" => "Source User",
    "format" => "user",
    "hidden" => 0,
    "required" => 0,
    "sample" => "thirdlane-101",
    "transform" => "user_src",
    "type" => "string"
  },
  "user_dst" => {
    "description" => "Destination User",
    "format" => "user",
    "hidden" => 0,
    "required" => 0,
    "sample" => "thirdlane-103",
    "transform" => "user_dst",
    "type" => "string"
  }
);

our %out_hash_map_caller_id = (
  "name" => {
    "description" => "Caller Name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Jhon Snow",
    "transform" => "name",
    "type" => "string"
  },
  "number" => {
    "description" => "Caller Number",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "901",
    "transform" => "number",
    "type" => "string"
  }
);

my %in_hash_map = (
);

our %requests = (
	types => [
	    {
		"type" => "call_history",
		"structure" => \%out_hash_map
	    },
	    {
		"type" => "caller_id",
		"structure" => \%out_hash_map_caller_id
	    }
	],
	operations => [
	    {
		"path" => "/api/user/call_history",
		"method" => "GET",
		"summary" => "Get User's Call History",
		"notes" => "Get User's Call History",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "call_history"
			}
		]
	    }
	]
);

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{version} = $VERSION;

    $self->{mapper} = new Asterisk::API::Mapper::Generic;

    my $out_mapping = $self->{mapper}->prepare( \%out_hash_map );
    $self->{defaults}->{out} = $out_mapping->{defaults};
    $self->{map}->{out} = $out_mapping->{map};
    $self->{type}->{out} = $out_mapping->{type};
    $self->{reverse}->{out} = $out_mapping->{reverse};

    my $in_mapping = $self->{mapper}->prepare( \%in_hash_map );
    $self->{defaults}->{in} = $in_mapping->{defaults};
    $self->{map}->{in} = $in_mapping->{map};
    $self->{type}->{in} = $in_mapping->{type};
    $self->{reverse}->{in} = $in_mapping->{reverse};

    return $self;
}

sub info {
    my ( $self ) = @_;

    my $doc = new Asterisk::API::Documentation::Generic;

    return $doc->prepare_types( \%requests );
}

sub map_out {
    my ($self, $hash) = @_;

    my $out_hash = {};

    my @directions_str = ("outgoing", "incoming");

    my $hash_map = \%out_hash_map;

    my $in_date_format = DateTime::Format::Strptime->new(
            time_zone => 'local',
            pattern   => '%Y-%m-%d %H:%M:%S',
    );

    my $out_date_format = DateTime::Format::Strptime->new(
            time_zone => $hash->{tz},
            pattern   => $hash->{date_time_format} || '%a %b %d %r %Z %Y', # ctime(3) format?
    );

    delete $hash->{date_time_format} if exists $hash->{date_time_format};

    eval {

    my $out_date = $in_date_format->parse_datetime($hash->{calldate});
    $out_date->set_time_zone($hash->{tz});
    $out_date->set_formatter( $out_date_format );

    $out_hash->{date} = sprintf "%s", $out_date;
    $out_hash->{date_epoch} = $out_date->epoch();

	if (exists $hash->{current_ext}) {
	    my $direction = $hash->{current_ext} eq $hash->{src} ? 0 : 1;
	    $out_hash->{direction} = $directions_str[$direction];
	}

	if (exists $hash->{current_user}) {
	    my $direction = $hash->{current_user} eq $hash->{user_src} ? 0 : 1;
	    $out_hash->{direction} = $directions_str[$direction];
	}

    if ( my ( $name, $number ) = $hash->{clid} =~ m/^"(.*?)"\s+<(.*?)>$/ ) {
	$out_hash->{caller_id} = {
	    name   => $name,
	    number => $number
	};
    } else {
	$out_hash->{caller_id} = { number => $hash->{clid} };
    }

    foreach my $key (keys %{$hash}) {
	$out_hash->{$self->{'map'}->{'out'}->{$key}} = $hash->{$key} if exists $self->{'map'}->{'out'}->{$key};
    }

#    foreach my $key (keys %{$hash}) {
#	my $outkey = $hash_map->{$key};
#	$out_hash->{$outkey} = $hash->{$key} if exists $hash_map->{$key};
#    }

    };

    $out_hash = $self->{mapper}->fix_type($out_hash, $self->{type}->{out}, $self->{reverse}->{out});

    return $out_hash;
}

sub map_out_natural {
    my ($self, $hash) = @_;
return $hash;
    my $out_hash = {};

    foreach my $key (keys %{$hash}) {
	$out_hash->{$self->{'map'}->{'out'}->{$key}} = $hash->{$key} if exists $self->{'map'}->{'out'}->{$key};
    }

    return $out_hash;
}

sub map_in {
    my ($self, $hash) = @_;
return $hash;
    my $out_hash = {};

    foreach my $key (keys %{$hash}) {
	$out_hash->{$self->{'map'}->{'in'}->{$key}} = $hash->{$key} if exists $self->{'map'}->{'in'}->{$key};
    }

    return $out_hash;
}

1;
