package Asterisk::API::Structures::ContactSource;

use strict;
use warnings;

use Asterisk::API::Mapper::Generic;
use Asterisk::API::Documentation::Generic;

our $VERSION = "1";

our %out_hash_map = (
  "id" => {
    "description" => "Source ID",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 1,
    "transform" => "id",
    "type" => "integer"
  },
  "type" => {
    "description" => "Source name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "contact_source_type",
    "type" => "string"
  },
  "description" => {
    "description" => "Source description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "description",
    "type" => "string"
  },
  "status" => {
    "description" => "Source status",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "status",
    "type" => "integer"
  },
  "data" => {
    "description" => "Source parameters",
    "format" => "array",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "data",
    "type" => "source_parameter"
  }
);

my %in_hash_map = (
  "id" => {
    "description" => "Source ID",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 1,
    "transform" => "id",
    "type" => "integer"
  },
  "contact_source_type" => {
    "description" => "Source name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "type",
    "type" => "string"
  },
  "description" => {
    "description" => "Source description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "description",
    "type" => "string"
  },
  "status" => {
    "description" => "Source status",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "status",
    "type" => "integer"
  },
  "data" => {
    "description" => "Source parameters",
    "format" => "array",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "data",
    "type" => "source_parameter"
  }
);

our %contact_source_parameter_type = (
  "server_url" => {
    "description" => "Source URL",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "server_url",
    "type" => "string"
  },
  "api_token" => {
    "description" => "Source token",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "api_token",
    "type" => "string"
  },
  "username" => {
    "description" => "Source username",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "username",
    "type" => "string"
  },
  "password" => {
    "description" => "Source password",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "password",
    "type" => "string"
  }

);

our %requests = (
	types => [
	    {
		"type" => "source_parameter",
		"structure" => \%contact_source_parameter_type
	    },
	    {
		"type" => "contact_source",
		"structure" => \%out_hash_map
	    }
	],
	operations => [
	    {
		"path" => "/api/contact_source/",
		"method" => "GET",
		"summary" => "Get Contact Sources",
		"notes" => "List all Contact Sources",
		"consumes" => [
			"application/json"
		],
		"parameters" => [ ],
		"responses" => [
			{
				"code" => 200,
				"type" => "source_parameter"
			}
		]
	    },
	    {
		"path" => "/api/contact_source/{contactSourceId}",
		"method" => "GET",
		"summary" => "Get single Contact Source",
		"notes" => "Get exactly one Contact Source",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "contactSourceId",
			"description" => "Contact Source ID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "1"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "source_parameter"
			}
		]
	    },
	    {
		"path" => "/api/contact_source/",
		"method" => "POST",
		"summary" => "Add Source",
		"notes" => "Add Contact Source",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "source_parameter",
			"parameter" => "request",
			"sample" => q({ "contact_source_type" : "salesforce", "description" : "Own Salesforce org", "status": 1, "data" : { "username" : "sfuser@org.com", "password" : "securepass" } })
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "source_parameter"
			}
		]
	    },
	    {
		"path" => "/api/contact_source/{contactSourceId}",
		"method" => "PUT",
		"summary" => "Update Source",
		"notes" => "Update Contact Source record",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "source_parameter",
			"parameter" => "request",
			"sample" => q({ "contact_source_type" : "salesforce", "description" : "Own Salesforce org", "status": 1, "data" : { "username" : "sfuser@org.com", "password" : "securepass" } })
		    },
		    {
			"name" => "contactSourceId",
			"description" => "Contact Source ID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "1"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "source_parameter"
			}
		]
	    },
	    {
		"path" => "/api/contact_source/{contactSourceId}",
		"method" => "DELETE",
		"summary" => "Delete Contact Source",
		"notes" => "Delete Contact Source record",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "contactSourceId",
			"description" => "Contact Source ID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "1"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "source_parameter"
			}
		]
	    }

	]
);

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{version} = $VERSION;

    $self->{mapper} = new Asterisk::API::Mapper::Generic;

    my $out_mapping = $self->{mapper}->prepare( \%out_hash_map );
    $self->{defaults}->{out} = $out_mapping->{defaults};
    $self->{map}->{out} = $out_mapping->{map};
    $self->{type}->{out} = $out_mapping->{type};
    $self->{reverse}->{out} = $out_mapping->{reverse};

    my $in_mapping = $self->{mapper}->prepare( \%in_hash_map );
    $self->{defaults}->{in} = $in_mapping->{defaults};
    $self->{map}->{in} = $in_mapping->{map};
    $self->{type}->{in} = $in_mapping->{type};
    $self->{reverse}->{in} = $in_mapping->{reverse};

    return $self;
}


sub info {
    my ( $self ) = @_;

    my $doc = new Asterisk::API::Documentation::Generic;

    return $doc->prepare_types( \%requests );
}

sub map_out {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{out} } ) { # fill default values
	    $out_hash->{$self->{map}->{out}->{$key}} = $self->{defaults}->{out}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{out}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{out}->{$key} && exists $hash->{$key};
    }

    $out_hash = $self->{mapper}->fix_type($out_hash, $self->{type}->{out}, $self->{reverse}->{out});

    return $out_hash;
}


sub map_in {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{in} } ) { # fill default values
	    $out_hash->{$self->{map}->{in}->{$key}} = $self->{defaults}->{in}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{in}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{in}->{$key} && exists $hash->{$key};
    }

    return $out_hash;
}

1;
