package Asterisk::API::Structures::Dids;

use strict;
use warnings;

use Asterisk::API::Mapper::Generic;
use Asterisk::API::Documentation::Generic;

our $VERSION = "1";

our %out_hash_map = (
  "assigned" => {
    "description" => "Assigned",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "assigned_to",
    "type" => "string"
  },
  "note" => {
    "description" => "Description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Company X main number",
    "transform" => "description",
    "type" => "string"
  },
  "number" => {
    "description" => "Number",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "did",
    "type" => "string"
  }

);

my %in_hash_map = (
  "assigned_to" => {
    "description" => "Assigned to",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "assigned",
    "type" => "string"
  },
  "description" => {
    "description" => "Description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Description text",
    "transform" => "note",
    "type" => "string"
  },
  "did" => {
    "description" => "Did",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "number",
    "type" => "string"
  }

);

our %requests = (
	types => [
	    {
		"type" => "did",
		"structure" => \%out_hash_map
	    }
	],
	operations => [
	    {
		"path" => "/api/dids/",
		"method" => "GET",
		"summary" => "Get DIDs",
		"notes" => "Get all DIDs",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "did"
			}
		]
	    },
	    {
		"path" => "/api/dids/{did}",
		"method" => "GET",
		"summary" => "Get one DID",
		"notes" => "Get one DID",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "did",
			"description" => "DID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "14155551212"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "did"
			}
		]
	    },
	    {
		"path" => "/api/dids/",
		"method" => "POST",
		"summary" => "Add DID",
		"notes" => "Add DID",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "did",
			"parameter" => "request",
			"sample" => q({"assigned_to" : "thirdlane", "did" : "14155551212", "description" : "DID's description" })
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "did"
			}
		]
	    },
	    {
		"path" => "/api/dids/{did}",
		"method" => "PUT",
		"summary" => "Update DID",
		"notes" => "Update DID",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "DID",
			"required" => 1,
			"type" => "did",
			"parameter" => "request",
			"sample" => q({ "description" : "New description"})
		    },
		    {
			"name" => "did",
			"description" => "DID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "14155551212"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "did"
			}
		]
	    },
	    {
		"path" => "/api/dids/{did}",
		"method" => "DELETE",
		"summary" => "Delete DID",
		"notes" => "Delete DID",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "did",
			"description" => "DID",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "14155551212"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "did"
			}
		]
	    }
	]
);

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{version} = $VERSION;

    $self->{mapper} = new Asterisk::API::Mapper::Generic;

    my $out_mapping = $self->{mapper}->prepare( \%out_hash_map );
    $self->{defaults}->{out} = $out_mapping->{defaults};
    $self->{map}->{out} = $out_mapping->{map};
    $self->{type}->{out} = $out_mapping->{type};
    $self->{reverse}->{out} = $out_mapping->{reverse};

    my $in_mapping = $self->{mapper}->prepare( \%in_hash_map );
    $self->{defaults}->{in} = $in_mapping->{defaults};
    $self->{map}->{in} = $in_mapping->{map};
    $self->{type}->{in} = $in_mapping->{type};
    $self->{reverse}->{in} = $in_mapping->{reverse};

    return $self;
}


sub info {
    my ( $self ) = @_;

    my $doc = new Asterisk::API::Documentation::Generic;

    return $doc->prepare_types( \%requests );
}

sub map_out {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{out} } ) { # fill default values
	    $out_hash->{$self->{map}->{out}->{$key}} = $self->{defaults}->{out}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{out}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{out}->{$key} && exists $hash->{$key};
    }

    $out_hash = $self->{mapper}->fix_type($out_hash, $self->{type}->{out}, $self->{reverse}->{out});

    return $out_hash;
}


sub map_in {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{in} } ) { # fill default values
	    $out_hash->{$self->{map}->{in}->{$key}} = $self->{defaults}->{in}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{in}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{in}->{$key} && exists $hash->{$key};
    }

    return $out_hash;
}

1;
