package Asterisk::API::Structures::SingleHuntList;

use strict;
use warnings;

use Asterisk::API::Mapper::Generic;
use Asterisk::API::Documentation::Generic;

our $VERSION = "1";

our %out_hash_map = (
  "args" => {
    "description" => "Destination on no-answer arguments",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => "sales,10,9000\@default,opts,noanswer",
    "transform" => "args",
    "type" => "string"
  },
  "command" => {
    "description" => "Destination on no-answer",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => "tl-queue-timeout-voicemail",
    "transform" => "command",
    "type" => "string"
  },
  "description" => {
    "description" => "Description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "description",
    "type" => "string"
  },
  "list" => {
    "description" => "Huntlist Item",
    "format" => "json",
    "hidden" => 0,
    "required" => 1,
    "sample" => "[ { \"type\" : \"custom\", \"command\" : \"hangup\" } ]",
    "transform" => "list",
    "type" => "huntlist_item"
  },
  "name" => {
    "description" => "Huntlist Name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => 201,
    "transform" => "name",
    "type" => "string"
  }

);

my %in_hash_map = (
  "args" => {
    "description" => "Destination on no-answer arguments",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => "sales,10,9000\@default,opts,noanswer",
    "transform" => "args",
    "type" => "string"
  },
  "command" => {
    "description" => "Destination on no-answer",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => "tl-queue-timeout-voicemail",
    "transform" => "command",
    "type" => "string"
  },
  "description" => {
    "description" => "Description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "description",
    "type" => "string"
  },
  "list" => {
    "description" => "Huntlist Item",
    "format" => "huntlist_item",
    "hidden" => 0,
    "required" => 1,
    "sample" => "[ { \"type\" : \"custom\", \"command\" : \"hangup\" } ]",
    "transform" => "list",
    "type" => "huntlist_item"
  },
  "name" => {
    "description" => "Huntlist Name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => 201,
    "transform" => "name",
    "type" => "string"
  }

);

our %out_hash_map_list_item = (
	extensions => "extensions",
	options => "options",
	timeout => "timeout",
	type => {
	    type => "string",
	    format => "one of [devices, ringgroup, queue, custom]",
	    description => "Huntlist Item",
	    required => 1,
	    hidden => 0,
	    transform => "type",
	    sample => "custom"
	},
	phones => {
	    type => "string",
	    format => "raw",
	    description => "Phones",
	    required => 1,
	    hidden => 0,
	    transform => "phones",
	    sample => ""
	},
	queues => {
	    type => "string",
	    format => "raw",
	    description => "Queues",
	    required => 1,
	    hidden => 0,
	    transform => "queues",
	    sample => ""
	},
	ogm => {
	    type => "string",
	    format => "raw",
	    description => "Announcement to caller",
	    required => 1,
	    hidden => 0,
	    transform => "ogm",
	    sample => "sampleopen"
	},
	command => {
	    type => "string",
	    format => "raw",
	    description => "Command",
	    required => 1,
	    hidden => 0,
	    transform => "command",
	    sample => ""
	}
);

our %requests = (
	types => [
	    {
		"type" => "huntlist",
		"structure" => \%out_hash_map
	    },
	    {
		"type" => "huntlist_item",
		"structure" => \%out_hash_map_list_item
	    }
	],
	operations => [
	    {
		"path" => "/api/huntlist/",
		"method" => "GET",
		"summary" => "Get Huntlist",
		"notes" => "Get all Huntlist",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "huntlist"
			}
		]
	    },
	    {
		"path" => "/api/huntlist/{name}",
		"method" => "GET",
		"summary" => "Get Huntlist parameters",
		"notes" => "Get single Huntlist parameters",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "name",
			"description" => "Huntlist Name",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "10"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "huntlist"
			}
		]
	    },
	    {
		"path" => "/api/huntlist/",
		"method" => "POST",
		"summary" => "Add Huntlist",
		"notes" => "Add new Huntlist",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "huntlist",
			"parameter" => "request",
			"sample" => q(
{
    "name" : "1",
    "args" : "sales,10,9000@default,opts,tpr",
    "command" : "tl-queue-timeout-voicemail",
    "description" : "desc",
    "list" : [
	{
	    "extensions" : "SIP/9000",
	    "options" : "i",
	    "timeout" : "10",
	    "type" : "devices",
	    "phones" : "076",
	    "queues" : "",
	    "ogm" : "sampleopen",
	    "command" : ""
	}
    ]
}
)
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "huntlist"
			}
		]
	    },
	    {
		"path" => "/api/huntlist/{name}",
		"method" => "PUT",
		"summary" => "Update single Huntlist",
		"notes" => "Update single Huntlist",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "huntlist",
			"parameter" => "request",
			"sample" => q(
{
    "name" : "1",
    "args" : "sales,10,9000@default,opts,tpr",
    "command" : "tl-queue-timeout-voicemail",
    "list" : [
	{
	    "extensions" : "SIP/9001",
	    "options" : "i",
	    "timeout" : "10",
	    "type" : "devices",
	    "phones" : "076",
	    "queues" : "",
	    "ogm" : "sampleopen",
	    "command" : ""
	}
    ]
}
)
		    },
		    {
			"name" => "name",
			"description" => "Huntlist Name",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "10"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "huntlist"
			}
		]
	    },
	    {
		"path" => "/api/huntlist/{name}",
		"method" => "DELETE",
		"summary" => "Delete single Huntlist",
		"notes" => "Delete one Huntlist",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "name",
			"description" => "Huntlist Name",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "10"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "huntlist"
			}
		]
	    }
	]
);

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{version} = $VERSION;

    $self->{mapper} = new Asterisk::API::Mapper::Generic;

    my $out_mapping = $self->{mapper}->prepare( \%out_hash_map );
    $self->{defaults}->{out} = $out_mapping->{defaults};
    $self->{map}->{out} = $out_mapping->{map};
    $self->{type}->{out} = $out_mapping->{type};
    $self->{reverse}->{out} = $out_mapping->{reverse};

    my $in_mapping = $self->{mapper}->prepare( \%in_hash_map );
    $self->{defaults}->{in} = $in_mapping->{defaults};
    $self->{map}->{in} = $in_mapping->{map};
    $self->{type}->{in} = $in_mapping->{type};
    $self->{reverse}->{in} = $in_mapping->{reverse};

    return $self;
}


sub info {
    my ( $self ) = @_;

    my $doc = new Asterisk::API::Documentation::Generic;

    return $doc->prepare_types( \%requests );
}

sub map_out {
    my ( $self, $hash, $params ) = @_;
#return $hash;
    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{out} } ) { # fill default values
	    $out_hash->{$self->{map}->{out}->{$key}} = $self->{defaults}->{out}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{out}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{out}->{$key} && exists $hash->{$key};
    }

    $out_hash = $self->{mapper}->fix_type($out_hash, $self->{type}->{out}, $self->{reverse}->{out});

    return $out_hash;
}


sub map_in {
    my ( $self, $hash, $params ) = @_;
return $hash;
    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{in} } ) { # fill default values
	    $out_hash->{$self->{map}->{in}->{$key}} = $self->{defaults}->{in}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{in}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{in}->{$key} && exists $hash->{$key};
    }

    return $out_hash;
}

sub in_to_out {
    my ($self, $hash) = @_;

    $hash = [ $hash ] if ( ( defined $hash ) && ( ref $hash eq "HASH" ) );

    my $out_hash = [ ];

    foreach my $list_item ( @{ $hash } ) {

	my $items = { };
	my $list = { };
	my $command = $list_item->{command};
	$command = $list_item->{action} if ( ( $list_item->{action} ) && ( !$command ) );
	$command =~ s/,$//sg;
	my $args = $list_item->{args};

	foreach my $item ( keys %{ $list_item } ) {
	    next if $item =~ m/\Q$command/;
	    next if $list_item->{$item} =~ m/\Q$command/;
	    if ( my ($item_name, $item_seq) = $item =~ m/^(.*?)_(\d)$/ ) {
		$item_name =~ s/^gtype$/type/;
		$items->{$item_seq}->{$item_name} = $list_item->{$item};
	    }
	}

	$list->{command} = $command;
	$list->{args} = $args;
	$list->{name} = $list_item->{name};
	$list->{description} = $list_item->{description} if exists $list_item->{description};

	foreach my $item ( sort {$a <=> $b} keys %{ $items } ) {
	    push @{ $list->{list} }, $items->{$item};
	}

	push @{ $out_hash }, $list;
    }

    $out_hash = $out_hash->[0] if scalar @{$out_hash} == 1;

    return $out_hash;
}

sub out_to_in {
    my ($self, $hash) = @_;

    my $list = $hash->{list};
    my $list_iter = 0;
    my $list_trans = { };

    my $out_hash = { };

    foreach my $list_item ( @{ $list } ) {
	foreach my $item_name ( keys %{ $list_item } ) {
	    my $item_name_trans = $item_name;
	    $item_name_trans =~ s/^type$/gtype/;
	    $list_trans->{$item_name_trans . "_" . $list_iter} = $list_item->{$item_name};
	}
	$list_iter++;
    }

    $list_trans->{command} = $hash->{command};
    $list_trans->{action_0} = $hash->{command};
    $list_trans->{args} = $hash->{args};
    $list_trans->{name} = $hash->{name};
    $list_trans->{description} = $hash->{description} if exists $hash->{description};

    my $arg_iter = 0;
    foreach my $arg ( split ",", $list_trans->{args} ) {
	$list_trans->{$list_trans->{command} . "_0_". $arg_iter++} = $arg;
    }

    return $list_trans;
}

1;
