package Asterisk::API::Structures::SinglePickupGroup;

use strict;
use warnings;

use Asterisk::API::Mapper::Generic;
use Asterisk::API::Documentation::Generic;

our $VERSION = "1";

our %out_hash_map = (
  "description" => {
    "description" => "Pickup Group Description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "description",
    "type" => "string"
  },
  "id" => {
    "description" => "Pickup Id",
    "format" => "int64",
    "hidden" => 1,
    "required" => 0,
    "sample" => 1,
    "transform" => "id",
    "type" => "integer"
  },
  "name" => {
    "description" => "Pickup Group Name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => "pickup1",
    "transform" => "name",
    "type" => "string"
  },
  "options" => {
    "description" => "Pickup Group Options",
    "format" => "raw",
    "hidden" => 1,
    "required" => 0,
    "sample" => "",
    "transform" => "options",
    "type" => "string"
  },
  "phones" => {
    "description" => "Pickup Group Phones",
    "format" => "comma separated devices",
    "hidden" => 0,
    "required" => 0,
    "sample" => "SIP/9000,SIP/9001",
    "transform" => "phones",
    "type" => "string"
  }

);

my %in_hash_map = (
  "description" => {
    "description" => "Pickup Group Description",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "",
    "transform" => "description",
    "type" => "string"
  },
  "id" => {
    "description" => "Pickup Id",
    "format" => "int64",
    "hidden" => 1,
    "required" => 0,
    "sample" => 1,
    "transform" => "id",
    "type" => "integer"
  },
  "name" => {
    "description" => "Pickup Group Name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 1,
    "sample" => "pickup1",
    "transform" => "name",
    "type" => "string"
  },
  "options" => {
    "description" => "Pickup Group Options",
    "format" => "raw",
    "hidden" => 1,
    "required" => 0,
    "sample" => "",
    "transform" => "options",
    "type" => "string"
  },
  "phones" => {
    "description" => "Pickup Group Phones",
    "format" => "comma separated devices",
    "hidden" => 0,
    "required" => 0,
    "sample" => "SIP/9000,SIP/9001",
    "transform" => "phones",
    "type" => "string"
  }

);

our %requests = (
	types => [
	    {
		"type" => "pickupgroup",
		"structure" => \%out_hash_map
	    }
	],
	operations => [
	    {
		"path" => "/api/pickupgroup/",
		"method" => "GET",
		"summary" => "Get Pickup Groups",
		"notes" => "Get all Pickup Groups",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "pickupgroup"
			}
		]
	    },
	    {
		"path" => "/api/pickupgroup/{name}",
		"method" => "GET",
		"summary" => "Get Pickup Group parameters",
		"notes" => "Get single Pickup Group parameters",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "name",
			"description" => "Pickup Group Name",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "pickup0-thirdlane"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "pickupgroup"
			}
		]
	    },
	    {
		"path" => "/api/pickupgroup/",
		"method" => "POST",
		"summary" => "Add Pickup Group",
		"notes" => "Add new Pickup Group",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "Pickup Group",
			"parameter" => "request",
			"sample" => q({ "name" : "pickup0", "description" : "pickup group 0", "phones" : "SIP/9000,SIP/9001" })
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "pickupgroup"
			}
		]
	    },
	    {
		"path" => "/api/pickupgroup/{name}",
		"method" => "PUT",
		"summary" => "Update single Pickup Group",
		"notes" => "Update single Pickup Group",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "body",
			"description" => "No description",
			"required" => 1,
			"type" => "Pickup Group",
			"parameter" => "request",
			"sample" => q({ "description" : "changed description", "phones" : "SIP/9000,SIP/9001" })
		    },
		    {
			"name" => "name",
			"description" => "Pickup Group Name",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "pickup0-thirdlane"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "pickupgroup"
			}
		]
	    },
	    {
		"path" => "/api/pickupgroup/{name}",
		"method" => "DELETE",
		"summary" => "Delete single Pickup Group",
		"notes" => "Delete one Pickup Group",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "name",
			"description" => "Pickup Group Name",
			"required" => 1,
			"type" => "string",
			"parameter" => "path",
			"sample" => "pickup0-thirdlane"
		    }
			],
		"responses" => [
			{
				"code" => 200,
				"type" => "pickupgroup"
			}
		]
	    }
	]
);

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{version} = $VERSION;

    $self->{mapper} = new Asterisk::API::Mapper::Generic;

    my $out_mapping = $self->{mapper}->prepare( \%out_hash_map );
    $self->{defaults}->{out} = $out_mapping->{defaults};
    $self->{map}->{out} = $out_mapping->{map};
    $self->{type}->{out} = $out_mapping->{type};
    $self->{reverse}->{out} = $out_mapping->{reverse};

    my $in_mapping = $self->{mapper}->prepare( \%in_hash_map );
    $self->{defaults}->{in} = $in_mapping->{defaults};
    $self->{map}->{in} = $in_mapping->{map};
    $self->{type}->{in} = $in_mapping->{type};
    $self->{reverse}->{in} = $in_mapping->{reverse};

    return $self;
}


sub info {
    my ( $self ) = @_;

    my $doc = new Asterisk::API::Documentation::Generic;

    return $doc->prepare_types( \%requests );
}

sub map_out {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{out} } ) { # fill default values
	    $out_hash->{$self->{map}->{out}->{$key}} = $self->{defaults}->{out}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{out}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{out}->{$key} && exists $hash->{$key};
    }

    $out_hash = $self->{mapper}->fix_type($out_hash, $self->{type}->{out}, $self->{reverse}->{out});

    return $out_hash;
}


sub map_in {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{in} } ) { # fill default values
	    $out_hash->{$self->{map}->{in}->{$key}} = $self->{defaults}->{in}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{in}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{in}->{$key} && exists $hash->{$key};
    }

    return $out_hash;
}

1;
