package Asterisk::API::Structures::VoiceMail;

use strict;
use warnings;

use Asterisk::API::Mapper::Generic;
use Asterisk::API::Documentation::Generic;

our $VERSION = "1";

our %out_hash_map = (
  "caller_id_number" => {
    "description" => "CallerID number",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 2003,
    "transform" => "caller_id_number",
    "type" => "string"
  },
  "caller_name" => {
    "description" => "CallerID name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Foo Bar",
    "transform" => "caller_name",
    "type" => "string"
  },
  "download_url" => {
    "description" => "Voice message download URL",
    "format" => "URL",
    "hidden" => 0,
    "required" => 0,
    "sample" => "https://localhost:10000/asterisk/get_message_from_db.cgi?id=1&download=true",
    "transform" => "download_url",
    "type" => "string"
  },
  "duration" => {
    "description" => "Message duration",
    "format" => "M:S",
    "hidden" => 0,
    "required" => 0,
    "sample" => "1:28",
    "transform" => "duration",
    "type" => "integer"
  },
  "folder" => {
    "description" => "Message folder",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "INBOX",
    "transform" => "folder",
    "type" => "string"
  },
  "msgid" => {
    "description" => "Message ID",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 1,
    "transform" => "id",
    "type" => "string"
  },
  "origdate" => {
    "description" => "Issue datetime",
    "format" => "ctime",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Sat Aug 23 12:59:42 PM EDT 2014",
    "transform" => "orig_date",
    "type" => "string"
  },
  "origdate_epoch" => {
    "description" => "UTC epoch value",
    "format" => "unixtime",
    "hidden" => 0,
    "required" => 0,
    "sample" => "1325376000",
    "transform" => "orig_date_epoch",
    "type" => "integer"
  },
  "play_url" => {
    "description" => "Voice message play URL",
    "format" => "URL",
    "hidden" => 0,
    "required" => 0,
    "sample" => "https://localhost:10000/asterisk/get_message_from_db.cgi?id=1",
    "transform" => "play_url",
    "type" => "string"
  }

);

my %in_hash_map = (
  "caller_id_number" => {
    "description" => "CallerID number",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 2003,
    "transform" => "caller_id_number",
    "type" => "string"
  },
  "caller_name" => {
    "description" => "CallerID name",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Foo Bar",
    "transform" => "caller_name",
    "type" => "string"
  },
  "download_url" => {
    "description" => "Voice message download URL",
    "format" => "URL",
    "hidden" => 0,
    "required" => 0,
    "sample" => "https://localhost:10000/asterisk/get_message_from_db.cgi?id=1&download=true",
    "transform" => "download_url",
    "type" => "string"
  },
  "duration" => {
    "description" => "Message duration",
    "format" => "M:S",
    "hidden" => 0,
    "required" => 0,
    "sample" => "1:28",
    "transform" => "duration",
    "type" => "integer"
  },
  "folder" => {
    "description" => "Message folder",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => "INBOX",
    "transform" => "folder",
    "type" => "string"
  },
  "id" => {
    "description" => "Message ID",
    "format" => "raw",
    "hidden" => 0,
    "required" => 0,
    "sample" => 1,
    "transform" => "msgid",
    "type" => "string"
  },
  "orig_date" => {
    "description" => "Issue datetime",
    "format" => "ctime",
    "hidden" => 0,
    "required" => 0,
    "sample" => "Sat Aug 23 12:59:42 PM EDT 2014",
    "transform" => "origdate",
    "type" => "string"
  },
  "play_url" => {
    "description" => "Voice message play URL",
    "format" => "URL",
    "hidden" => 0,
    "required" => 0,
    "sample" => "https://localhost:10000/asterisk/get_message_from_db.cgi?id=1",
    "transform" => "play_url",
    "type" => "string"
  }

);

our %requests = (
	types => [
	    {
		"type" => "voicemail",
		"structure" => \%out_hash_map
	    }
	],
	operations => [
	    {
		"path" => "/api/voicemail/",
		"method" => "GET",
		"summary" => "Get voicemail messages list",
		"notes" => "Get voicemail messages list",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "voicemail"
			}
		]
	    },
	    {
		"path" => "/api/voicemail/{messageId}",
		"method" => "GET",
		"summary" => "Get one voicemail message",
		"notes" => "Get one voicemail message",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "messageId",
			"description" => "Message ID",
			"required" => 1,
			"type" => "integer",
			"parameter" => "path"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "voicemail"
			}
		]
	    },
	    {
		"path" => "/api/voicemail/{messageId}",
		"method" => "DELETE",
		"summary" => "Delete voicemail message",
		"notes" => "Delete voicemail message",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "messageId",
			"description" => "Message ID",
			"required" => 1,
			"type" => "integer",
			"parameter" => "path"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "voicemail"
			}
		]
	    },
	    {
		"path" => "/api/voicemail/{messageId}",
		"method" => "PATCH",
		"summary" => "Mark voicemail message as readed",
		"notes" => "Mark voicemail message as readed",
		"consumes" => [
			"application/json"
		],
		"parameters" => [
		    {
			"name" => "messageId",
			"description" => "Message ID",
			"required" => 1,
			"type" => "integer",
			"parameter" => "path"
		    }
		],
		"responses" => [
			{
				"code" => 200,
				"type" => "voicemail"
			}
		]
	    }
	]
);

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{version} = $VERSION;

    $self->{mapper} = new Asterisk::API::Mapper::Generic;

    my $out_mapping = $self->{mapper}->prepare( \%out_hash_map );
    $self->{defaults}->{out} = $out_mapping->{defaults};
    $self->{map}->{out} = $out_mapping->{map};
    $self->{type}->{out} = $out_mapping->{type};
    $self->{reverse}->{out} = $out_mapping->{reverse};

    my $in_mapping = $self->{mapper}->prepare( \%in_hash_map );
    $self->{defaults}->{in} = $in_mapping->{defaults};
    $self->{map}->{in} = $in_mapping->{map};
    $self->{type}->{in} = $in_mapping->{type};
    $self->{reverse}->{in} = $in_mapping->{reverse};

    return $self;
}


sub info {
    my ( $self ) = @_;

    my $doc = new Asterisk::API::Documentation::Generic;

    return $doc->prepare_types( \%requests );
}

sub map_out {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{out} } ) { # fill default values
	    $out_hash->{$self->{map}->{out}->{$key}} = $self->{defaults}->{out}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{out}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{out}->{$key} && exists $hash->{$key};
    }

    $out_hash = $self->{mapper}->fix_type($out_hash, $self->{type}->{out}, $self->{reverse}->{out});

    return $out_hash;
}


sub map_in {
    my ( $self, $hash, $params ) = @_;

    if ( exists $params->{direct} && $params->{direct} ) {
	return $hash;
    }

    my $out_hash = { };

    if ( exists $params->{defaults} && $params->{defaults} ) {
	foreach my $key ( keys %{ $self->{defaults}->{in} } ) { # fill default values
	    $out_hash->{$self->{map}->{in}->{$key}} = $self->{defaults}->{in}->{$key};
	}
    }

    foreach my $key ( keys %{ $hash } ) {
	$out_hash->{$self->{map}->{in}->{$key}} = $hash->{$key}
	    if exists $self->{map}->{in}->{$key} && exists $hash->{$key};
    }

    return $out_hash;
}

1;
