package Integration::CRM::CLI;

use strict;
use warnings;


use JSON::XS;
use Types::Serialiser;

use IPC::Open3;
use Data::Dumper;
use POSIX qw(strftime);

our $VERSION = '1';

our $DEBUG = 0;

my %object_prefixes = (
	ACCOUNT                     => '001',
	CONTACT                     => '003',
	LEAD                        => '00Q',
	TASK                        => '00T'
);

sub new {
    my( $class, $args ) = @_;
    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    $self->{config}->{script} = '/opt/thirdlane/perl_lib/crmapi.pl' unless $self->{config}->{script};

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    die "connection error" unless $self->_connect;

    return $self;
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub conf {
    my $self = shift;

    return $self->{config};
}

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
    print Data::Dumper->Dump([@_]);
}

sub _command {
    my ( $self, $cmd, $args ) = @_;

    my $pid = open3( \*WRITE, \*READ, 0, $self->conf->{script}, $cmd ); #if \*ERROR is false(arg:3), STDERR is sent to STDOUT

    my $packet = $self->json->encode($args);

    print "Sent packet:".$self->dump($packet) if $DEBUG;

    print WRITE "$packet\n";

    my $answer = <READ>;

    print "Answer:".$self->dump($answer) if $DEBUG;

    eval {
	$answer = $self->json->decode($answer);
    };

    print "Decode error:".$@ if $@ && $DEBUG;

    print "Decoded answer:".$self->dump($answer) if $DEBUG;

    return $answer;
}

sub _connect {
    my $self = shift;

    return $self->_command("ping");
}

sub get_leads {
    my $self = shift;
    return $self->find('Leads', { all => 1 });
}

sub get_contacts {
    my $self = shift;
    return $self->find('Contacts', { all => 1 });
}

sub get_calls {
    my $self = shift;
    return $self->find('Calls', { all => 1 });
}

sub detect_type_by_id {
    my ( $self, $id ) = @_;

    my $records = $self->find('Leads', { Id => $id });
    return 'LEAD' if scalar @{ $records };

    $records = $self->find('Contacts', { Id => $id });
    return 'CONTACT' if scalar @{ $records };

    $records = $self->find('Calls', { Id => $id });
    return 'TASK' if scalar @{ $records };

    $records = $self->find('Accounts', { Id => $id });
    return 'ACCOUNT' if scalar @{ $records };

    return 'UNKNOWN';
}

sub get_fields {
    my ( $self, $object ) = @_;

    return $self->_command("fields", { object => $object });
}

sub update {
    my( $self, $args ) = @_;

    return $self->_command("upsert", { object => $args->{object}, record => $args->{record} });
}

sub insert {
    my( $self, $args ) = @_;

    return $self->_command("upsert", { object => $args->{object}, record => $args->{record} });
}

sub upsert {
    my( $self, $args ) = @_;

    return $self->update($args) if exists $args->{record}->{Id};
    return $self->insert($args);
}

sub get_record_by_id {
    my( $self, $object, $id ) = @_;
    return $self->find($object, { Id => $id });
}

sub find {
    my( $self, $object, $args ) = @_;

    my $response = $self->_command("find", { object => $object, args => $args });

    return undef unless $response;
    return undef if ( ( ref $response eq "HASH" ) && ( exists $response->{error} ) );

    return $response;
}

sub query {
    my( $self, $object, $args ) = @_;
    return $self->find($object, $args);
}

sub query_all {
    my( $self, $query ) = @_;

    my $leads = $self->find('Leads', { FirstName => $query, LastName => $query, Email => $query, Phone => $query });
    my $contacts = $self->find('Contacts', { FirstName => $query, LastName => $query, Email => $query, Phone => $query } );

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub find_phone {
    my( $self, $phone ) = @_;

    my $leads = $self->find('Leads', { Phone => $phone });
    my $contacts = $self->find('Contacts', { Phone => $phone } );

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub query_tasks_by_rec_id {
    my( $self, $rec_id ) = @_;

    return $self->find('Calls', { Call => $rec_id });
}

sub query_tasks_by_id {
    my( $self, $id ) = @_;

    return $self->find('Calls', { Id => $id });
}

sub query_leads_by_id {
    my( $self, $id ) = @_;

    return $self->find('Leads', { Id => $id });
}

sub query_contacts_by_id {
    my( $self, $id ) = @_;

    return $self->find('Contacts', { Id => $id });
}

sub query_accounts_by_id {
    my( $self, $id ) = @_;

    return $self->find('Accounts', { Id => $id });
}

1;
