package Integration::CRM::Freshdesk;

use strict;
use warnings;


use JSON::XS;
use Types::Serialiser;

use LWP::UserAgent;
use HTTP::Request;
use URI::Encode;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);

our $VERSION = '2';

our $DEBUG = 0;

my %object_prefixes = (
	ACCOUNT                     => '001',
	CONTACT                     => '003',
	LEAD                        => '00Q',
	TASK                        => '00T'
);

sub new {
    my( $class, $args ) = @_;
    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    if ( $self->conf->{user}->{token} ) {
	$self->conf->{user}->{login} = $self->conf->{user}->{token};
	$self->conf->{user}->{password} = "X";
    }

    $self->{id} = $self->conf->{id} || 0;

    $self->{ua} = LWP::UserAgent->new(ssl_opts => { verify_hostname => 0 });

    $self->{ua}->timeout(10);
    $self->{ua}->env_proxy;

    $self->{uri_encode} = URI::Encode->new( { encode_reserved => 1 } );

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    die "no url" unless $args->{url};

    die "connection error" unless $self->_connect;

    return $self;
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub _w {
    my $cond = shift;
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($cond);
}

sub ua {
    my $self = shift;

    $self->{ua}->credentials( $self->conf->{url} . ":443", 'Freshdesk', $self->conf->{user}->{login}, $self->conf->{user}->{password} );

    return $self->{ua};
}

sub get {
    my $self = shift;
    my $url = shift;

    my $req = HTTP::Request->new(GET => $url);
    $req->authorization_basic($self->conf->{user}->{login}, $self->conf->{user}->{password});

    my $response = $self->ua->request($req);

    if ($response->is_success) {
	print $response->decoded_content if $DEBUG;
    }
    else {
	die $response->status_line;
    }

    my $decoded_response = eval { $self->json->decode($response->decoded_content) };

    return undef if $@;
    return $decoded_response;
}

sub conf {
    my $self = shift;

    return $self->{config};
}

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
    print Data::Dumper->Dump([@_]);
}

sub _connect {
    my $self = shift;

    my $url = sprintf ('https://%s/api/v2/tickets',
	$self->conf->{url}
    );

    my $response = $self->get($url);

    return $self if ref $response eq "ARRAY";

    return undef if $@;
    return undef unless exists $response->{code};

    return $self;
}

sub get_by_phone {
    my $self = shift;
    my $type = shift;
    my $phone = shift;

    my $url = sprintf ('https://%s/api/v2/contacts.json?%s=%s',
	$self->conf->{url},
	$type,
	_w($phone)
    );

    my $decoded_response = $self->get($url);

    my @result;

    if ( ref $decoded_response eq "ARRAY" ) {
	foreach my $record ( @{ $decoded_response } ) {
	    push @result, $self->_convert_record_contact($record);
	}
    }

    return \@result;
}

sub _format_url_args {
    my ( $self, $args) = @_;

    my $result;

    foreach my $arg ( keys %{ $args }) {
	$arg =~ s{\"}{}sg;
	push @{ $result }, sprintf("%s=%s", $arg, $args->{$arg});
    }

    return join "&", @{ $result };
}

sub _convert_record_contact {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => $object->{'name'},
	'LastName' => '',
	'Description' => $object->{'description'},
	'Email' => $object->{'email'},
	'Phone' => $object->{'phone'},
	'Mobile' => $object->{'mobile'},
	'Department' => '',
	'Id' => [ $object->{'id'} ],
	'type' => 'Contact',
	'contact_source' => 'freshdesk',
	'contact_source_contact_id' => $object->{'id'},
	'contact_source_type' => 'Freshdesk',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub find_phone {
    my( $self, $phone ) = @_;

    my @result;

    my $contacts = $self->get_by_phone('phone', $phone);
    push @result, @{$contacts};

    $contacts = $self->get_by_phone('mobile', $phone);
    push @result, @{$contacts};

    if ( my ( $phone_without_lp ) = $phone =~ m/^\+(.*?)$/ ) {
	$contacts = $self->get_by_phone('phone', $phone_without_lp);
	push @result, @{$contacts};

	$contacts = $self->get_by_phone('mobile', $phone_without_lp);
	push @result, @{$contacts};
    } else {
	$contacts = $self->get_by_phone('phone', "+" . $phone);
	push @result, @{$contacts};

	$contacts = $self->get_by_phone('mobile', "+" . $phone);
	push @result, @{$contacts};
    }

    my %result = map { $_->{Id}->[0] => $_ } @result;
    @result = values %result;

    return \@result;
}

sub query_all {
    my( $self, $query ) = @_;

    my $result = $self->find_phone($query);

    return $result;
}

1;
