package Integration::CRM::PipeDrive;

use strict;
use warnings;


use JSON::XS;
use Types::Serialiser;
use HTTP::Request::Common;
use LWP::UserAgent;
use HTTP::Request;
use URI::Encode;
use URI;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);

our $VERSION = '1';
our $DEBUG = 0;

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    $self->{id} = $self->conf->{id} || 0;

    $self->{ua} = LWP::UserAgent->new(ssl_opts => { verify_hostname => 0 });

    $self->{ua}->timeout(10);
    $self->{ua}->env_proxy;

    $self->{uri_encode} = URI::Encode->new( { encode_reserved => 1 } );

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    $self->{crm_domain} = "pipedrive.com";
    $self->{crm_api_subdomain} = "api";

    $self->{url} = sprintf("https://%s.%s", $self->{crm_api_subdomain}, $self->{crm_domain});

    $self->{api_version} = "v1";

    $self->pdump(['Creating Pipedrive:', $args]) if $DEBUG;

    die "connection error" unless $self->_connect;

    return $self;
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub _w {
    my $cond = shift;
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($cond);
}

sub ua {
    my $self = shift;

    return $self->{ua};
}

sub get {
    my ( $self, $url, $query, $additional_data ) = @_;

    my $uri = URI->new($self->{url});
    $uri->path_segments("v1", split /\//, $url);
    $query->{api_token} = $self->conf->{user}->{token} if exists $self->conf->{user}->{token} && defined $self->conf->{user}->{token} && length $self->conf->{user}->{token};
    $uri->query_form($query) if defined $query;

    printf("Getting URL: %s\n", $uri->as_string) if $DEBUG;

    my $req = HTTP::Request::Common::GET($uri->as_string);

    my $response = $self->ua->request($req);

    if ($response->is_success) {
	print $response->decoded_content if $DEBUG;
    }
    else {
	die $response->status_line;
    }

    my $decoded_response = eval { $self->json->decode($response->decoded_content) };

    return undef if $@;

    if ( defined $decoded_response && ref $decoded_response eq "HASH" && exists $decoded_response->{success} && $decoded_response->{success} ) {
	if ( exists $decoded_response->{data} ) {
#	    $self->pdump($decoded_response->{data}) if $DEBUG;
	    if ( $additional_data ) {
		return $decoded_response;
	    }
	    return $decoded_response->{data};
	} else {
#	    $self->pdump($decoded_response) if $DEBUG;
	    return $decoded_response;
	}
    }

    return undef;
}

sub post {
    my ( $self, $url, $query, $content, $additional_data  ) = @_;

    my $uri = URI->new($self->{url});
    $uri->path_segments("v1", split /\//, $url);

    $query->{api_token} = $self->conf->{user}->{token} if exists $self->conf->{user}->{token} && defined $self->conf->{user}->{token} && length $self->conf->{user}->{token};
    $uri->query_form($query) if defined $query;

    printf("Posting URL: %s\n", $uri->as_string) if $DEBUG;

    $self->pdump($content) if $DEBUG;

    my $req = HTTP::Request::Common::POST($uri->as_string, Content => $content);

    $self->pdump($req) if $DEBUG;

    my $response = $self->ua->request($req);

    $self->pdump($response) if $DEBUG;

    if ($response->is_success) {
	print $response->decoded_content if $DEBUG;
    }
    else {
	die $response->status_line;
    }

    my $decoded_response = eval { $self->json->decode($response->decoded_content) };

    return undef if $@;

    if ( defined $decoded_response && ref $decoded_response eq "HASH" && exists $decoded_response->{success} && $decoded_response->{success} ) {
	if ( exists $decoded_response->{data} ) {
#	    $self->pdump($decoded_response->{data}) if $DEBUG;
	    if ( $additional_data ) {
		return $decoded_response;
	    }
	    return $decoded_response->{data};
	} else {
#	    $self->pdump($decoded_response) if $DEBUG;
	    return $decoded_response;
	}
    }

    return undef;
}

sub put {
    my ( $self, $url, $query, $content, $additional_data  ) = @_;

    my $uri = URI->new($self->{url});
    $uri->path_segments("v1", split /\//, $url);

    $query->{api_token} = $self->conf->{user}->{token} if exists $self->conf->{user}->{token} && defined $self->conf->{user}->{token} && length $self->conf->{user}->{token};
    $uri->query_form($query) if defined $query;

    printf("Putting URL: %s\n", $uri->as_string) if $DEBUG;

    $self->pdump($content) if $DEBUG;

    my $req = HTTP::Request::Common::PUT($uri->as_string, Content => $content);

    $self->pdump($req) if $DEBUG;

    my $response = $self->ua->request($req);

    $self->pdump($response) if $DEBUG;

    if ($response->is_success) {
	print $response->decoded_content if $DEBUG;
    }
    else {
	die $response->status_line;
    }

    my $decoded_response = eval { $self->json->decode($response->decoded_content) };

    return undef if $@;

    if ( defined $decoded_response && ref $decoded_response eq "HASH" && exists $decoded_response->{success} && $decoded_response->{success} ) {
	if ( exists $decoded_response->{data} ) {
#	    $self->pdump($decoded_response->{data}) if $DEBUG;
	    if ( $additional_data ) {
		return $decoded_response;
	    }
	    return $decoded_response->{data};
	} else {
#	    $self->pdump($decoded_response) if $DEBUG;
	    return $decoded_response;
	}
    }

    return undef;
}

sub conf {
    my $self = shift;

    return $self->{config};
}

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
    print Data::Dumper->Dump([@_]);
}

sub _connect {
    my $self = shift;

    $self->pdump(['Connecting Pipedrive', $self->conf() ]) if $DEBUG;

    if ( exists $self->conf->{user}->{login} && defined $self->conf->{user}->{login} && length $self->conf->{user}->{login} &&
	exists $self->conf->{user}->{password} && defined $self->conf->{user}->{password} && length $self->conf->{user}->{password}
    ) {
	my $response = $self->post("authorizations", { }, [ email => $self->conf->{user}->{login}, password => $self->conf->{user}->{password} ]);

	if ( defined $response && ref $response eq "ARRAY" && scalar @{ $response } && exists $response->[0]->{api_token} && defined $response->[0]->{api_token} && length $response->[0]->{api_token} ) {
	    $self->conf->{user}->{token} = $response->[0]->{api_token};
	}
    }

    return undef unless exists $self->conf->{user}->{token} && defined $self->conf->{user}->{token} && length $self->conf->{user}->{token};

    my $response = $self->get("userSettings");

    return undef unless defined $response;

    $response = $self->get("users/self");

    if ( defined $response ) {
	$self->{crm_user} = $response;
	
	if ( exists $response->{company_id} ) {
	    if ( exists $response->{companies} && exists $response->{companies}->{ $response->{company_id} } ) {
		$self->{crm_subdomain} = $response->{companies}->{ $response->{company_id} }->{domain};
		printf("received subdomain: %s\n", $self->{crm_subdomain}) if $DEBUG;
	    }
	}
    }

    return $self;
}

sub get_by_phone {
    my $self = shift;
    my $type = shift;
    my $phone = shift;

    my $url = sprintf ('https://%s/api/v2/contacts.json?%s=%s',
	$self->conf->{url},
	$type,
	_w($phone)
    );

    my $decoded_response = $self->get($url);

    my @result;

    if ( ref $decoded_response eq "ARRAY" ) {
	foreach my $record ( @{ $decoded_response } ) {
	    push @result, $self->_convert_record_contact($record);
	}
    }

    return \@result;
}

sub _format_url_args {
    my ( $self, $args) = @_;

    my $result;

    foreach my $arg ( keys %{ $args }) {
	$arg =~ s{\"}{}sg;
	push @{ $result }, sprintf("%s=%s", $arg, $args->{$arg});
    }

    return join "&", @{ $result };
}

sub detect_type_by_id {
    my ( $self, $id ) = @_;

    return "Contacts";
}

sub get_field_list {
    my ( $self, $object ) = @_;

    my $response;

    if ( $object eq 'Calls' ) {
	$response = $self->get("activityFields");
    }
    elsif ( $object eq 'Contacts' ) {
	$response = $self->get("personFields");
    }
    elsif ( $object eq 'Leads' ) {
	$response = $self->get("personFields");
    }
    elsif ( $object eq 'Accounts' ) {
	$response = $self->get("organizationFields");
    }
    else {
	return [ ];
    }

    return [ ] unless defined $response;

    return $response;
}

sub get_fields {
    my ( $self, $object ) = @_;

    my @fields;

    foreach my $field ( @{ $self->get_field_list($object) } ) {
	push @fields, $field->{key};
    }

    return \@fields;
}

sub get_url {
    my( $self, $id ) = @_;

    return sprintf("https://%s.%s/person/%s", $self->{crm_subdomain}, $self->{crm_domain}, $id);
}

sub fix_object {
    my( $self, $object ) = @_;

    if (exists $object->{id}) {
	$object->{contact_source} = $self->{id};
	$object->{contact_source_type} = "Pipedrive";

	if (ref $object->{Id} eq "ARRAY") {
	    $object->{URL} = $self->get_url($object->{id}->[0]);
	} else {
	    $object->{URL} = $self->get_url($object->{id});
	}
    }

    $object = $self->bless_hash($object);

    return $object;
}

sub upsert {
    my( $self, $args ) = @_;

    my $want_object = delete $args->{want_object};

#    $args->{record} = $self->_convert_to_record('Contacts', $args->{record});
    my $object = $args->{record};
    my $company = $object->{"Department"} || $object->{"Company"} || $object->{"AccountName"};
    my $company_id = $object->{"WhatId"};

    unless ( defined $company_id && length $company_id ) {
	my $response = eval { $self->get("organizations/find", { term => $company }) };
	if ( defined $response ) {
	    foreach my $organization ( @{ $response } ) {
		if (lc $organization->{name} eq lc $company) {
		    $company_id = $organization->{id};
		    last;
		}
	    }
	}
    }

    unless ( defined $company_id && length $company_id ) {
	my $response = eval { $self->post("organizations", [ name => $company, visible_to => 3 ]) };
	if ( defined $response && exists $response->{id} && length $response->{id} ) {
	    $company_id = $response->{id};
	}
    }

    if ( $args->{record}->{Id} ) {
	my $response = eval { $self->put("persons/".$args->{record}->{Id}, [ name => $args->{record}->{FirstName} . " " . $args->{record}->{LastName}, org_id => $company_id, email => $args->{record}->{Email}, phone => $args->{record}->{Phone} ]) };

	return undef unless defined $response;
	return $self->_convert_record('Contacts', $response) if $want_object;
	return $response->{id};
    } else {
	my $response = eval { $self->post("persons", [ name => $args->{record}->{FirstName} . " " . $args->{record}->{LastName}, org_id => $company_id, email => $args->{record}->{Email}, phone => $args->{record}->{Phone}, visible_to => 3 ]) };

	return undef unless defined $response;
	return $self->_convert_record('Contacts', $response) if $want_object;
	return $response->{id};
    }
}

sub find {
    my( $self, $args ) = @_;

    my @result;

    my $response = $self->get("searchResults", { term => $args->{query}->{what}, item_type => "person", start => 0, exact_match => 0 });

    if ( defined $response && ref $response eq "ARRAY" && scalar @{ $response } ) {
	foreach my $object ( @{ $response } ) {
	    if ( exists $object->{type} && $object->{type} eq "person" ) {
		my $person = eval { $self->get("persons/" . $object->{id}) };

		$person = $self->_convert_record('Contacts', $person) if defined $person;

		push @result, $person if defined $person;
	    }
	}
    }

    return \@result;
}

sub query {
    my( $self, $args ) = @_;

    my @result;
    my @returns;

    return $self->find({
	query => {
	    what => $args->{query}->{criteria},
	    where => "phone"
	}
    });
}

sub get_records {
    my( $self, $args ) = @_;

    my @result;

    my $persons = eval { $self->get("persons") };

    foreach my $person ( @{ $persons } ) {
	$person = $self->_convert_record('Contacts', $person) if defined $person;
	push @result, $person if defined $person;
    }

    return \@result;
}

sub find_phone {
    my( $self, $phone ) = @_;

    return $self->find({
	query => {
	    what => $phone,
	    where => "phone"
	}
    });
}

sub query_by_fields {
    my( $self, $query, $fields ) = @_;

    return $self->find({
	query => {
	    what => $query,
	    where => "phone"
	}
    });

    my $contact_fields = {
	"fullname" => { name => $query },
	"firstname"=> { name => $query },
	"lastname" => { name => $query },
	"comapny"  => { org => $query },
	"phone"    => { phone => $query },
	"email"    => { email => $query }
    };

    my $contacts_query_criteria = [ ];
    map { push @{ $contacts_query_criteria }, $contact_fields->{$_} if exists $contact_fields->{$_} } @{ $fields };

    my $contacts = $self->query({
	query => {
	    criteria => join ' OR ', @{ $contacts_query_criteria }
	},
	object => 'Contact'
    });


    my @result = ( @{$contacts} );

    return \@result;
}

sub query_all {
    my( $self, $query, $fields ) = @_;

    return $self->find({
	query => {
	    what => $query
	}
    });

    return $self->query_by_fields($query, $fields) if defined $fields && ref $fields eq 'ARRAY' && scalar @{ $fields };
}

sub query_tasks_by_rec_id {
    my( $self, $rec_id ) = @_;

    return $self->query({
	query => {
#	    fields => ["Id", "AccountId", "CallDisposition", "Description", "OwnerId", "Subject", "CallType", "WhatId", "WhoId"],
	    criteria => "CallDisposition = '" . $rec_id . "'"
	},
	object => "Task"
    });
}

sub query_tasks_by_id {
    my( $self, $id ) = @_;

    return $self->query({
	query => {
#	    fields => ["Id", "AccountId", "CallDisposition", "Description", "OwnerId", "Subject", "CallType", "WhatId", "WhoId"],
	    criteria => "Id = '" . $id . "'"
	},
	object => "Task"
    });
}

sub query_leads_by_id {
    my( $self, $id ) = @_;

    return $self->query({
	query => {
#	    fields => ["Id", "Company", "Description", "Email", "FirstName", "LastName", "Name", "Phone", "Title"],
	    criteria => "Id = '" . $id . "'"
	},
	object => "Lead"
    });
}

sub query_contacts_by_id {
    my( $self, $id ) = @_;

    return $self->query({
	query => {
#	    fields => ["Id", "Department", "Description", "Email", "FirstName", "LastName", "Name", "Phone", "Title"],
	    criteria => "Id = '" . $id . "'"
	},
	object => "Contact"
    });
}

sub query_accounts_by_id {
    my( $self, $id ) = @_;

    return $self->query({
	query => {
#	    fields => ["Id", "Department", "Description", "Email", "FirstName", "LastName", "Name", "Phone", "Title"],
	    criteria => "Id = '" . $id . "'"
	},
	object => "Account"
    });
}

sub bless_hash {
    my ($self, $object) = @_;
    my $result = { };

    eval {
	foreach my $key ( keys %{ $object } ) {
	    next if $key =~ m/\_{2,}/;
	    $result->{$key} = $object->{$key};
	    $result->{$key} = '' unless defined $object->{$key};
	}
    };

    return $result;
}

sub _convert_record_contact {
    my $self = shift;
    my $object = shift;

    my $result = {
	"FirstName" => $object->{first_name},
	"LastName" => $object->{last_name},
	"Department" => $object->{org_name},
	"Status" => "",
	"Description" => "",
	"AccountId" => $object->{company_id},
	"AccountName" => $object->{org_name},
	"type" => 'Contact',
	"contact_source_contact_id" => $object->{'id'},
	"contact_source_type" => 'Pipedrive',
	"contact_source" => $self->{id}
    };

    $result->{URL} = sprintf("https://%s.%s/person/%s", $self->{crm_subdomain}, $self->{crm_domain}, $object->{'id'});

    foreach my $phone ( @{ $object->{phone} } ) {
	$result->{ucfirst $phone->{label}} = $phone->{value};
    }

    foreach my $email ( @{ $object->{email} } ) {
	$result->{Email} = $email->{value};
    }

    $result->{Id} = [ $object->{id} ] if exists $object->{id};

    return $result;
}

sub _convert_record_account {
    my $self = shift;
    my $object = shift;

    my $result = {
	'Name' => $object->{'name'},
	'WhoId' => $object->{'owner_id'}->{'id'},
	'Id' => $object->{'company_id'},
	'AccountId' => $object->{'company_id'},
	'contact_source_type' => "Pipedrive",
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record {
    my $self = shift;
    my $object_type = shift;
    my $object = shift;

    my $result;

    if ( $object_type eq 'Contacts' ) {
	$result = $self->_convert_record_contact($object);
    }
    elsif ( $object_type eq 'Calls' ) {
	$result = $self->_convert_record_call($object);
    }
    else {
	$result = $object;
    }

    return $self->bless_hash($result);
}

sub _task_to_call_record {
    my( $self, $object ) = @_;

    $object->{CallDurationInSeconds} = 0 unless $object->{CallDurationInSeconds};

    my $result = {
	"Subject" => $object->{Subject},
	"Call Type" => $object->{Direction} || "Inbound",
	"Description" => $object->{Description},
	"Call Result" => md5_hex($object->{CallDisposition}),
	"Call Purpose" => "Contact",
	"Call Start Time" => $object->{Time} ? strftime("\%F \%H:\%M:\%S", localtime(int($object->{Time}))) : strftime("\%F \%H:\%M:\%S", localtime),
	"Call Duration (in seconds)" => $object->{CallDurationInSeconds},
	"Call Duration" => sprintf ("%02d:%02d",
	    int($object->{CallDurationInSeconds} / 60),
	    int($object->{CallDurationInSeconds} % 60)
	)
    };

    if ( $object->{WhoId} ) {

	if ( $self->detect_type_by_id($object->{WhoId}) eq "CONTACT" ) {
	    $result->{"SEMODULE"} = "Contacts";
	    $result->{"CONTACTID"} = $object->{WhoId};
	}
	else {
	    $result->{"RELATEDTOID"} = $object->{WhoId},
	    $result->{"SEID"} = $object->{WhoId},
	    $result->{"SEMODULE"} = "Leads",
	}
    }

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _contact_to_contacts_record {
    my( $self, $object ) = @_;

    my $result = {
	"First Name" => $object->{FirstName},
	"Last Name" => $object->{LastName},
	"Email" => $object->{Email},
	"Phone" => $object->{Phone},
	"Mobile" => $object->{Mobile},
	"Other Phone" => $object->{Other},
	"Department" => $object->{Department},
	"Description" => $object->{Description}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _accounts_to_account_record {
    my( $self, $object ) = @_;

    my $result = {
	"name" => $object->{Name},
	"owner_id" => $object->{owner_id},
	"visible_to" => '3'
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _convert_to_record {
    my( $self, $type, $object ) = @_;

    my $result;

    if ( $type eq 'Calls' ) {
	$result = $self->_task_to_call_record($object);
    }
    elsif ( $type eq 'Contacts' ) {
	$result = $self->_contact_to_contacts_record($object);
    }
    elsif ( $type eq 'Leads' ) {
	$result = $self->_leads_to_lead_record($object);
    }
    elsif ( $type eq 'Accounts' ) {
	$result = $self->_accounts_to_account_record($object);
    }
    else {
	$result = $object;
    }
}

1;
