package Integration::CRM::Sugar;

use strict;
use warnings;


use JSON::XS;
use Types::Serialiser;

use Integration::CRM::Sugar::Adapter;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);
use URI::Encode;

our $VERSION = '1';

our $DEBUG = 0;

my %object_prefixes = (
	ACCOUNT                     => '001',
	CONTACT                     => '003',
	LEAD                        => '00Q',
	TASK                        => '00T'
);

sub new {
    my( $class, $args ) = @_;
    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    $self->{id} = $args->{id} || 0;

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    die unless $args->{user}->{login} && $args->{user}->{password} && $args->{url};
    die "connection error" unless $self->_connect;

    return $self;
}

sub json { shift->{json} }

sub crm { shift->{crm} }

sub conf { shift->{config} }

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
    print Data::Dumper->Dump([@_]);
}

sub _w {
    my( $cond ) = @_;
    my $new_cond = '%' . $cond . '%';
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($new_cond);
}

sub _connect {
    my $self = shift;

    $self->{crm} = Integration::CRM::Sugar::Adapter->new({
	url => $self->conf->{url},
	restuser => $self->conf->{user}->{login},
	restpasswd => md5_hex($self->conf->{user}->{password})
    });

    eval {
	$self->{crm}->login;
    };

    return undef if $@;

    return $self;
}

sub _nvhash {
    my $self = shift;
    my $object = shift;

    return { } unless exists $object->{name_value_list};

    my %result = map { $_ => $object->{name_value_list}->{$_}->{value} } keys %{ $object->{name_value_list} };

    return \%result;
}

sub _convert_record_lead {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => $object->{'first_name'},
	'LastName' => $object->{'last_name'},
	'Description' => $object->{'description'},
	'Status' => $object->{'status'},
	'Email' => $object->{'email'},
	'Phone' => $object->{'phone_work'},
	'Company' => $object->{'department'},
	'Id' => [ $object->{'id'} ],
	'type' => 'Lead',
	'contact_source' => 'sugar',
	'contact_source_contact_id' => $object->{'id'},
	'contact_source_type' => 'Sugar',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record_contact {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => $object->{'first_name'},
	'LastName' => $object->{'last_name'},
	'Description' => $object->{'description'},
	'Email' => $object->{'email'},
	'Phone' => $object->{'phone_work'},
	'Department' => $object->{'department'},
	'Id' => [ $object->{'id'} ],
	'type' => 'Contact',
	'AccountId' => $object->{'account_id'},
	'AccountName' => $object->{'account_name'},
	'contact_source' => 'sugar',
	'contact_source_contact_id' => $object->{'id'},
	'contact_source_type' => 'Sugar',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record_call {
    my $self = shift;
    my $object = shift;

    my $result = {
	'Status' => $object->{'status'},
	'Id' => [ $object->{'id'} ],
	'Subject' => $object->{'name'},
	'WhatId' => $object->{'parent_id'},
	'CallDurationInSeconds' => ( $object->{'duration_hours'} * 60 * 60 ) + ( $object->{'duration_minutes'} * 60 ),
	'CallDisposition' => $object->{'outlook_id'},
	'type' => 'Task',
	'WhoId' => $object->{'parent_id'},
	'Description' => $object->{'description'},
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record_account {
    my $self = shift;
    my $object = shift;

    my $result = {
	'Name' => $object->{'name'},
	'WhoId' => $object->{'parent_id'},
	'Id' => $object->{'id'},
	'AccountId' => $object->{'id'},
	'Phone' => $object->{'phone_office'},
	'contact_source_type' => 'Sugar',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record {
    my $self = shift;
    my $object_type = shift;
    my $object = shift;

    my $result;

    if ( $object_type eq 'Leads' ) {
	$result = $self->_convert_record_lead($object);
    }
    elsif ( $object_type eq 'Contacts' ) {
	$result = $self->_convert_record_contact($object);
    }
    elsif ( $object_type eq 'Calls' ) {
	$result = $self->_convert_record_call($object);
    }
    elsif ( $object_type eq 'Accounts' ) {
	$result = $self->_convert_record_account($object);
    }
    else {
	$result = $object;
    }

    return $result;
}

sub get_records {
    my $self = shift;
    my $object = shift;

    return $self->find($object);
}

sub get_record_by_id {
    my $self = shift;
    my $object = shift;
    my $id = shift;

    return $self->find($object, { id => $id });
}

sub get_leads {
    my $self = shift;
    return $self->get_records('Leads');
}

sub get_contacts {
    my $self = shift;
    return $self->get_records('Contacts');
}

sub get_calls {
    my $self = shift;
    return $self->get_records('Calls');
}

sub detect_type_by_id {
    my ( $self, $id ) = @_;

    my $records = $self->get_record_by_id('Leads', $id);
    return 'LEAD' if scalar @{ $records };

    $records = $self->get_record_by_id('Contacts', $id);
    return 'CONTACT' if scalar @{ $records };

    $records = $self->get_record_by_id('Calls', $id);
    return 'TASK' if scalar @{ $records };

    $records = $self->get_record_by_id('Accounts', $id);
    return 'ACCOUNT' if scalar @{ $records };

    return 'UNKNOWN';
}

sub get_field_list {
    my ( $self, $object ) = @_;

    my $response = $self->crm->get_module_fields($object);

    return undef unless $response;

    my @fields = keys %{ $response->{module_fields} };

    return \@fields;
}

sub get_fields {
    my ( $self, $object ) = @_;

    return $self->get_field_list($object);
}

sub update {
    my( $self, $args ) = @_;

    my $id = delete $args->{record}->{Id};

    return $self->crm->update_module_entry($args->{object}, $id, $args->{record});
}

sub insert {
    my( $self, $args ) = @_;

    return $self->crm->create_module_entry($args->{object}, $args->{record});
}

sub _task_to_call_record {
    my( $self, $object ) = @_;

    $object->{CallDurationInSeconds} = 0 unless $object->{CallDurationInSeconds};

    my $result = {
	"name" => $object->{Subject},
	"direction" => "Inbound",
	"description" => $object->{Description},
	"outlook_id" => $object->{CallDisposition},
	"date_entered" => strftime("\%F \%H:\%M:\%S", localtime),
	"duration_hours" => int($object->{CallDurationInSeconds} / 3600),
	"duration_minutes" => int(($object->{CallDurationInSeconds} - ((int($object->{CallDurationInSeconds} / 3600)) * 3600)) / 60)
    };

    if ( $object->{WhoId} ) {
	if ( $self->detect_type_by_id($object->{WhoId}) eq "CONTACT" ) {
	    $result->{"parent_type"} = "Contacts";
	}
	else {
	    $result->{"parent_type"} = "Leads",
	}
	$result->{"parent_id"} = $object->{WhoId};
    }

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _contact_to_contacts_record {
    my( $self, $object ) = @_;

    my $result = {
	"first_name" => $object->{FirstName},
	"last_name" => $object->{LastName},
	"email" => $object->{Email},
	"phone_work" => $object->{Phone},
	"department" => $object->{Department}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _leads_to_lead_record {
    my( $self, $object ) = @_;

    my $result = {
	"first_name" => $object->{FirstName},
	"last_name" => $object->{LastName},
	"email" => $object->{Email},
	"phone_work" => $object->{Phone},
	"department" => $object->{Company},
	"status" => $object->{Status}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _accounts_to_account_record {
    my( $self, $object ) = @_;

    my $result = {
	"name" => $object->{Name},
	"phone_office" => $object->{Phone}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _convert_to_record {
    my( $self, $type, $object ) = @_;

    my $result;

    if ( $type eq 'Calls' ) {
	$result = $self->_task_to_call_record($object);
    }
    elsif ( $type eq 'Contacts' ) {
	$result = $self->_contact_to_contacts_record($object);
    }
    elsif ( $type eq 'Leads' ) {
	$result = $self->_leads_to_lead_record($object);
    }
    elsif ( $type eq 'Accounts' ) {
	$result = $self->_accounts_to_account_record($object);
    }
    else {
	$result = $object;
    }
}

sub upsert {
    my( $self, $args ) = @_;

    my $want_object = delete $args->{want_object};

    $args->{object} = 'Calls' if $args->{object} eq 'Task';
    $args->{object} = 'Contacts' if $args->{object} eq 'Contact';
    $args->{object} = 'Leads' if $args->{object} eq 'Lead';
    $args->{object} = 'Accounts' if $args->{object} eq 'Account';

    $args->{record} = $self->_convert_to_record($args->{object}, $args->{record});

    my $result = { };

    if ( exists $args->{record}->{Id} ) {
	$result = $self->update($args) 
    } else {
	$result = $self->insert($args);
    }
	
    if ($want_object && (defined $result) && (ref ($result) ne "HASH") && (ref ($result) ne "ARRAY")) {
	$result = $self->get_record_by_id($args->{object}, $result);
    }

    if ( ( ref ($result) eq "ARRAY" ) && ( scalar @{ $result } == 1 ) ) {
	$result = $result->[0];
    }

    return $result;
}

sub find {
    my( $self, $object, $args ) = @_;

    $object = 'Calls' if $object eq 'Task';
    $object = 'Contacts' if $object eq 'Contact';
    $object = 'Leads' if $object eq 'Lead';
    $object = 'Accounts' if $object eq 'Account';

    my $query;
    if ( defined $args ) {
	if ( ref $args eq "HASH" ) {
	    foreach my $arg ( keys %{ $args } ) {
		$query .= sprintf(" %s.%s = '%s'", lc $object, $arg, $args->{$arg});
	    }
	} else {
	    $query = $args;
	}
    }

    my $response = $self->crm->get_module_entries($object, $query);
    return [ ] unless $response;

    my $rows = $response;
    $rows = [ $rows ] if ref $rows eq "HASH";

    my $result = [ ];

    foreach my $row ( @{ $rows} ) {
#	$self->pdump($row);
	push @{ $result }, $self->_convert_record($object, $self->_nvhash($row));
    }

    return $result;
}

sub query {
    my( $self, $object, $args ) = @_;
    return $self->find($object, $args);
}

sub find_phone {
    my( $self, $phone ) = @_;

    $phone =~ s/[^0-9]//g;
    my $query = _w( join ( "\%", split (//, $phone ) ) );

#    my $leads = $self->find( 'Leads', { phone_work => $phone } );
    my $leads = $self->find('Leads', "leads.phone_work like \"$query\" or leads.phone_mobile like \"$query\" or leads.phone_other like \"$query\" or leads.phone_fax like \"$query\" or leads.phone_home like \"$query\"");

#    my $contacts = $self->find( 'Contacts', { phone_work => $phone } );
    my $contacts = $self->find('Contacts', "contacts.phone_work like \"$query\" or contacts.phone_other like \"$query\" or contacts.phone_mobile like \"$query\" contacts.phone_home like \"$query\" contacts.phone_fax like \"$query\"");

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub query_by_field {
    my( $self, $query, $fields ) = @_;

    $query = _w($query);

    my $lead_fields = {
        "fullname" => qq{leads.first_name like "$query" or leads.last_name like "$query"},
        "firstname"=> qq{leads.first_name like "$query"},
        "lastname" => qq{leads.last_name like "$query"},
        "comapny"  => qq{leads.account_name like "$query"},
        "phone"    => qq{leads.phone_work like "$query" or leads.phone_other like "$query" or leads.phone_mobile like "$query" leads.phone_home like "$query" leads.phone_fax like "$query"},
        "email"    => qq{leads.email like "$query"}
    };

    my $contact_fields = {
        "fullname" => qq{contacts.first_name like "$query" or contacts.last_name like "$query"},
        "firstname"=> qq{contacts.first_name like "$query"},
        "lastname" => qq{contacts.last_name like "$query"},
        "comapny"  => qq{contacts.department like "$query"},
        "phone"    => qq{contacts.phone_work like "$query" or contacts.phone_other like "$query" or contacts.phone_mobile like "$query" contacts.phone_home like "$query" contacts.phone_fax like "$query"},
        "email"    => qq{contacts.email like "$query"}
    };

    my $contacts_query_criteria = [ ];
    map { push @{ $contacts_query_criteria }, $contact_fields->{$_} if exists $contact_fields->{$_} } @{ $fields };

    my $leads_query_criteria = [ ];
    map { push @{ $leads_query_criteria }, $lead_fields->{$_} if exists $lead_fields->{$_} } @{ $fields };

    my $leads = $self->find('Leads', join " or ", @{ $leads_query_criteria });
    my $contacts = $self->find('Contacts', join " or ", @{ $contacts_query_criteria });

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub query_all {
    my( $self, $query, $fields ) = @_;

    return $self->query_by_field($query, $fields) if defined $fields && ref $fields eq "ARRAY" && scalar @{ $fields };

    $query = _w($query);

    my $leads = $self->find('Leads', "leads.phone_work like \"$query\" or leads.phone_mobile like \"$query\" or leads.phone_other like \"$query\" or leads.phone_fax like \"$query\" or leads.phone_home like \"$query\" or leads.first_name like \"$query\" or leads.last_name like \"$query\" or leads.email like \"$query\" or leads.account_name like \"$query\"");
    my $contacts = $self->find('Contacts', "contacts.phone_work like \"$query\" or contacts.phone_other like \"$query\" or contacts.phone_mobile like \"$query\" contacts.phone_home like \"$query\" contacts.phone_fax like \"$query\" or contacts.first_name like \"$query\" or contacts.last_name like \"$query\" or contacts.email like \"$query\" or contacts.department like \"$query\"");

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub query_tasks_by_rec_id {
    my( $self, $rec_id ) = @_;

    return $self->find('Calls', { outlook_id => $rec_id });
}

sub query_tasks_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Calls', $id);
}

sub query_leads_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Leads', $id);
}

sub query_contacts_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Contacts', $id);
}

sub query_accounts_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Accounts', $id);
}

1;
