package Integration::CRM::Sugar::Adapter;


use JSON::XS;
use Types::Serialiser;
use Data::Dumper;
use LWP::UserAgent;
use HTTP::Request::Common;

our $LEADS = 'Leads';
our $NOTES = 'Notes';
our $ACCOUNTS = 'Accounts';
our $CONTACTS = 'Contacts';
our $OPPORTUNITIES = 'Opportunities';
our $CAMPAIGNS = 'Campaigns';
our $PROSPECTLISTS = 'ProspectLists';
our $EMAILMARKETINGS = 'EmailMarketing';
our $USERS = 'Users';
our $CURRENCIES = 'Currencies';

our $DEBUG = 0;

sub new {
    my( $class, $args ) = @_;

    my $self = bless( { }, $class );

    $self->{config} = $args;
    $self->{application} = 'net_sugarcrm';
    $self->{max_results} = $args->{max_results} || 1000;

    $self->{globalua} = LWP::UserAgent->new(
        agent => "SugarCRMclient",
        keep_alive => 1
    );

    $self->{globalua}->env_proxy;
    $self->{globalua}->default_header('Accept' => 'application/json');

    return $self;
}

sub config { shift->{config} }

sub globalua { shift->{globalua} }

sub _sessionid { shift->{_sessionid} }

sub application { shift->{application} }

sub max_results { shift->{max_results} }

sub _module_id_for_search {
    my $self = shift;

    my $ret_val = {
	$LEADS => 'leads.id',
	$NOTES => 'notes.id',
	$ACCOUNTS => 'accounts.id',
	$CONTACTS => 'contacts.id',
	$OPPORTUNITIES => 'opportunities.id',
	$CAMPAIGNS => 'campaigns.id',
	$PROSPECTLISTS => 'prospect_lists.id',
	$EMAILMARKETINGS => 'email_marketing.id',
    };

    return $ret_val;
}

sub sessionid {
    my ($self) = @_;
    $self->login if (!defined($self->_sessionid));

    return $self->_sessionid;
}

sub _rest_request {
    my ($self, $method, $rest_data) = @_;

    my $res = $self->_rest_request_no_json($method, $rest_data);

    my $response = eval {
        JSON::XS->new->decode($res->content);
    };

    return {} if $@;

    return $response;
}

sub _rest_request_no_json {
    my ($self, $method, $rest_data) = @_;

    my $req = POST($self->{config}->{url}, Accept=>'application/json', Content => [
        method          => $method,
        input_type      => 'json',
        response_type   => 'json',
        rest_data       => $rest_data
    ]);

    my $res = $self->globalua->request($req);

    if ($res->is_error) {
    }

    return $res;
}

sub login {
    my ($self) = @_;

    my $application = $self->application;

    my $user_auth = {
            user_name => $self->config->{restuser},
            password => $self->config->{restpasswd}
    };

#    $user_auth->{encryption} = $self->encryption if ($self->encryption);

    my $rest_data = encode_json({
        user_auth => $user_auth,
	application => $application});

    my $response = $self->_rest_request('login', $rest_data);

    my $sessionid=$response->{id};
    $self->{_sessionid} = $sessionid;

    return $sessionid;
}

sub logout {
    my ($self) = @_;

    if (!defined($self->_sessionid)) {
        return;
    }
    my $rest_data = encode_json({
            session => $self->_sessionid,
        });

    my $res = eval {
        $self->_rest_request_no_json('logout', $rest_data);
    };

    if ($res->content ne 'null') {
    } else {
        $self->_sessionid = undef;
    }

    return;
}

sub get_available_modules {
    my ($self) = @_;
    my $rest_data = encode_json({
            session => $self->sessionid,
        });

    my $response = $self->_rest_request('get_available_modules', $rest_data);

    return $response;
}

sub get_module_fields {
    my ($self, $module_name) = @_;

    my $rest_data = '{ "session": "'.$self->sessionid.'", "module_name": "'.$module_name.'" }';
    my $response = $self->_rest_request('get_module_fields', $rest_data);

    return $response;
}

sub create_module_entry {
    my ($self, $module, $attributes) = @_;

    my $rest_data = '{"session": "'.$self->sessionid.'", "module_name": "'.$module.'", "name_value_list": '.
        encode_json($attributes).
        ',"track_view": "false" }';

    my $response = $self->_rest_request('set_entry', $rest_data);

    return $response->{id};
}

sub get_module_entries {
    my ($self, $module, $query) = @_;

    $query =~ s/"/\\"/xg;

    my $rest_data = '{"session": "'.$self->sessionid.'", "module_name": "'.$module.'", "query": "'.$query.'", "order_by": "", "offset": "", "select_fields": "", "link_name_to_fields_array": "", "max_results": "'.$self->max_results.'" } ';
    my $response = $self->_rest_request('get_entry_list', $rest_data);

    if ($response->{total_count} == 0) {
        return [];
    }

#    return [map { Integration::CRM::Sugar::Entry->new($_) } @{$response->{entry_list}}];
    return [map { $self->ret($_) } @{$response->{entry_list}}];
}

sub ret {
    my ( $self, $obj ) = @_;
    return $obj;
}

sub get_module_ids {
    my ($self, $module, $query) = @_;

    my $entries = $self->get_module_entries($module, $query);
    my @entriesids = map { $_->{id} } @$entries;

    return \@entriesids;
}

sub get_unique_module_id {
    my ($self, $module, $query) = @_;

    my $entries = $self->get_module_ids($module, $query);
    return () if ($#$entries == -1);

    return $entries->[0];
}

sub get_module_entry {
    my ($self, $module, $id) = @_;

    my $rest_data = '{"session": "'.$self->sessionid.'", "module_name": "'.$module.'", "query": "'.$self->_module_id_for_search->{$module}.' = \"'.$id.'\"" } ';
    my $response = $self->_rest_request('get_entry_list', $rest_data);

    if ($response->{total_count} == 0) {
        return ();
    }

#    return Integration::CRM::Sugar::Entry->new($response->{entry_list}->[0]);
    return $self->ret($response->{entry_list}->[0]);
}

sub get_module_attribute {
    my ($self, $module, $id, $attribute) = @_;
    my $entry = $self->get_module_entry($module, $id);

    if (defined $entry && exists($entry->{name_value_list}{$attribute})) {
        return $entry->{name_value_list}{$attribute}{value};
    }

    return ();
}

sub delete_module_entry_by_id {
    my ($self, $module, $id) = @_;
    my $rest_data = '{"session": "'.$self->sessionid.'", "module_name": "'.$module.'", "name_value_list": { "id" : "'.$id.'", "deleted" : "1" } } ';
    my $response = $self->_rest_request('set_entry', $rest_data);

    if ($response->{id} ne $id) {
        return ();
    }

    return 1;

}

sub get_module_link_ids {
    my ($self, $module, $link, $id, $query, $relatedfields) = @_;

    $query = '' if (!$query);
    $relatedfields = [ 'id' ] if (ref $relatedfields ne 'ARRAY');
    my $relatedfieldsjson;
    {
	local $Data::Dumper::Indent = 0;
	local $Data::Dumper::Useqq = 1;
	local $Data::Dumper::Terse = 1;
	$relatedfieldsjson = Dumper($relatedfields);
    }

    my $sessionid = $self->sessionid;
    my $rest_data = qq({"session":"$sessionid","module_name":"$module","module_id":"$id","link_field_name":"$link","related_module_query":"$query","related_fields":$relatedfieldsjson});

    my $response = $self->_rest_request('get_relationships', $rest_data);

    my @entriesids = map { $_->{id} } @{$response->{entry_list}};

    return \@entriesids;
}

sub update_module_entry {
    my ($self, $module, $id, $attributes) = @_;
    $$attributes{id} = $id;
    my $rest_data = '{"session": "'.$self->sessionid.'", "module_name": "'.$module.'", "name_value_list": '.
        encode_json($attributes).
        ' } ';
    my $response = $self->_rest_request('set_entry', $rest_data);

    if ($response->{id} ne $id) {
        return ();
    }

    return 1;
}

sub create_lead {
    my ($self, $attributes) = @_;

    return $self->create_module_entry($LEADS, $attributes);
}

sub get_leads {
    my ($self, $query) = @_;
    return $self->get_module_entries($LEADS, $query);
}

sub get_lead_ids {
    my ($self, $query) = @_;
    return $self->get_module_ids($LEADS, $query);
}

sub get_unique_lead_id {
    my ($self, $query) = @_;
    return $self->get_unique_module_id($LEADS, $query);
}

sub get_lead {
    my ($self, $leadid) = @_;
    return $self->get_module_entry($LEADS, $leadid);
}

sub get_lead_attribute {
    my ($self, $leadid, $attribute) = @_;
    return $self->get_module_attribute($LEADS, $leadid, $attribute);
}

sub delete_lead_by_id {
    my ($self, $leadid) = @_;
    return $self->delete_module_entry_by_id($LEADS, $leadid);
}

sub update_lead {
    my ($self, $leadid, $attributes) = @_;
    return $self->update_module_entry($LEADS, $leadid, $attributes);
}

sub create_contact {
    my ($self, $attributes) = @_;
    return $self->create_module_entry($CONTACTS, $attributes);
}

sub get_contacts {
    my ($self, $query) = @_;
    return $self->get_module_entries($CONTACTS, $query);
}

sub get_contact_ids {
    my ($self, $query) = @_;
    return $self->get_module_ids($CONTACTS, $query);
}

sub get_unique_contact_id {
    my ($self, $query) = @_;
    return $self->get_unique_module_id($CONTACTS, $query);
}

sub get_contact {
    my ($self, $contactid) = @_;
    return $self->get_module_entry($CONTACTS, $contactid);
}

sub get_contact_attribute {
    my ($self, $contactid, $attribute) = @_;
    return $self->get_module_attribute($CONTACTS, $contactid, $attribute);
}

sub get_contact_account_ids {
    my ($self, $id) = @_;
    return $self->get_module_link_ids($CONTACTS, 'accounts', $id);
}

sub delete_contact_by_id {
    my ($self, $contactid) = @_;
    return $self->delete_module_entry_by_id($CONTACTS, $contactid);
}

sub update_contact {
    my ($self, $contactid, $attributes) = @_;
    return $self->update_module_entry($CONTACTS, $contactid, $attributes);
}

sub create_account {
    my ($self, $attributes) = @_;
    return $self->create_module_entry($ACCOUNTS, $attributes);
}

sub get_accounts {
    my ($self, $query) = @_;
    return $self->get_module_entries($ACCOUNTS, $query);
}

sub get_account_ids {
    my ($self, $query) = @_;
    return $self->get_module_ids($ACCOUNTS, $query);
}

sub get_unique_account_id {
    my ($self, $query) = @_;
    return $self->get_unique_module_id($ACCOUNTS, $query);
}

sub get_account {
    my ($self, $accountid) = @_;
    return $self->get_module_entry($ACCOUNTS, $accountid);
}

sub get_account_attribute {
    my ($self, $accountid, $attribute) = @_;
    return $self->get_module_attribute($ACCOUNTS, $accountid, $attribute);
}

sub get_account_contact_ids {
    my ($self, $id) = @_;
    return $self->get_module_link_ids($ACCOUNTS, 'contacts', $id);
}

sub delete_account_by_id {
    my ($self, $accountid) = @_;
    return $self->delete_module_entry_by_id($ACCOUNTS, $accountid);
}

sub update_account {
    my ($self, $accountid, $attributes) = @_;
    return $self->update_module_entry($ACCOUNTS, $accountid, $attributes);
}

sub create_note {
    my ($self, $attributes) = @_;

    return $self->create_module_entry($NOTES, $attributes);
}

sub get_note {
    my ($self, $noteid) = @_;
    return $self->get_module_entry($NOTES, $noteid);
}

sub get_note_attribute {
    my ($self, $noteid, $attribute) = @_;
    return $self->get_module_attribute($NOTES, $noteid, $attribute);
}

sub delete_note_by_id {
    my ($self, $noteid) = @_;
    return $self->delete_module_entry_by_id($OPPORTUNITIES, $noteid);
}

sub DESTROY {
    my ($self) = @_;

    $self->logout;

    return;
}

sub update {
    my ($self, $entry) = @_;
    return $self->update_module_entry(
        $entry->{module_name},
        $entry->{id},
        $entry->{name_value_list});
}

1;
