package Integration::CRM::VTiger;

use strict;
use warnings;


use JSON::XS;
use Types::Serialiser;
use Integration::CRM::VTiger::Adapter;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);
use URI::Encode;

our $VERSION = '1';

our $DEBUG = 0;

my %object_prefixes = (
	ACCOUNT                     => '001',
	CONTACT                     => '003',
	LEAD                        => '00Q',
	TASK                        => '00T'
);

sub new {
    my( $class, $args ) = @_;
    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    $self->{id} = $args->{id} || 0;

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    die unless $args->{user}->{login} && $args->{user}->{password} && $args->{url};
    die "connection error" unless $self->_connect;

    return $self;
}

sub json { shift->{json} }

sub crm { shift->{crm} }

sub conf { shift->{config} }

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
    print Data::Dumper->Dump([@_]);
}

sub _w {
    my( $cond ) = @_;
    my $new_cond = '%' . $cond . '%';
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($new_cond);
}

sub _connect {
    my $self = shift;

    $self->{crm} = Integration::CRM::VTiger::Adapter->new( { url => $self->conf->{url} } );

    eval {
	$self->{crm_session} = $self->{crm}->getSession($self->conf->{user}->{login}, $self->conf->{user}->{password});
	$self->{session} = $self->{crm_session}->{'sessionName'};
    };

    return undef if $@;

    return $self;
}

sub _convert_record_lead {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => $object->{'firstname'},
	'LastName' => $object->{'lastname'},
	'Description' => $object->{'description'},
	'Status' => $object->{'leadstatus'},
	'Email' => $object->{'email'},
	'Phone' => $object->{'phone'},
	'Company' => $object->{'company'},
	'Id' => [ $object->{'id'} ],
	'type' => 'Lead',
	'contact_source_contact_id' => $object->{'id'},
	'contact_source_type' => 'VTiger',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record_contact {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => $object->{'firstname'},
	'LastName' => $object->{'lastname'},
	'Description' => $object->{'description'},
	'Email' => $object->{'email'},
	'Phone' => $object->{'phone'},
	'Department' => $object->{'department'},
	'Id' => [ $object->{'id'} ],
	'type' => 'Contact',
	'AccountId' => $object->{'account_id'},
	'AccountName' => $object->{'account_name'},
	'contact_source_contact_id' => $object->{'id'},
	'contact_source_type' => 'VTiger',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record_call {
    my $self = shift;
    my $object = shift;

    my $result = {
	'Status' => $object->{'eventstatus'},
	'Id' => [ $object->{'id'} ],
	'Subject' => $object->{'subject'},
	'WhatId' => $object->{'contact_id'} || $object->{'parent_id'},
	'CallDurationInSeconds' => ( ( ($object->{'duration_minutes'} || 0) * 60 ) + ( ($object->{'duration_hours'} || 0) * 60 * 60 ) ),
	'CallDisposition' => $object->{'location'},
	'type' => 'Task',
	'WhoId' => $object->{'contact_id'} || $object->{'parent_id'},
	'Description' => $object->{'description'},
	'contact_source_type' => 'VTiger',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record_account {
    my $self = shift;
    my $object = shift;

    my $result = {
	'Name' => $object->{'accountname'},
	'WhoId' => $object->{'assigned_user_id'},
	'Id' => $object->{'id'},
	'AccountId' => $object->{'id'},
	'Phone' => $object->{'phone'},
	'contact_source_type' => 'VTiger',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record {
    my $self = shift;
    my $object_type = shift;
    my $object = shift;

    my $result;

    if ( $object_type eq 'Leads' ) {
	$result = $self->_convert_record_lead($object);
    }
    elsif ( $object_type eq 'Contacts' ) {
	$result = $self->_convert_record_contact($object);
    }
    elsif ( $object_type eq 'Events' ) {
	$result = $self->_convert_record_call($object);
    }
    elsif ( $object_type eq 'Accounts' ) {
	$result = $self->_convert_record_account($object);
    }
    else {
	$result = $object;
    }

    return $result;
}

sub get_records {
    my $self = shift;
    my $object = shift;

    return $self->find($object);
}

sub get_record_by_id {
    my $self = shift;
    my $object = shift;
    my $id = shift;

    return $self->find($object, { id => $id });
}

sub get_leads {
    my $self = shift;
    return $self->get_records('Leads');
}

sub get_contacts {
    my $self = shift;
    return $self->get_records('Contacts');
}

sub get_calls {
    my $self = shift;
    return $self->get_records('Events');
}

sub detect_type_by_id {
    my ( $self, $id ) = @_;

    my $records = $self->get_record_by_id('Leads', $id);
    return 'LEAD' if scalar @{ $records };

    $records = $self->get_record_by_id('Contacts', $id);
    return 'CONTACT' if scalar @{ $records };

    $records = $self->get_record_by_id('Calls', $id);
    return 'TASK' if scalar @{ $records };

    $records = $self->get_record_by_id('Accounts', $id);
    return 'ACCOUNT' if scalar @{ $records };

    return 'UNKNOWN';
}

sub get_field_list {
    my ( $self, $object ) = @_;

    my $description = $self->crm->describe($self->{session}, $object);

    return undef unless $description;

    my @fields = map { $_->{name} } @{ $description->{'fields'} };

    return \@fields;
}

sub get_fields {
    my ( $self, $object ) = @_;

    return $self->get_field_list($object);
}

sub update {
    my( $self, $args ) = @_;

    my $id = delete $args->{record}->{Id};
    my $object = $args->{object};

    $args->{record}->{assigned_user_id} = $self->{crm_session}->{userId};

    my $retrieved = $self->crm->retrieve($self->{session}, $id);

    foreach my $key ( keys %{ $args->{record} } ) {
	$retrieved->{$key} = $args->{record}->{$key};
    }

    my $rows = $self->crm->update($self->{session}, $retrieved);
    $rows = [ $rows ] if ref $rows eq "HASH";

    my $result = [ ];

    foreach my $row ( @{ $rows} ) {
#	$self->pdump($row);
	push @{ $result }, $self->_convert_record($args->{object}, $row);
    }

    return $result;
}

sub insert {
    my( $self, $args ) = @_;

    $args->{record}->{assigned_user_id} = $self->{crm_session}->{userId};

    my $rows = $self->crm->create( $self->{session}, $args->{object}, $args->{record} );
    $rows = [ $rows ] if ref $rows eq "HASH";

    my $result = [ ];

    foreach my $row ( @{ $rows} ) {
#	$self->pdump($row);
	push @{ $result }, $self->_convert_record($args->{object}, $row);
    }

    return $result;
}

sub _task_to_call_record {
    my( $self, $object ) = @_;

    $object->{CallDurationInSeconds} = 0 unless $object->{CallDurationInSeconds};

    my $result = {
	"subject" => $object->{Subject},
	"activitytype" => "Call",
	"visibility" => "Public",
	"eventstatus" => "Held",
	"description" => $object->{Description},
	"location" => $object->{CallDisposition},
	"date_start" => strftime("\%F", localtime),
	"time_start" => strftime("\%H:\%M:\%S", localtime),
	"time_end" => strftime("\%H:\%M:\%S", localtime),
	"due_date" => strftime("\%F", localtime),
	"duration_minutes" => int($object->{CallDurationInSeconds} / 60),
	"duration_hours" => int($object->{CallDurationInSeconds} / 3600)
    };

    if ( $object->{WhoId} ) {
	if ( $self->detect_type_by_id($object->{WhoId}) eq "CONTACT" ) {
#	    $result->{"parent_type"} = "Contacts";
	    $result->{"contact_id"} = $object->{WhoId};
	}
	else {
#	    $result->{"parent_type"} = "Leads",
	    $result->{"parent_id"} = $object->{WhoId};
	}
#	$result->{"contact_id"} = $object->{WhoId};
    }

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _contact_to_contacts_record {
    my( $self, $object ) = @_;

    my $result = {
	"firstname" => $object->{FirstName},
	"lastname" => $object->{LastName},
	"email" => $object->{Email},
	"phone" => $object->{Phone},
	"department" => $object->{Department}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _leads_to_lead_record {
    my( $self, $object ) = @_;

    my $result = {
	"firstname" => $object->{FirstName},
	"lastname" => $object->{LastName},
	"email" => $object->{Email},
	"phone" => $object->{Phone},
	"company" => $object->{Company},
	"leadstatus" => $object->{Status}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _accounts_to_account_record {
    my( $self, $object ) = @_;

    my $result = {
	"accountname" => $object->{Name},
	"phone" => $object->{Phone}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _convert_to_record {
    my( $self, $type, $object ) = @_;

    my $result;

    if ( $type eq 'Events' ) {
	$result = $self->_task_to_call_record($object);
    }
    elsif ( $type eq 'Contacts' ) {
	$result = $self->_contact_to_contacts_record($object);
    }
    elsif ( $type eq 'Leads' ) {
	$result = $self->_leads_to_lead_record($object);
    }
    elsif ( $type eq 'Accounts' ) {
	$result = $self->_accounts_to_account_record($object);
    }
    else {
	$result = $object;
    }
}

sub upsert {
    my( $self, $args ) = @_;

    my $want_object = delete $args->{want_object};

    $args->{object} = 'Events' if $args->{object} eq 'Task';
    $args->{object} = 'Contacts' if $args->{object} eq 'Contact';
    $args->{object} = 'Leads' if $args->{object} eq 'Lead';
    $args->{object} = 'Accounts' if $args->{object} eq 'Account';

    $args->{record} = $self->_convert_to_record($args->{object}, $args->{record});

    my $result = { };

    if ( exists $args->{record}->{Id} ) {
        $result = $self->update($args) 
    } else {
        $result = $self->insert($args);
    }

    if ($want_object && (defined $result) && (ref ($result) ne "HASH") && (ref ($result) ne "ARRAY")) {
        $result = $self->get_record_by_id($args->{object}, $result);
    }

    if ( ( ref ($result) eq "ARRAY" ) && ( scalar @{ $result } == 1 ) ) {
        $result = $result->[0];
    }

    return $result;
}

sub find {
    my( $self, $object, $args ) = @_;

    $object = 'Events' if $object eq 'Task';
    $object = 'Contacts' if $object eq 'Contact';
    $object = 'Leads' if $object eq 'Lead';
    $object = 'Accounts' if $object eq 'Account';

    my $query = "";
    if ( defined $args ) {
	if ( ref $args eq "HASH" ) {
	    foreach my $arg ( keys %{ $args } ) {
		$query .= sprintf(" %s = '%s'", $arg, $args->{$arg});
	    }
	} else {
	    $query = $args;
	}
	$query = " where " . $query;
    }

    my $crm_query = sprintf("select * from %s %s;", $object, $query);

    my $response = $self->crm->query($self->{session}, $crm_query);

    return [ ] unless $response;

    my $rows = $response;
    $rows = [ $rows ] if ref $rows eq "HASH";

    my $result = [ ];

    foreach my $row ( @{ $rows} ) {
#	$self->pdump($row);
	push @{ $result }, $self->_convert_record($object, $row);
    }

    return $result;
}

sub query {
    my( $self, $object, $args ) = @_;

    return $self->find($object, $args);
}

sub find_phone {
    my( $self, $phone ) = @_;

    $phone =~ s/[^0-9]//g;
    $phone = "\%" . join ( "\%", split (//, $phone ) ) . "\%";

    my $leads = $self->find('Leads', "phone like \'$phone\'");
    my $contacts = $self->find('Contacts', "phone like \'$phone\'");

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub query_by_field {
    my( $self, $query, $fields ) = @_;

    $query = _w($query);

    my $lead_fields = {
	"fullname" => qq{firstname like '$query' or lastname like '$query'},
	"firstname"=> qq{firstname like '$query'},
	"lastname" => qq{lastname like '$query'},
	"comapny"  => qq{company like '$query'},
	"phone"    => qq{phone like '$query'},
	"email"    => qq{email like '$query'}
    };

    my $contact_fields = {
	"fullname" => qq{firstname like '$query' or lastname like '$query'},
	"firstname"=> qq{firstname like '$query'},
	"lastname" => qq{lastname like '$query'},
	"comapny"  => qq{department like '$query'},
	"phone"    => qq{phone like '$query' or fax like '$query' or mobile like '$query' or homephone like '$query' or otherphone like '$query'},
	"email"    => qq{email like '$query'}
    };

    my $contacts_query_criteria = [ ];
    map { push @{ $contacts_query_criteria }, $contact_fields->{$_} if exists $contact_fields->{$_} } @{ $fields };

    my $leads_query_criteria = [ ];
    map { push @{ $leads_query_criteria }, $lead_fields->{$_} if exists $lead_fields->{$_} } @{ $fields };

    my $leads = $self->find('Leads', join " or ", @{ $leads_query_criteria });
    my $contacts = $self->find('Contacts', join " or ", @{ $contacts_query_criteria });

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;

}

sub query_all {
    my( $self, $query, $fields ) = @_;

    return $self->query_by_field($query, $fields) if defined $fields && ref $fields eq "ARRAY" && scalar @{ $fields };

    $query = _w($query);

    my $leads = $self->find('Leads', "phone like \'$query\' or firstname like \'$query\' or lastname like \'$query\' or email like \'$query\' or company like \'$query\'");
    my $contacts = $self->find('Contacts', "phone like \'$query\' or fax like \'$query\' or mobile like \'$query\' or homephone like \'$query'\ or otherphone like \'$query'\ or firstname like \'$query\' or lastname like \'$query\' or email like \'$query\' or department like \'$query\'");

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub query_tasks_by_rec_id {
    my( $self, $rec_id ) = @_;

    return $self->find('Events', { location => $rec_id });
}

sub query_tasks_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Events', $id);
}

sub query_leads_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Leads', $id);
}

sub query_contacts_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Contacts', $id);
}

sub query_accounts_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Accounts', $id);
}

1;
