package Integration::CRM::VTiger::Adapter;

use strict;
use warnings;

use Digest::MD5;
use LWP::UserAgent;
use JSON::XS;
use Types::Serialiser;

our $DEBUG = 0;

sub new {
    my ( $class, $args ) = @_;

    my $self  = {
        ua   => LWP::UserAgent->new,
        json => JSON::XS->new->allow_nonref,
        ctx  => Digest::MD5->new,
        url  => $args->{url}
    };

    bless $self, $class;

    $self->{'ua'}->agent($args->{agent_name} || "Thirdlane CRM Client");

    return $self;
}

sub getSession {
    my $self     = shift;
    my $username = shift;
    my $pin      = shift;
    my $params   = '?operation=getchallenge&username=' . $username;
    my $session;

    my $req = HTTP::Request->new( GET => $self->{'url'} . $params );
    my $res = $self->{'ua'}->request($req);

    my $challenge = {};
    if ( $res->is_success ) {
        $challenge = $self->{'json'}->decode( $res->content );
        $self->{'ctx'}->add( $challenge->{result}{token} . $pin );
        my $ak = $self->{'ctx'}->hexdigest;
        $session = $self->_login( $username, $ak );
    }
    else {
        die "Error in server response: " . $res->status_line, "\n";
    }
    return $session;
}

sub _login {
    my $self = shift;
    my $un   = shift;
    my $ak   = shift;

    my $req = HTTP::Request->new( POST => $self->{'url'} );
    $req->content_type('application/x-www-form-urlencoded');
    $req->content( 'operation=login&username=' . $un . '&accessKey=' . $ak );

    my $res          = $self->{'ua'}->request($req);
    my $jsonResponse = {};
    if ( $res->is_success ) {
        $jsonResponse = $self->{'json'}->decode( $res->content );
        die( 'Login fault: ' . $jsonResponse->{error}{message} ) unless $jsonResponse->{'success'} eq 'true';
        my $result = $jsonResponse->{result};

        return $result;
    }
    else {
        die("Connection error\n");
    }
}

sub describe {
    my $self      = shift;
    my $sessionId = shift;
    my $module    = shift;

    my $params = '?sessionName=' . $sessionId . '&elementType=' . $module . '&operation=describe';

    my $result = $self->_get($params);

    return $result;
}

sub create {
    my $self       = shift;
    my $sessionId  = shift;
    my $moduleName = shift;
    my $data       = shift;

    my $objectJson = $self->{'json'}->encode($data);

    my $params     = {
        (
            sessionName => $sessionId,
            operation   => 'create',
            element     => $objectJson,
            elementType => $moduleName
        )
    };

    my $result = $self->_post($params);

    return $result;
}

sub delete {
    my $self      = shift;
    my $sessionId = shift;
    my $id        = shift;

    my $params = {
        (
            sessionName => $sessionId,
            id          => $id,
            operation   => 'delete'
        )
    };

    $self->_post($params);

    return 'deleted';
}

sub update {
    my $self      = shift;
    my $sessionId = shift;
    my $data      = shift;

    my $objectJson = $self->{'json'}->encode($data);

    my $params = {
        (
            sessionName => $sessionId,
            operation   => 'update',
            element     => $objectJson,
        )
    };

    my $result = $self->_post($params);

    return $result;
}

sub query {
    my $self      = shift;
    my $sessionId = shift;
    my $query     = shift;

    my $params = '?sessionName=' . $sessionId . '&operation=query&query=' . $query;

    my $result = $self->_get($params);

    return $result;
}

sub retrieve {
    my $self      = shift;
    my $sessionId = shift;
    my $id        = shift;

    my $params = '?sessionName=' . $sessionId . '&operation=retrieve&id=' . $id;

    my $result = $self->_get($params);

    return $result;
}

sub get_modules {
    my $self      = shift;
    my $sessionId = shift;

    my $params    = '?sessionName=' . $sessionId . '&operation=listtypes';

    my $result    = $self->_get($params);

    return $result;
}

sub _get {
    my $self   = shift;
    my $params = shift;

    my $req = HTTP::Request->new( GET => $self->{'url'} . $params );

    printf("URL: %s\n", $self->{'url'} . $params) if $DEBUG;

    my $res = $self->{'ua'}->request($req);

    my $jsonResponse = { };
    if ( $res->is_success ) {
        $jsonResponse = $self->{'json'}->decode( $res->content );
        die( 'Service fault! ' . $jsonResponse->{'error'}{'message'} ) unless $jsonResponse->{'success'} eq 'true';

	print $self->{'json'}->encode($jsonResponse) if $DEBUG;

        return $jsonResponse->{result};
    }
    else {
        die("Connection error\n");
    }
}

sub _post {
    my $self   = shift;
    my $params = shift;

    my $res = $self->{'ua'}->post( $self->{'url'}, $params );
    my $jsonResponse = {};
    if ( $res->is_success ) {
        $jsonResponse = $self->{'json'}->decode( $res->content );
        die( 'POST fault: ' . $jsonResponse->{error}{message} . "\n" . $self->{'url'} ) unless $jsonResponse->{'success'} eq 'true';

        return $jsonResponse->{result};
    }
    else {
        die("Connection error (POST)\n$@\n");
    }
}

1;
