package Integration::CRM::X2CRM;

use strict;
use warnings;


use JSON::XS;
use Types::Serialiser;
use LWP::UserAgent;
use URI::Encode;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);

our $VERSION = '2';

our $DEBUG = 0;

my %object_prefixes = (
	ACCOUNT                     => '001',
	CONTACT                     => '003',
	LEAD                        => '00Q',
	TASK                        => '00T'
);

sub new {
    my( $class, $args ) = @_;
    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    my $netloc = $args->{url};

    if ( $netloc =~ m{^http\://(.*?)$} ) {
	$netloc = $1 . ":80";
    }
    elsif ( $netloc =~ m{^https\://(.*?)$} ) {
	$netloc = $1 . ":443";
    }

    $self->{config}->{url} .= "/index.php/api2";

    $self->{config}->{netloc} = $netloc;
    $self->{config}->{realm} = "X2Engine API v2";

    $self->{id} = $args->{id} || 0;

    my $ua = LWP::UserAgent->new;
    $ua->timeout(10);
    $ua->env_proxy;

    $self->{ua} = LWP::UserAgent->new(ssl_opts => { verify_hostname => 0 });
    $self->{ua}->timeout(10);
    $self->{ua}->env_proxy;

    $self->{ua}->credentials( $self->conf->{netloc}, $self->conf->{realm}, $self->conf->{user}->{login}, $self->conf->{user}->{token} );

#    $self->{uri_encode} = URI::Encode->new( { encode_reserved => 1 } );

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    die "no token" unless $args->{user}->{token};

    die "connection error" unless $self->_connect;

    return $self;
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub ua {
    my $self = shift;

    return $self->{ua};
}

sub conf {
    my $self = shift;

    return $self->{config};
}

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
    print Data::Dumper->Dump([@_]);
}

sub _connect {
    my $self = shift;

    my $url = sprintf ('%s/appInfo.json',
	$self->conf->{url}
    );

    $self->{ua}->credentials( $self->conf->{netloc}, $self->conf->{realm}, $self->conf->{user}->{login}, $self->conf->{user}->{token} );
    my $response = $self->ua->get($url);

    if ($response->is_success) {
	print $response->decoded_content if $DEBUG;
    }
    else {
	die $response->status_line;
    }

    my $decoded_response = eval { $self->json->decode($response->decoded_content) };

    return undef if $@;
    return undef unless exists $decoded_response->{message};

    return $self;
}

sub _format_url_args {
    my ( $self, $args) = @_;

    my $result;

    foreach my $arg ( keys %{ $args }) {
	$arg =~ s{\"}{}sg;
	push @{ $result }, sprintf("%s=%s", $arg, $args->{$arg});
    }

    return join "&", @{ $result };
}

sub _get_json {
    my ( $self, $args) = @_;

    my $object = ucfirst $args->{object};

    my $id = "";
    $id = $args->{id} if exists $args->{id} && defined $args->{id} && length $args->{id};

    if ( defined $id && length $id && $id =~ m{\_\d+} ) {
	($id) = $id =~ m{^.*?\_(\d+)$};
    }

    $id .= ".json" if defined $id && length $id;

    my $opt_args = "";

    if ( exists $args->{args} && defined $args->{args} && length $args->{args} ) {
	$opt_args = "?" . $args->{args};
    }

    my $url = sprintf ('%s/%s/%s%s',
	$self->conf->{url},
	$object,
	$id,
	$opt_args
    );

    printf ("\nGET URL:%s\n", $url) if $DEBUG;

    $self->{ua}->credentials( $self->conf->{netloc}, $self->conf->{realm}, $self->conf->{user}->{login}, $self->conf->{user}->{token} );
    my $response = $self->ua->get($url);

    if ($response->is_success) {
	print $response->decoded_content if $DEBUG;
    }
    else {
	die sprintf ("url:%s, status:%s, response:%s", $url, $response->status_line, $response->decoded_content );
    }

    my $decoded_response = $self->json->decode($response->decoded_content);

    return undef if ref $decoded_response eq "HASH" && exists $decoded_response->{error};

    return $decoded_response;
}

sub _post_json {
    my ( $self, $args) = @_;

    my $object = ucfirst $args->{object};

    my $id = "";
    $id = $args->{id} if exists $args->{id} && defined $args->{id} && length $args->{id};

    if ( defined $id && length $id && $id =~ m{\_\d+} ) {
	($id) = $id =~ m{^.*?\_(\d+)$};
    }

    $id .= ".json" if defined $id && length $id;

    my $opt_args = "";

    if ( exists $args->{args} && defined $args->{args} && length $args->{args} ) {
	$opt_args = "?" . $args->{args};
    }

    my $url = sprintf ('%s/%s/%s%s',
	$self->conf->{url},
	$object,
	$id,
	$opt_args
    );

    my $body = eval { $self->json->encode($args->{record}) };

    printf ("\nPOST URL:%s\n", $url) if $DEBUG;
    printf ("\nPOST Data:%s\n", $body) if $DEBUG;

    $self->{ua}->credentials( $self->conf->{netloc}, $self->conf->{realm}, $self->conf->{user}->{login}, $self->conf->{user}->{token} );

    my $req = HTTP::Request->new();
    $req->uri($url);
    $req->content_type("application/json");
    $req->content($body);

    if ( $id ) {
	$req->method("PUT");
    } else {
	$req->method("POST");
    }

    my $response = $self->ua->request($req);

    $self->pdump($response) if $DEBUG;

    if ($response->is_success) {
	print $response->decoded_content if $DEBUG;
    }
    else {
	die sprintf ("url:%s, status:%s, body:%s, response:%s", $url, $response->status_line, $body, $response->decoded_content );
    }

    my $decoded_response = $self->json->decode($response->decoded_content);

    $self->pdump($decoded_response) if $DEBUG;

    return undef if ref $decoded_response eq "HASH" && exists $decoded_response->{error};

    return $decoded_response->{id} if ref $decoded_response eq "HASH" && exists $decoded_response->{id};

    return undef;
}

sub _convert_record_contact {
    my $self = shift;
    my $object = shift;

    my $company = $object->{'company'};
    my ($company_name, $account_id) = $company =~ m{^(.*?)\_(\d+)$};

    my $result = {
	'FirstName' => $object->{'firstName'},
	'LastName' => $object->{'lastName'},
	'Description' => $object->{'backgroundInfo'},
	'Email' => $object->{'email'},
	'Phone' => $object->{'phone'},
	'Mobile' => $object->{'phone2'},
	'Department' => $company_name,
	'Id' => [ $object->{'id'} ],
	'type' => 'Contact',
	'AccountId' => $account_id,
	'AccountName' => $object->{'company'},
	'contact_source_contact_id' => $object->{'id'},
	'contact_source_type' => 'X2CRM',
	'contact_source' => $self->{id}
    };

    unless ( $company_name && $account_id ) {
	$result->{'Department'} = $company;
	delete $result->{'AccountId'};
	delete $result->{'AccountName'};
    }

    return $result;
}

sub _convert_record_call {
    my $self = shift;
    my $object = shift;

    my $result = {
	'Status' => $object->{'eventStatus'},
	'Id' => [ $object->{'id'} ],
	'Subject' => $object->{'subject'},
	'WhatId' => $object->{'associationId'},
	'CallDurationInSeconds' => $object->{'timeSpent'},
	'CallDisposition' => $object->{'subject'},
	'type' => 'Task',
	'WhoId' => $object->{'associationId'},
	'Description' => $object->{'actionDescription'},
	'Direction' => 'Outbound',
	'Time' => $object->{'createDate'},
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record_account {
    my $self = shift;
    my $object = shift;

    my $result = {
	'Name' => $object->{'name'},
	'WhoId' => $object->{'primaryContact'},
	'Id' => $object->{'id'},
	'AccountId' => $object->{'id'},
	'Phone' => $object->{'phone'},
	'contact_source_type' => 'X2CRM',
	'contact_source' => $self->{id}
    };

    return $result;
}

sub _convert_record {
    my $self = shift;
    my $object_type = shift;
    my $object = shift;

    my $result;

    if ( $object_type eq 'Contacts' ) {
	$result = $self->_convert_record_contact($object);
    }
    elsif ( $object_type eq 'Leads' ) {
	$result = $self->_convert_record_call('Contacts');
    }
    elsif ( $object_type eq 'Actions' ) {
	$result = $self->_convert_record_call($object);
    }
    elsif ( $object_type eq 'Accounts' ) {
	$result = $self->_convert_record_account($object);
    }
    else {
	$result = $object;
    }

    return $result;
}

sub get_records {
    my $self = shift;
    my $object = shift;

    my $response = $self->_get_json( { object => $object } );

    return undef unless $response;
    return undef if ref $response eq "HASH" && exists $response->{error};
    return [ ] if ref $response eq "ARRAY" && scalar @{ $response } == 0;

    my $rows = $response;
    $rows = [ $rows ] if ref $rows eq "HASH";

    my $result;

    foreach my $item ( @{ $rows } ) {
	$item = $self->_convert_record($object, $item);
	push @{ $result }, $item;
    }

    return $result;
}

sub get_record_by_id {
    my $self = shift;
    my $object = shift;
    my $id = shift;

    my $response = $self->_get_json({ object => $object, id => $id });

    return undef unless $response;
    return undef if ref $response eq "HASH" && exists $response->{error};
    return [ ] if ref $response eq "ARRAY" && scalar @{ $response } == 0;

    my $rows = $response;
    $rows = [ $rows ] if ref $rows eq "HASH";

    my $result;

    foreach my $item ( @{ $rows} ) {
	$item = $self->_convert_record($object, $item);
	push @{ $result }, $item;
    }

    return $result;
}

sub get_leads {
    my $self = shift;
    return $self->get_records('Contacts');
}

sub get_contacts {
    my $self = shift;
    return $self->get_records('Contacts');
}

sub get_calls {
    my $self = shift;
    return $self->get_records('Actions');
}

sub detect_type_by_id {
    my ( $self, $id ) = @_;

    my $records = $self->get_record_by_id('Contacts', $id);
    return 'CONTACT' if scalar @{ $records };

    $records = $self->get_record_by_id('Actions', $id);
    return 'TASK' if scalar @{ $records };

    $records = $self->get_record_by_id('Accounts', $id);
    return 'ACCOUNT' if scalar @{ $records };

    return 'UNKNOWN';
}

sub update {
    my( $self, $args ) = @_;

    my $id = delete $args->{record}->{id};

    return $self->_post_json({ object => $args->{object}, id => $id, record => $args->{record} });
}

sub insert {
    my( $self, $args ) = @_;

    return $self->_post_json({ object => $args->{object}, record => $args->{record} });
}

sub _task_to_call_record {
    my( $self, $object ) = @_;

    $object->{CallDurationInSeconds} = 0 unless $object->{CallDurationInSeconds};

    my $result = {
	"subject" => md5_hex($object->{CallDisposition}),
	"actionDescription" => $object->{Description},
	"associationType" => "contacts",
	"complete" => "Yes",
	"createDate" => ($object->{Time} ? int($object->{Time}) : time),
	"dueDate" => ($object->{Time} ? int($object->{Time}) : time) - $object->{CallDurationInSeconds},
	"lastUpdated" => ($object->{Time} ? int($object->{Time}) : time),
	"completeDate" => ($object->{Time} ? int($object->{Time}) : time),
	"timeSpent" => $object->{CallDurationInSeconds},
	"type" => "call",
	"visibility" => "1"
    };

    if ( $object->{WhoId} ) {
	$result->{"associationId"} = $object->{WhoId};
    }

    $result->{id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _contact_to_contacts_record {
    my( $self, $object ) = @_;

    my $result = {
	"firstName" => $object->{FirstName},
	"lastName" => $object->{LastName},
	"email" => $object->{Email},
	"phone" => $object->{Phone},
	"phone2" => $object->{Mobile},
	"company" => ( defined $object->{Department} ? $object->{Department} : $object->{Company} ),
	"backgroundInfo" => $object->{Description},
	"visibility" => "1"
    };

    $result->{id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _accounts_to_account_record {
    my( $self, $object ) = @_;

    my $result = {
	"name" => $object->{Name},
	"phone" => $object->{Phone},
	"visibility" => "1"
    };

    $result->{id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _convert_to_record {
    my( $self, $type, $object ) = @_;

    my $result;

    if ( $type eq 'Actions' ) {
	$result = $self->_task_to_call_record($object);
    }
    elsif ( $type eq 'Contacts' ) {
	$result = $self->_contact_to_contacts_record($object);
    }
    elsif ( $type eq 'Accounts' ) {
	$result = $self->_accounts_to_account_record($object);
    }
    else {
	$result = $object;
    }
}

sub upsert {
    my( $self, $args ) = @_;

    my $want_object = delete $args->{want_object};

    $args->{object} = 'Actions' if $args->{object} eq 'Task';
    $args->{object} = 'Contacts' if $args->{object} eq 'Contact';
    $args->{object} = 'Contacts' if $args->{object} eq 'Lead';
    $args->{object} = 'Accounts' if $args->{object} eq 'Account';

    $args->{record} = $self->_convert_to_record($args->{object}, $args->{record});

    my $result = { };

    if ( exists $args->{record}->{id} ) {
	$result = eval { $self->update($args) };
    } else {
	$result = eval { $self->insert($args) };
    }

    if ($want_object && (defined $result) && (ref ($result) ne "HASH") && (ref ($result) ne "ARRAY")) {
	$result = $self->get_record_by_id($args->{object}, $result);
    }

    if ( ( ref ($result) eq "ARRAY" ) && ( scalar @{ $result } == 1 ) ) {
	$result = $result->[0];
    }

    return $result;
}

sub find {
    my( $self, $object, $args ) = @_;

    $self->pdump($args) if $DEBUG;

    my $response = $self->_get_json({ object => $object, args => $args });

    $self->pdump($response) if $DEBUG;

    return undef unless $response;
    return undef if ref $response eq "HASH" && exists $response->{error};
    return [ ] if ref $response eq "ARRAY" && scalar @{ $response } == 0;

    my $rows = $response;
    $rows = [ $rows ] if ref $rows eq "HASH";

    my $result;

    foreach my $item ( @{ $rows} ) {
	$item = $self->_convert_record($object, $item);
	push @{ $result }, $item;
    }

    return $result;
}

sub query {
    my( $self, $object, $args ) = @_;

    return $self->find($object, $args);
}

sub _w {
    my( $cond ) = @_;
    my $new_cond = '%' . $cond . '%';
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($new_cond);
}

sub find_phone {
    my( $self, $phone ) = @_;

    my $contacts = $self->find('Contacts', sprintf('_partial=1&_escape=0&_or=1&phone=%s&phone2=%s', _w($phone), _w($phone)));

    my @result = @{$contacts};

    return \@result;
}

sub query_all_by_field {
    my( $self, $query, $fields ) = @_;

    my $equery = _w($query);

    my $contact_fields = {
        "fullname" => qq{firstName=$equery&lastName=$equery},
        "firstname"=> qq{firstName=$equery},
	"lastname" => qq{lastName=$equery},
	"comapny"  => qq{company=$equery},
	"phone"    => qq{phone=$equery&phone2=$equery},
	"email"    => qq{email=$equery},
    };

    my $contacts_query_criteria = [ ];
    map { push @{ $contacts_query_criteria }, $contact_fields->{$_} if exists $contact_fields->{$_} } @{ $fields };

    my $contacts_query_criteria_url_params = join "&", @{ $contacts_query_criteria };

    my $contacts_query_criteria_url = "_partial=1&_escape=0";
    $contacts_query_criteria_url .= "&_or=1" if $contacts_query_criteria_url_params =~ m/\&/;
    $contacts_query_criteria_url .= "&" . $contacts_query_criteria_url_params;

    my $contacts = $self->find('Contacts', $contacts_query_criteria_url);

    return $contacts;
}

sub query_all {
    my( $self, $query, $fields ) = @_;

    return $self->query_all_by_field($query, $fields) if defined $fields && ref $fields eq "ARRAY" && scalar @{ $fields };

    my $equery = _w($query);

    my $contacts = $self->find('Contacts', sprintf('_partial=1&_escape=0&_or=1&phone=%s&phone2=%s&company=%s&email=%s&name=%s&firstName=%s&lastName=%s',
	$equery, $equery, $equery, $equery, $equery, $equery, $equery)
    );

    my @result = @{$contacts};

    return \@result;
}

sub query_tasks_by_rec_id {
    my( $self, $rec_id ) = @_;

    return $self->find('Actions', sprintf('_partial=1&_escape=0&_or=1&subject=%s', md5_hex($rec_id)));
}

sub query_tasks_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Actions', $id);
}

sub query_leads_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Contacts', $id);
}

sub query_contacts_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Contacts', $id);
}

sub query_accounts_by_id {
    my( $self, $id ) = @_;

    return $self->get_record_by_id('Accounts', $id);
}

1;
