package Integration::Contact::Source::Google::Contacts;

use utf8;
use strict;
use warnings;


use URI;
use JSON::XS;
use Types::Serialiser;
use LWP::UserAgent;
use HTTP::Request;
use URI::Encode;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);

use lib '/opt/thirdlane/perl_lib/lib';

use Thirdlane::Auth::OAuth2;

our $VERSION = '1';

our $DEBUG = 0;

sub new {
    my( $class, $args ) = @_;
    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    unless ( exists $args->{credentials}->{user_id} && defined $args->{credentials}->{user_id} && length $args->{credentials}->{user_id} ) {
	die "no token";
    }

    $self->{id} = $self->conf->{id} || 0;

    $self->{uri_encode} = URI::Encode->new( { encode_reserved => 1 } );

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    $self->{client} = Thirdlane::Auth::OAuth2->new( { provider => 'google', debug => $args->{debug} } );

    die "connection error" unless $self->_connect;

    return $self;
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub _w {
    my $cond = shift;
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($cond);
}

sub ua {
    my $self = shift;

    return $self->{ua}->();
}

sub get {
    my $self = shift;
    my $url = shift;

    my $req = HTTP::Request->new( GET => $url );

    return {} unless $self->ua;

    my $response = $self->ua->request( $req );

    if ( defined $response ) {
        if ( my $status = $response->header('Tlintstatus') ) {
            return { error => { provider => $response->header('Tlprovider'), status => $status } };
        }

        if ( $response->is_success() ) {
            my $response_content = $response->decoded_content();
            return  eval { $self->json->decode( $response_content ) };
        }
    }

    return {};
}

sub conf {
    my $self = shift;

    return $self->{config};
}

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
#    print Data::Dumper->Dump([@_]);
}

sub _connect {
    my $self = shift;

    $self->{ua} = sub { $self->{client}->build_user_agent( { user_id => $self->{config}->{credentials}->{user_id}, provider => 'google' } ); };

    return $self->{ua} = undef unless $self->{ua}->();

    return $self->{ua};
}

sub pmatch {
    my $self = shift;
    my ( $obj, $phone ) = @_;

	foreach my $record ( keys %{ $obj } ) {
	    if ( index( lc $obj->{$record}, lc $phone ) != -1 ) {
		return 1;
	    }
	}
    return 0;
}


sub get_by_phone {
    my $self = shift;
#    my $type = shift;
    my $phone = shift;

    return [ ] unless defined $phone && length $phone;

#    my $cache_results = $self->{client}->cache->get('oauth::provider::' . $self->{client}->provider->name . '::contacts::phone::' . $phone);

#    unless ( defined $cache_results ) {
    my @result;
    my $uri = URI->new( q{https://people.googleapis.com/v1/people/me/connections} );

    my $query = {
	'requestMask.includeField' => 
	    'person.names,person.email_addresses,person.phone_numbers,person.organizations,person.photos,person.urls,person.taglines,person.biographies,person.metadata',
	'sortOrder' => 
	    'FIRST_NAME_ASCENDING',
	'pageSize' => 
	    '500'
    };
#    nextPageToken


    my $decoded_response = { };

    while (1) {

    $query->{nextPageToken} = $decoded_response->{nextPageToken} if defined $decoded_response && exists $decoded_response->{nextPageToken} && defined $decoded_response->{nextPageToken} && length $decoded_response->{nextPageToken};

    $uri->query_form($query);

    $decoded_response = $self->get($uri->as_string);

#    last if defined $decoded_response && exists $decoded_response->{error}; #RESOURCE_EXHAUSTED

    if ( ref $decoded_response->{connections} eq "ARRAY" ) {
	foreach my $record ( @{ $decoded_response->{connections} } ) {
	    my $conversion = $self->_convert_record_contact($record);
	    if ( $self->pmatch( $conversion, $phone ) ) {
		push @result, $conversion;
	    }
	}
    }

    last unless defined $decoded_response && exists $decoded_response->{nextPageToken} && defined $decoded_response->{nextPageToken} && length $decoded_response->{nextPageToken};

    }
#    $self->{client}->cache->set('oauth::provider::' . $self->{client}->provider->name . '::contacts::phone::' . $phone, \@result, 60*5)
#	if scalar @result;

    return \@result;
#    }

#    return defined $cache_results ? $cache_results : [ ];
}

sub _convert_record_contact {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => '',
	'LastName' => '',
	'Description' => '',
	'Email' => '',
	'Phone' => '',
	'Company' => '',
	'Id' => $object->{resourceName},
	'type' => 'Contact',
	'contact_source' => 'google',
	'contact_source_contact_id' => $object->{resourceName},
	'contact_source_type' => 'Google',
	'contact_source' => $self->{id}
    };

    if ( exists $object->{names} && defined $object->{names} && ref $object->{names} eq "ARRAY" ) {
	foreach my $item ( @{ $object->{names} } ) {
	    if ( $item->{metadata}->{source}->{type} eq q{CONTACT} ) {
		$result->{DisplayName} = $item->{displayName} if exists $item->{displayName};
		$result->{LastName} = $item->{familyName};
		$result->{FirstName} = $item->{givenName};
	    }
	}
    }

    if ( exists $object->{photos} && defined $object->{photos} && ref $object->{photos} eq "ARRAY" ) {
	foreach my $item ( @{ $object->{photos} } ) {
	    if ( $item->{metadata}->{source}->{type} eq q{CONTACT} ) {
		$result->{Image} = $item->{url} if exists $item->{url};
	    }
	}
    }

    if ( exists $object->{emailAddresses} && defined $object->{emailAddresses} && ref $object->{emailAddresses} eq "ARRAY" ) {
	foreach my $item ( @{ $object->{emailAddresses} } ) {
	    if ( $item->{metadata}->{source}->{type} eq q{CONTACT} ) {
		$result->{Email} = $item->{value} if exists $item->{value};
	    }

	    $result->{"Email_" . $item->{value}} = $item->{value} if exists $item->{value};
	}
    }

    if ( exists $object->{phoneNumbers} && defined $object->{phoneNumbers} && ref $object->{phoneNumbers} eq "ARRAY" ) {
	foreach my $item ( @{ $object->{phoneNumbers} } ) {
	    if ( $item->{metadata}->{primary} ) {
		$result->{Phone} = $item->{canonicalForm} if exists $item->{canonicalForm};
	    }

	    if ( $item->{type} eq "business" ) { $result->{Work} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "mobile" ) { $result->{Mobile} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "home" ) { $result->{Home} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "other" ) { $result->{Other} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "assistant" ) { $result->{AssistantPhone} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "homeFax" ) { $result->{HomeFax} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "businessFax" ) { $result->{Fax} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "otherFax" ) { $result->{FaxPhone} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "pager" ) { $result->{Pager} = $item->{canonicalForm};
	    } elsif ( $item->{type} eq "radio" ) { $result->{Radio} = $item->{canonicalForm}; }
	}
    }

    if ( exists $object->{organizations} && defined $object->{organizations} && ref $object->{organizations} eq "ARRAY" ) {
	foreach my $item ( @{ $object->{organizations} } ) {
	    if ( $item->{metadata}->{source}->{type} eq q{CONTACT} || $item->{metadata}->{primary} ) {
		$result->{Company} = $item->{name} if exists $item->{name};
	    }

	    $result->{"Department_" . $item->{name}} = $item->{name} if exists $item->{name};
	}
    }

    if ( exists $object->{urls} && defined $object->{urls} && ref $object->{urls} eq "ARRAY" ) {
	foreach my $item ( @{ $object->{urls} } ) {
	    if ( $item->{metadata}->{source}->{type} eq q{CONTACT} || $item->{metadata}->{primary} ) {
		$result->{URL} = $item->{value} if exists $item->{value};
	    }
	}
    }

    if ( exists $object->{biographies} && defined $object->{biographies} && ref $object->{biographies} eq "ARRAY" ) {
	foreach my $item ( @{ $object->{biographies} } ) {
	    if ( $item->{metadata}->{source}->{type} eq q{CONTACT} || $item->{metadata}->{primary} ) {
		$result->{Description} = $item->{value} if exists $item->{value};
	    }
	}
    }

    return $result;
}

sub find_phone {
    my( $self, $phone ) = @_;
    return $self->get_by_phone($phone);
    my @result;

    my $contacts = $self->get_by_phone('phone', $phone);
    push @result, @{$contacts};

    $contacts = $self->get_by_phone('mobile', $phone);
    push @result, @{$contacts};

    if ( my ( $phone_without_lp ) = $phone =~ m/^\+(.*?)$/ ) {
	$contacts = $self->get_by_phone('phone', $phone_without_lp);
	push @result, @{$contacts};

	$contacts = $self->get_by_phone('mobile', $phone_without_lp);
	push @result, @{$contacts};
    } else {
	$contacts = $self->get_by_phone('phone', "+" . $phone);
	push @result, @{$contacts};

	$contacts = $self->get_by_phone('mobile', "+" . $phone);
	push @result, @{$contacts};
    }

    my %result = map { $_->{Id}->[0] => $_ } @result;
    @result = values %result;

    return \@result;
}

sub query_all {
    my( $self, $query, $fields ) = @_;

    return $self->get_by_phone($query);
    my $result = $self->find_phone($query);

    return $result;
}

1;
