package Integration::Contact::Source::Microsoft::Contacts;

use utf8;
use strict;
use warnings;


use URI;
use JSON::XS;
use Types::Serialiser;
use LWP::UserAgent;
use HTTP::Request;
use URI::Encode;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);

use lib '/opt/thirdlane/perl_lib/lib';

use Thirdlane::Auth::OAuth2;

our $VERSION = '1';

our $DEBUG = 0;

sub new {
    my( $class, $args ) = @_;
    my $self = bless( { }, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config} = $args;
    $self->{version} = $VERSION;

    unless ( exists $args->{credentials}->{user_id} && defined $args->{credentials}->{user_id} && length $args->{credentials}->{user_id} ) {
	die "no token";
    }

    $self->{id} = $self->conf->{id} || 0;

    $self->{uri_encode} = URI::Encode->new( { encode_reserved => 1 } );

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    $self->{client} = Thirdlane::Auth::OAuth2->new( { provider => 'microsoft', debug => $args->{debug} } );

    die "connection error" unless $self->_connect;

    return $self;
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub _w {
    my $cond = shift;
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($cond);
}

sub ua {
    my $self = shift;

    return $self->{ua}->();
}

sub get {
    my $self = shift;
    my $url = shift;

    my $req = HTTP::Request->new( GET => $url );

    return {} unless $self->ua;

    my $response = $self->ua->request( $req );

    if ( defined $response ) {
        if ( my $status = $response->header('Tlintstatus') ) {
            return { error => { provider => $response->header('Tlprovider'), status => $status } };
        }
        if ( $response->is_success() ) {
            my $response_content = $response->decoded_content();
            return  eval { $self->json->decode( $response_content ) };
        }
    }

    return {};
}

sub conf {
    my $self = shift;

    return $self->{config};
}

sub dump {
    my $self = shift;
    return Data::Dumper->Dump([@_]);
}

sub pdump {
    my $self = shift;
#    print Data::Dumper->Dump([@_]);
}

sub _connect {
    my $self = shift;

    $self->{ua} = sub { $self->{client}->build_user_agent( { user_id => $self->{config}->{credentials}->{user_id}, provider => 'microsoft' } ); };
    return undef unless $self->{ua}->();
    return $self->{ua};
}

sub get_contact_folders {
    my $self = shift;

    my $uri = URI->new( q{https://graph.microsoft.com/beta/me/contactFolders} );

    my $decoded_response = $self->get($uri->as_string);

    my @result;

    return $decoded_response if exists $decoded_response->{error};

    if ( ref $decoded_response->{value} eq "ARRAY" ) {
	foreach my $record ( @{ $decoded_response->{value} } ) {
	    push @result, $record->{id} if exists $record->{id};
	}
    }

    return \@result;
}

sub get_contact_object {
    my $self = shift;
    my $contact_id = shift;

    return undef unless defined $contact_id && length $contact_id;

    my $uri = URI->new( q{https://graph.microsoft.com/beta/me/contacts/} . $contact_id );

    my $decoded_response = $self->get($uri->as_string);

    return $decoded_response if exists $decoded_response->{error};

    if ( defined $decoded_response && exists $decoded_response->{id} ) {
	return $decoded_response;
    }

    return undef;
}

sub get_by_phone {
    my $self = shift;
#    my $type = shift;
    my $phone = shift;

    return [ ] unless defined $phone && length $phone;

#    my $cache_results = $self->{client}->cache->get('oauth::provider::' . $self->{client}->provider->name . '::contacts::phone::' . $phone);
#    unless ( defined $cache_results ) {
    my @result;

    my $folders = $self->get_contact_folders();

    return $folders if ref $folders eq 'HASH' && exists $folders->{error};

    if ( defined $folders && ref $folders eq "ARRAY" && scalar @{ $folders } ) {
	foreach my $folder ( @{ $folders } ) {
	    my $decoded_response = { };
	    while (1) {
	    my $uri = URI->new( q{https://graph.microsoft.com/v1.0/me/contactFolders/} . $folder . q{/contacts} );

	    my $query = { '$search' => '"' . $phone . '"' };
	    ( $query->{'$skip'} ) = $decoded_response->{'@odata.nextLink'} =~ m/^.*?\=(\d+)$/ if defined $decoded_response && exists $decoded_response->{'@odata.nextLink'} && defined $decoded_response->{'@odata.nextLink'} && length $decoded_response->{'@odata.nextLink'};
	    $uri->query_form($query) if defined $phone && length $phone;

	    $decoded_response = $self->get($uri->as_string);

	    return $decoded_response if exists $decoded_response->{error};

	    if ( ref $decoded_response->{value} eq "ARRAY" ) {
		foreach my $record ( @{ $decoded_response->{value} } ) {
		    push @result, $self->_convert_record_contact($record);
		}
	    }
	    last unless defined $decoded_response && exists $decoded_response->{'@odata.nextLink'} && defined $decoded_response->{'@odata.nextLink'} && length $decoded_response->{'@odata.nextLink'};
	    }
	}
    }

    my $decoded_response = { };
    while (1) {
    my $uri = URI->new( q{https://graph.microsoft.com/v1.0/me/contacts} );
    my $query = { '$search' => '"' . $phone . '"' };
    ( $query->{'$skip'} ) = $decoded_response->{'@odata.nextLink'} =~ m/^.*?\=(\d+)$/ if defined $decoded_response && exists $decoded_response->{'@odata.nextLink'} && defined $decoded_response->{'@odata.nextLink'} && length $decoded_response->{'@odata.nextLink'};
    $uri->query_form($query) if defined $phone && length $phone;

    $decoded_response = $self->get($uri->as_string);

    return $decoded_response if exists $decoded_response->{error};

    if ( ref $decoded_response->{value} eq "ARRAY" ) {
	foreach my $record ( @{ $decoded_response->{value} } ) {
	    push @result, $self->_convert_record_contact($record);
	}
    }
    last unless defined $decoded_response && exists $decoded_response->{'@odata.nextLink'} && defined $decoded_response->{'@odata.nextLink'} && length $decoded_response->{'@odata.nextLink'};
    }

    $decoded_response = { };
    while (1) {
    my $uri = URI->new( q{https://graph.microsoft.com/beta/me/people} );
    my $query = { '$search' => '"' . $phone . '"' };
    ( $query->{'$skip'} ) = $decoded_response->{'@odata.nextLink'} =~ m/^.*?\=(\d+)$/ if defined $decoded_response && exists $decoded_response->{'@odata.nextLink'} && defined $decoded_response->{'@odata.nextLink'} && length $decoded_response->{'@odata.nextLink'};
    $uri->query_form($query) if defined $phone && length $phone;

    $decoded_response = $self->get($uri->as_string);

    return $decoded_response if exists $decoded_response->{error};

    if ( ref $decoded_response->{value} eq "ARRAY" ) {
	foreach my $record ( @{ $decoded_response->{value} } ) {
	    push @result, $self->_convert_record_people_beta($record);
	}
    }
    last unless defined $decoded_response && exists $decoded_response->{'@odata.nextLink'} && defined $decoded_response->{'@odata.nextLink'} && length $decoded_response->{'@odata.nextLink'};
    }

    my $contacts_hash = { };

    foreach my $record ( @result ) {
	$contacts_hash->{$record->{Id}} = $record if exists $record->{Id};
    }

    @result = ( );

    foreach my $record_id ( keys %{ $contacts_hash } ) {
	push @result, $contacts_hash->{$record_id};
    }

#    $self->{client}->cache->set('oauth::provider::' . $self->{client}->provider->name . '::contacts::phone::' . $phone, \@result, 60*5)
#        if scalar @result;

    return \@result;
#    }

#    return defined $cache_results ? $cache_results : [ ];
}

sub _convert_record_contact {
    my $self = shift;
    my ( $object ) = @_;

    my $full_contact_object = $self->get_contact_object( $object->{id} );

    return $full_contact_object if exists $full_contact_object->{error};

    return $self->_convert_record_contact_beta( $full_contact_object ) if defined $full_contact_object;

    my $result = {
	'FirstName' => $object->{givenName} || "",
	'LastName' => $object->{surname} || "",
	'Description' => $object->{personalNotes} || "",
	'Image' => '',
	'Email' => '',
	'Phone' => "",
	'Company' => $object->{companyName} || "",
	'Id' => $object->{id},
	'type' => 'Contact',
	'contact_source' => 'microsoft',
	'contact_source_contact_id' => $object->{id},
	'contact_source_type' => 'Microsoft',
	'contact_source' => $self->{id}
    };

    $result->{Email} = join ( ",", map { $_->{address} } @{ $object->{emailAddresses} } )
	if exists $object->{emailAddresses}
	    && defined $object->{emailAddresses};


    $result->{Home} = join ( ",", @{ $object->{homePhones} } )
	if exists $object->{homePhones}
	    && defined $object->{homePhones};

    $result->{mobile} = $object->{mobilePhone}
	if exists $object->{mobilePhone}
	    && defined $object->{mobilePhone};

    $result->{Work} = join ( ",", @{ $object->{businessPhones} } )
	if exists $object->{businessPhones}
	    && defined $object->{businessPhones};

    return $result;
}


sub _convert_record_contact_beta {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => $object->{givenName} || "",
	'LastName' => $object->{surname} || "",
	'Description' => $object->{personNotes} || "",
	'Image' => '',
	'Email' => ( ref $object->{emailAddresses} eq "ARRAY" && scalar @{ $object->{emailAddresses} } ? $object->{emailAddresses}->[0]->{address} : "" ),
	'Phone' => "",
	'Company' => $object->{companyName} || "",
	'Id' => $object->{id},
	'type' => 'Contact',
	'contact_source' => 'microsoft',
	'contact_source_contact_id' => $object->{id},
	'contact_source_type' => 'Microsoft',
	'contact_source' => $self->{id}
    };

    if ( ref $object->{phones} eq "ARRAY" ) {
	foreach my $phone ( @{ $object->{phones} } ) {
	    if ( $phone->{type} eq "business" ) { $result->{Work} = $phone->{number};
	    } elsif ( $phone->{type} eq "mobile" ) { $result->{Mobile} = $phone->{number};
	    } elsif ( $phone->{type} eq "home" ) { $result->{Home} = $phone->{number};
	    } elsif ( $phone->{type} eq "other" ) { $result->{Other} = $phone->{number};
	    } elsif ( $phone->{type} eq "assistant" ) { $result->{AssistantPhone} = $phone->{number};
	    } elsif ( $phone->{type} eq "homeFax" ) { $result->{HomeFax} = $phone->{number};
	    } elsif ( $phone->{type} eq "businessFax" ) { $result->{Fax} = $phone->{number};
	    } elsif ( $phone->{type} eq "otherFax" ) { $result->{FaxPhone} = $phone->{number};
	    } elsif ( $phone->{type} eq "pager" ) { $result->{Pager} = $phone->{number};
	    } elsif ( $phone->{type} eq "radio" ) { $result->{Radio} = $phone->{number};
	    }
	}
    }

    return $result;
}

sub _convert_record_people_beta {
    my $self = shift;
    my $object = shift;

    my $result = {
	'FirstName' => $object->{givenName} || "",
	'LastName' => $object->{surname} || "",
	'Description' => $object->{personalNotes} || "",
	'Image' => '',
	'Email' => ( ref $object->{emailAddresses} eq "ARRAY" && scalar @{ $object->{emailAddresses} } ? $object->{emailAddresses}->[0]->{address} : "" ),
	'Phone' => "",
	'Company' => $object->{companyName} || "",
	'Id' => $object->{id},
	'type' => 'Contact',
	'contact_source' => 'microsoft',
	'contact_source_contact_id' => $object->{id},
	'contact_source_type' => 'Microsoft',
	'contact_source' => $self->{id}
    };

    if ( ref $object->{phones} eq "ARRAY" ) {
	foreach my $phone ( @{ $object->{phones} } ) {
	    if ( $phone->{type} eq "business" ) { $result->{Work} = $phone->{number};
	    } elsif ( $phone->{type} eq "mobile" ) { $result->{Mobile} = $phone->{number};
	    } elsif ( $phone->{type} eq "home" ) { $result->{Home} = $phone->{number};
	    } elsif ( $phone->{type} eq "other" ) { $result->{Other} = $phone->{number};
	    } elsif ( $phone->{type} eq "assistant" ) { $result->{AssistantPhone} = $phone->{number};
	    } elsif ( $phone->{type} eq "homeFax" ) { $result->{HomeFax} = $phone->{number};
	    } elsif ( $phone->{type} eq "businessFax" ) { $result->{Fax} = $phone->{number};
	    } elsif ( $phone->{type} eq "otherFax" ) { $result->{FaxPhone} = $phone->{number};
	    } elsif ( $phone->{type} eq "pager" ) { $result->{Pager} = $phone->{number};
	    } elsif ( $phone->{type} eq "radio" ) { $result->{Radio} = $phone->{number};
	    }
	}
    }

    return $result;
}


sub find_phone {
    my( $self, $phone ) = @_;
    return $self->get_by_phone($phone);
    my @result;

    my $contacts = $self->get_by_phone('phone', $phone);
    push @result, @{$contacts};

    $contacts = $self->get_by_phone('mobile', $phone);
    push @result, @{$contacts};

    if ( my ( $phone_without_lp ) = $phone =~ m/^\+(.*?)$/ ) {
	$contacts = $self->get_by_phone('phone', $phone_without_lp);
	push @result, @{$contacts};

	$contacts = $self->get_by_phone('mobile', $phone_without_lp);
	push @result, @{$contacts};
    } else {
	$contacts = $self->get_by_phone('phone', "+" . $phone);
	push @result, @{$contacts};

	$contacts = $self->get_by_phone('mobile', "+" . $phone);
	push @result, @{$contacts};
    }

    my %result = map { $_->{Id}->[0] => $_ } @result;
    @result = values %result;

    return \@result;
}

sub query_all {
    my( $self, $query, $fields ) = @_;

    return $self->get_by_phone($query);
    my $result = $self->find_phone($query);

    return $result;
}

1;
