package Integration::Contact::Source::Multiple;

use utf8;
use strict;
use warnings;

use Integration::CRM;

use Cache::Memcached::Fast::Safe;

use Data::Dumper qw(Dumper);

use lib '/opt/thirdlane/perl_lib/lib';

our $VERSION = '1';

our $DEBUG = 0;

sub new {
    my ( $class, $args ) = @_;
    my $self = bless( {}, $class );

    $DEBUG = $args->{base_debug} if exists $args->{base_debug};

    $self->{config}  = $args;
    $self->{version} = $VERSION;

    $self->{sorces} = $args->{sources} || [];

    $self->{config}->{cache}->{enabled} =
      exists $args->{cache}
      ? ( ref $args->{cache} eq "" ? $args->{cache} : 0 )
      : 0;
    $self->{config}->{cache}->{enabled} =
      exists $args->{cache}->{enabled}
      ? $args->{cache}->{enabled}
      : $self->{config}->{cache}->{enabled};

    $self->{config}->{cache}->{ttl} =
      exists $args->{ttl} ? $args->{ttl} : 5 * 60;    # 5 mins cache ttl

    $self->{config}->{cache}->{enabled} = 0;

    $self->{config}->{credentials} = delete $args->{credentials};

    if ( $self->{config}->{cache}->{enabled} ) {
        $self->{cache} = new Cache::Memcached::Fast::Safe {
            servers            => ["127.0.0.1:11211"],
            serialize_methods  => [ \&Storable::freeze, \&Storable::thaw ],
            utf8               => 1,
            compress_threshold => -1,
            max_size           => 512 * 1024,
            hash_namespace     => 1,
            nowait             => 1
        };

        #$self->{cache}->set_compress_threshold(10_000);
        #$self->{cache}->enable_compress(1);
    }

    return $self;
}

sub add_source {
    my $self     = shift;
    my $settings = shift;

    my $id = $settings->{id} || 0;

    if ( exists $settings->{data} ) {
        foreach my $key ( keys %{ $settings->{data} } ) {
            $settings->{$key} = $settings->{data}->{$key};
        }
    }

    if ( exists $settings->{user} ) {
        foreach my $key ( keys %{ $settings->{user} } ) {
            $settings->{$key} = $settings->{user}->{$key};
        }
    }

    if (   ( exists $settings->{login} && defined $settings->{login} )
        && ( !exists $settings->{username} && !defined $settings->{username} ) )
    {
        $settings->{username} = $settings->{login};
    }

    if ( exists $settings->{contact_source_type} ) {
        $settings->{type} = $settings->{contact_source_type};
    }

    my $crm_handle = eval {
        Integration::CRM->new(
            {
                id    => $id,
                url   => $settings->{server_url},
                debug => (
                    exists $settings->{debug} ? $settings->{debug}
                    : $self->{config}->{debug}
                ),
                type        => lc $settings->{type},
                credentials => (
                    exists $settings->{credentials} ? $settings->{credentials}
                    : $self->{config}->{credentials}
                ),
                user => {
                    login    => $settings->{username},
                    password => $settings->{password},
                    token    => $settings->{api_token},
                    scope    => $settings->{api_scope}
                }
            }
        );
    };

    die $@ if $@;

    if ( $self->{config}->{cache}->{enabled} ) {
        my $cache_prefix = sprintf( "crm:%s:%s:%s:%s",
            $settings->{type}       || "",
            $settings->{server_url} || "",
            $settings->{username}   || "",
            $settings->{password}   || "",
            $settings->{api_token}  || "",
            $settings->{api_scope}  || "" );

        #simple crc?
        $cache_prefix = sprintf( "crm:%s:", unpack( "%64A*", $cache_prefix ) );
        push @{ $self->{sources} },
          {
            cache_prefix => $cache_prefix,
            handle       => $crm_handle,
            settings     => $settings
          }
          if defined $crm_handle;
    }
    else {
        push @{ $self->{sources} },
          { handle => $crm_handle, settings => $settings }
          if defined $crm_handle;
    }

    return $self->{sources};
}

sub sources {
    my $self = shift;
    return $self->{sources};
}

sub crm {
    my $self = shift;
    return $self->{crm};
}

sub detect_type_by_id {
    my ( $self, $id ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->detect_type_by_id($id);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub get_field_list {
    my ( $self, $object ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->get_field_list($object);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub get_fields {
    my ( $self, $object ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->get_field_list($object);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub _connect {
    my $self = shift;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->_connect();
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub upsert {
    my ( $self, $args ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->upsert($args);

        if ( ref $query_result eq "ARRAY" ) {
            push @{$result}, @{$query_result};
        }
        elsif ( ref $query_result eq "HASH" ) {
            push @{$result}, $query_result;
        }
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub find {
    my ( $self, $args ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->find($args);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub query {
    my ( $self, $args ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->query($args);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub query_tasks_by_rec_id {
    my ( $self, $rec_id ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->query_tasks_by_rec_id($rec_id);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub query_tasks_by_id {
    my ( $self, $id ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->query_tasks_by_id($id);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub query_leads_by_id {
    my ( $self, $id ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->query_leads_by_id($id);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub query_contacts_by_id {
    my ( $self, $id ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->query_contacts_by_id($id);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub query_accounts_by_id {
    my ( $self, $id ) = @_;

    my $result = [];

    foreach my $source ( @{ $self->{sources} } ) {
        my $query_result = $source->{handle}->query_accounts_by_id($id);
        push @{$result}, @{$query_result} if $query_result;
    }

    return $result->[0]
      if ref $result eq "ARRAY" && scalar @{$result} && defined $result->[0];
    return $result;
}

sub query_all {
    my ( $self, $query, $fields ) = @_;

    my $result = [];

    my $cache_key = $query;

    if ( defined $fields && ref $fields eq "ARRAY" && scalar @{$fields} ) {
        $cache_key = $query . ":fields:" . join ",", @{$fields};
    }

    foreach my $source ( @{ $self->{sources} } ) {
        if ( exists $self->{config}->{cache}->{enabled}
            && $self->{config}->{cache}->{enabled} )
        {
            my $cached_contact =
              $self->{cache}->get( $source->{cache_prefix} . $cache_key );

            unless ( defined $cached_contact
                && ref $cached_contact eq "ARRAY"
                && scalar @{$cached_contact} )
            {

                my $query_result =
                  $source->{handle}->query_all( $query, $fields );
                if (   defined $query_result
                    && ref $query_result eq 'HASH'
                    && exists $query_result->{error} )
                {
                    push @{$result}, $query_result;
                }
                else {

                    if (   defined $query_result
                        && ref $query_result eq "ARRAY"
                        && scalar @{$query_result} )
                    {

                        $self->{cache}
                          ->set( $source->{cache_prefix} . $cache_key,
                            $query_result, $self->{config}->{cache}->{ttl} );
                    }
                    push @{$result}, @{$query_result} if $query_result;
                }
            }
            else {
                push @{$result}, @{$cached_contact} if $cached_contact;
            }
        }
        else {
            my $query_result = $source->{handle}->query_all( $query, $fields );
            if (   defined $query_result
                && ref $query_result eq 'HASH'
                && exists $query_result->{error} )
            {
                push @{$result}, $query_result;
            }
            else {
                push @{$result}, @{$query_result} if $query_result;
            }
        }
    }

    return $result;
}

sub find_phone {
    my ( $self, $query ) = @_;

    my $result = [];

    my $cache_key = $query;

    foreach my $source ( @{ $self->{sources} } ) {
        if ( exists $self->{config}->{cache}->{enabled}
            && $self->{config}->{cache}->{enabled} )
        {
            my $cached_contact =
              $self->{cache}->get( $source->{cache_prefix} . $cache_key );

            unless ( defined $cached_contact
                && ref $cached_contact eq "ARRAY"
                && scalar @{$cached_contact} )
            {

                my $query_result = $source->{handle}->find_phone($query);
                if (   defined $query_result
                    && ref $query_result eq 'HASH'
                    && exists $query_result->{error} )
                {
                    push @{$result}, $query_result;
                }
                else {

                    if (   defined $query_result
                        && ref $query_result eq "ARRAY"
                        && scalar @{$query_result} )
                    {

                        $self->{cache}
                          ->set( $source->{cache_prefix} . $cache_key,
                            $query_result, $self->{config}->{cache}->{ttl} );
                    }
                    push @{$result}, @{$query_result} if $query_result;
                }
                push @{$result}, @{$query_result} if $query_result;
            }
            else {
                push @{$result}, @{$cached_contact} if $cached_contact;
            }
        }
        else {
            my $query_result = eval { $source->{handle}->find_phone($query) };
            if (   defined $query_result
                && ref $query_result eq 'HASH'
                && exists $query_result->{error} )
            {
                push @{$result}, $query_result;
            }
            else {
                push @{$result}, @{$query_result} if $query_result;
            }
        }
    }

    return $result;
}

1;
