package Integration::Contact::Source::Zoho::Contacts;

use utf8;
use strict;
use warnings;

use URI;
use JSON::XS;
use Types::Serialiser;
use LWP::UserAgent;
use HTTP::Request;
use URI::Encode;
use Data::Dumper;
use POSIX qw(strftime);
use Digest::MD5 qw(md5_hex);

use lib '/opt/thirdlane/perl_lib/lib';

use Thirdlane::Auth::OAuth2;

our $VERSION = '1';

our $DEBUG = 0;

my %object_prefixes = (
    ACCOUNT => '001',
    CONTACT => '003',
    LEAD    => '00Q',
    TASK    => '00T'
);

sub new {
    my ( $class, $args ) = @_;
    my $self = bless( {}, $class );

    $DEBUG = $args->{debug} if exists $args->{debug};

    $self->{config}  = $args;
    $self->{version} = $VERSION;

    unless ( exists $args->{credentials}->{user_id}
        && defined $args->{credentials}->{user_id}
        && length $args->{credentials}->{user_id} )
    {
        die "no token";
    }

    $self->{id} = $self->conf->{id} || 0;

    $self->{uri_encode} = URI::Encode->new( { encode_reserved => 1 } );

    $self->{json} = JSON::XS->new->allow_nonref->utf8;

    $self->{client} = Thirdlane::Auth::OAuth2->new(
        { provider => 'zoho', debug => $args->{debug} } );

    die "connection error" unless $self->_connect;

    return $self;
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub _w {
    my $cond = shift;
    my $uri = URI::Encode->new( { encode_reserved => 1 } );
    return $uri->encode($cond);
}

sub ua {
    my $self = shift;

    return $self->{ua}->();
}

sub get {
    my $self = shift;
    my $url  = shift;

    my $req = HTTP::Request->new( GET => $url );

    return {} unless $self->ua;

    my $response = $self->ua->request($req);

    if ( defined $response ) {
        if ( my $status = $response->header('Tlintstatus') ) {
            return {
                error => {
                    provider => $response->header('Tlprovider'),
                    status   => $status
                }
            };
        }

        if ( $response->is_success() ) {
            my $response_content = $response->decoded_content();
            return eval { $self->json->decode($response_content) };
        }
    }

    return {};
}

sub conf {
    my $self = shift;

    return $self->{config};
}

sub dump {
    my $self = shift;
    return Data::Dumper->Dump( [@_] );
}

sub pdump {
    my $self = shift;

    #    print Data::Dumper->Dump([@_]);
}

sub _connect {
    my $self = shift;

    $self->{ua} = sub {
        $self->{client}->build_user_agent(
            {
                user_id  => $self->{config}->{credentials}->{user_id},
                provider => 'zoho'
            }
        );
    };

    return $self->{ua} = undef unless $self->{ua}->();

    return $self->{ua};
}

sub pmatch {
    my $self = shift;
    my ( $obj, $phone ) = @_;

    foreach my $record ( keys %{$obj} ) {
        if ( index( lc $obj->{$record}, lc $phone ) != -1 ) {
            return 1;
        }
    }
    return 0;
}

sub get_by_phone {
    my $self = shift;

    #    my $type = shift;
    my $phone = shift;

    return [] unless defined $phone && length $phone;

#    my $cache_results = $self->{client}->cache->get('oauth::provider::' . $self->{client}->provider->name . '::contacts::phone::' . $phone);

    #    unless ( defined $cache_results ) {
    my @result;
    my $uri = URI->new(q{https://www.zohoapis.com/crm/v2/ALL/search});

    my $query = { word => $phone };

    my $decoded_response = {};

    $uri->query_form($query);

    $decoded_response = $self->get( $uri->as_string );

#    last if defined $decoded_response && exists $decoded_response->{error}; #RESOURCE_EXHAUSTED

    if ( ref $decoded_response->{data} eq "ARRAY" ) {
        foreach my $record ( @{ $decoded_response->{data} } ) {
            my $conversion = $self->_convert_record_contact($record);
            if ( $self->pmatch( $conversion, $phone ) ) {
                push @result, $conversion;
            }
        }
    }

#    $self->{client}->cache->set('oauth::provider::' . $self->{client}->provider->name . '::contacts::phone::' . $phone, \@result, 60*5)
#	if scalar @result;

    return \@result;

    #    }

    #    return defined $cache_results ? $cache_results : [ ];
}

sub _convert_record_contact {
    my $self   = shift;
    my $object = shift;

    my $result = {
        'FirstName'                 => '',
        'LastName'                  => '',
        'Description'               => '',
        'Email'                     => '',
        'Phone'                     => '',
        'Company'                   => '',
        'Id'                        => $object->{id},
        'type'                      => 'Contact',
        'contact_source'            => 'zoho',
        'contact_source_contact_id' => $object->{id},
        'contact_source_type'       => 'Zoho',
        'contact_source'            => $self->{id}
    };

    if (   exists $object->{Owner}
        && defined $object->{Owner}
        && ref $object->{Owner} eq "HASH" )
    {
        $result->{DisplayName} = $object->{Owner}->{name};
        $result->{LastName}    = ( split /\s/, $object->{Owner}->{name} )[1];
        $result->{FirstName}   = ( split /\s/, $object->{Owner}->{name} )[0];
        $result->{Email}       = $object->{Owner}->{email};
    }

    if (   exists $object->{organizations}
        && defined $object->{organizations}
        && ref $object->{organizations} eq "ARRAY" )
    {
        foreach my $item ( @{ $object->{organizations} } ) {
            if (   $item->{metadata}->{source}->{type} eq q{CONTACT}
                || $item->{metadata}->{primary} )
            {
                $result->{Company} = $item->{name} if exists $item->{name};
            }

            $result->{ "Department_" . $item->{name} } = $item->{name}
              if exists $item->{name};
        }
    }

    if (   exists $object->{urls}
        && defined $object->{urls}
        && ref $object->{urls} eq "ARRAY" )
    {
        foreach my $item ( @{ $object->{urls} } ) {
            if (   $item->{metadata}->{source}->{type} eq q{CONTACT}
                || $item->{metadata}->{primary} )
            {
                $result->{URL} = $item->{value} if exists $item->{value};
            }
        }
    }

    if (   exists $object->{biographies}
        && defined $object->{biographies}
        && ref $object->{biographies} eq "ARRAY" )
    {
        foreach my $item ( @{ $object->{biographies} } ) {
            if (   $item->{metadata}->{source}->{type} eq q{CONTACT}
                || $item->{metadata}->{primary} )
            {
                $result->{Description} = $item->{value}
                  if exists $item->{value};
            }
        }
    }

    return $result;
}

sub find_phone {
    my ( $self, $phone ) = @_;
    return $self->get_by_phone($phone);
}

sub query_all {
    my ( $self, $query, $fields ) = @_;

    return $self->get_by_phone($query);
}

sub _format_url_args {
    my ( $self, $args ) = @_;

    my $result;

    foreach my $arg ( keys %{$args} ) {
        $arg =~ s{\"}{}sg;
        push @{$result}, sprintf( "%s=%s", $arg, $args->{$arg} );
    }

    return join "&", @{$result};
}

sub _get_json {
    my ( $self, $uri, $args ) = @_;

    my $url = sprintf(
        '%s/%s?authtoken=%s&scope=%s',
        $self->conf->{url},
        $uri,
        $self->conf->{user}->{token},
        $self->conf->{scope}
    );

    $url .= '&' . $self->_format_url_args($args) if $args;

    printf( "\nGET URL:%s\n", $url ) if $DEBUG;

    my $response = $self->ua->get($url);

    if ( $response->is_success ) {
        print $response->decoded_content if $DEBUG;
    }
    else {
        die $response->status_line;
    }

    my $decoded_response = $self->json->decode( $response->decoded_content );

    return undef if exists $decoded_response->{response}->{error};

    return $decoded_response;
}

sub _post_json {
    my ( $self, $uri, $args, $body ) = @_;

    $body =~ s/\n//sg;

    my $base = $self->conf->{url};

    my $url = sprintf(
        '%s/%s?authtoken=%s&scope=%s&xmlData=%s',
        $base, $uri,
        $self->conf->{user}->{token},
        $self->conf->{scope}, $body
    );

    $url .= '&' . $self->_format_url_args($args) if scalar keys %{$args};

    printf( "\nPOST URL:%s\n",  $url )  if $DEBUG;
    printf( "\nPOST Data:%s\n", $body ) if $DEBUG;

    my $response = $self->ua->post($url);

    if ( $response->is_success ) {
        print $response->decoded_content if $DEBUG;
    }
    else {
        die $response->status_line;
    }

    my $decoded_response = $self->json->decode( $response->decoded_content );

    return undef if exists $decoded_response->{response}->{error};

    if ( exists $decoded_response->{response}->{result} ) {
        foreach my $record (
            @{ $decoded_response->{response}->{result}->{recorddetail}->{FL} } )
        {
            return $record->{content} if $record->{val} eq "Id";
        }
    }

    return undef;
}

sub _convert_record_lead {
    my $self   = shift;
    my $object = shift;

    my $result = {
        'FirstName'                 => $object->{'First Name'},
        'LastName'                  => $object->{'Last Name'},
        'Description'               => $object->{'Description'},
        'Status'                    => $object->{'Lead Status'},
        'Email'                     => $object->{'Email'},
        'Phone'                     => $object->{'Phone'},
        'Mobile'                    => $object->{'Mobile'},
        'Company'                   => $object->{'Company'},
        'Id'                        => [ $object->{'LEADID'} ],
        'type'                      => 'Lead',
        'contact_source_contact_id' => $object->{'LEADID'},
        'contact_source_type'       => 'Zoho',
        'contact_source'            => $self->{id}
    };

    return $result;
}

sub _convert_record_call {
    my $self   = shift;
    my $object = shift;

    my $result = {
        'Status'                => $object->{'Call Result'},
        'Id'                    => [ $object->{'ACTIVITYID'} ],
        'Subject'               => $object->{'Subject'},
        'WhatId'                => $object->{'ACTIVITYID'},
        'CallDurationInSeconds' => $object->{'Call Duration (in seconds)'},
        'CallDisposition'       => $object->{'Call Result'},
        'type'                  => 'Task',
        'WhoId'                 => $object->{'RELATEDTOID'},
        'Description'           => $object->{'Description'},
        'Direction'             => $object->{'Call Type'},
        'Time'                  => $object->{'Call Start Time'},
        'contact_source'        => $self->{id}
    };

    $result->{'WhoId'} = $object->{'CONTACTID'}
      if exists $object->{'CONTACTID'};
    $result->{'WhoId'} = $object->{'RELATEDTOID'}
      if $object->{'SEMODULE'} eq 'Leads';

    return $result;
}

sub _convert_record_account {
    my $self   = shift;
    my $object = shift;

    my $result = {
        'Name'                => $object->{'Account Name'},
        'WhoId'               => $object->{'SMOWNERID'},
        'Id'                  => $object->{'ACCOUNTID'},
        'AccountId'           => $object->{'ACCOUNTID'},
        'Phone'               => $object->{'Phone'},
        'Mobile'              => $object->{'Mobile'},
        'contact_source_type' => "Zoho",
        'contact_source'      => $self->{id}
    };

    return $result;
}

sub _convert_record {
    my $self        = shift;
    my $object_type = shift;
    my $object      = shift;

    my $result;

    if ( $object_type eq 'Leads' ) {
        $result = $self->_convert_record_lead($object);
    }
    elsif ( $object_type eq 'Contacts' ) {
        $result = $self->_convert_record_contact($object);
    }
    elsif ( $object_type eq 'Calls' ) {
        $result = $self->_convert_record_call($object);
    }
    elsif ( $object_type eq 'Accounts' ) {
        $result = $self->_convert_record_account($object);
    }
    else {
        $result = $object;
    }

    return $result;
}

sub get_records {
    my $self   = shift;
    my $object = shift;

    my $uri = sprintf( '%s/getRecords', $object );
    my $response = $self->_get_json($uri);

    return undef unless $response;
    return undef if exists $response->{response}->{error};
    return [] if exists $response->{response}->{nodata};

    my $rows = $response->{response}->{result}->{$object}->{row};
    $rows = [$rows] if ref $rows eq "HASH";

    my $result;

    foreach my $row ( @{$rows} ) {
        my $item;

        foreach my $record ( @{ $row->{FL} } ) {
            $item->{ $record->{val} } = $record->{content};
        }

        $item = $self->_convert_record( $object, $item );
        push @{$result}, $item;
    }

    return $result;
}

sub get_record_by_id {
    my $self   = shift;
    my $object = shift;
    my $id     = shift;

    my $uri = sprintf( '%s/getRecordById', $object );
    my $response = $self->_get_json( $uri, { id => $id } );

    return undef unless $response;
    return undef if exists $response->{response}->{error};
    return [] if exists $response->{response}->{nodata};

    my $rows = $response->{response}->{result}->{$object}->{row};
    $rows = [$rows] if ref $rows eq "HASH";

    my $result;

    foreach my $row ( @{$rows} ) {
        my $item;

        foreach my $record ( @{ $row->{FL} } ) {
            $item->{ $record->{val} } = $record->{content};
        }
        $item = $self->_convert_record( $object, $item );
        push @{$result}, $item;
    }

    return $result;
}

sub get_leads {
    my $self = shift;
    return $self->get_records('Leads');
}

sub get_contacts {
    my $self = shift;
    return $self->get_records('Contacts');
}

sub get_calls {
    my $self = shift;
    return $self->get_records('Calls');
}

sub detect_type_by_id {
    my ( $self, $id ) = @_;

    my $records = $self->get_record_by_id( 'Leads', $id );
    return 'LEAD' if scalar @{$records};

    $records = $self->get_record_by_id( 'Contacts', $id );
    return 'CONTACT' if scalar @{$records};

    $records = $self->get_record_by_id( 'Calls', $id );
    return 'TASK' if scalar @{$records};

    $records = $self->get_record_by_id( 'Accounts', $id );
    return 'ACCOUNT' if scalar @{$records};

    return 'UNKNOWN';
}

sub get_field_list {
    my ( $self, $object ) = @_;
    my $response = $self->_get_json( sprintf( '%s/getFields', $object ) );

    return undef unless $response;
    return undef if exists $response->{response}->{error};

    my @fields;

    $response->{$object}->{section} = [ $response->{$object}->{section} ]
      if ref $response->{$object}->{section} eq "HASH";

    foreach my $field ( @{ $response->{$object}->{section} } ) {
        $field->{FL} = [ $field->{FL} ] if ref $field->{FL} eq "HASH";

        foreach my $item ( @{ $field->{FL} } ) {
            push @fields, $item->{dv};
        }

    }

    return \@fields;
}

sub get_fields {
    my ( $self, $object ) = @_;

    return $self->get_field_list($object);
}

sub update {
    my ( $self, $args ) = @_;

    my $id = delete $args->{record}->{Id};

    my $object = sprintf( "<%s><row no=\"1\">", $args->{object} );

    foreach my $item ( keys %{ $args->{record} } ) {
        $object .= sprintf( "<FL val=\"%s\">%s</FL>",
            $item, $args->{record}->{$item} || "" );
    }

    $object .= sprintf( "</row></%s>", $args->{object} );

    return $self->_post_json( $args->{object} . '/updateRecords',
        { id => $id }, $object );
}

sub insert {
    my ( $self, $args ) = @_;

    my $object = sprintf( "<%s><row no=\"1\">", $args->{object} );

    foreach my $item ( keys %{ $args->{record} } ) {
        $object .= sprintf( "<FL val=\"%s\">%s</FL>",
            $item, $args->{record}->{$item} || "" );
    }

    $object .= sprintf( "</row></%s>", $args->{object} );

    return $self->_post_json( $args->{object} . '/insertRecords', {}, $object );
}

sub _task_to_call_record {
    my ( $self, $object ) = @_;

    $object->{CallDurationInSeconds} = 0
      unless $object->{CallDurationInSeconds};

    my $result = {
        "Subject"         => $object->{Subject},
        "Call Type"       => $object->{Direction} || "Inbound",
        "Description"     => $object->{Description},
        "Call Result"     => md5_hex( $object->{CallDisposition} ),
        "Call Purpose"    => "Contact",
        "Call Start Time" => $object->{Time}
        ? strftime( "\%F \%H:\%M:\%S", localtime( int( $object->{Time} ) ) )
        : strftime( "\%F \%H:\%M:\%S", localtime ),
        "Call Duration (in seconds)" => $object->{CallDurationInSeconds},
        "Call Duration"              => sprintf( "%02d:%02d",
            int( $object->{CallDurationInSeconds} / 60 ),
            int( $object->{CallDurationInSeconds} % 60 ) )
    };

    if ( $object->{WhoId} ) {

        if ( $self->detect_type_by_id( $object->{WhoId} ) eq "CONTACT" ) {
            $result->{"SEMODULE"}  = "Contacts";
            $result->{"CONTACTID"} = $object->{WhoId};
        }
        else {
            $result->{"RELATEDTOID"} = $object->{WhoId},
              $result->{"SEID"}      = $object->{WhoId},
              $result->{"SEMODULE"}  = "Leads",
              ;
        }
    }

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _contact_to_contacts_record {
    my ( $self, $object ) = @_;

    my $result = {
        "First Name"  => $object->{FirstName},
        "Last Name"   => $object->{LastName},
        "Email"       => $object->{Email},
        "Phone"       => $object->{Phone},
        "Mobile"      => $object->{Mobile},
        "Other Phone" => $object->{Other},
        "Department"  => $object->{Department},
        "Description" => $object->{Description}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _leads_to_lead_record {
    my ( $self, $object ) = @_;

    my $result = {
        "First Name"  => $object->{FirstName},
        "Last Name"   => $object->{LastName},
        "Email"       => $object->{Email},
        "Phone"       => $object->{Phone},
        "Mobile"      => $object->{Mobile},
        "Company"     => $object->{Company},
        "Lead Status" => $object->{Status},
        "Description" => $object->{Description}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _accounts_to_account_record {
    my ( $self, $object ) = @_;

    my $result = {
        "Account Name" => $object->{Name},
        "Phone"        => $object->{Phone}
    };

    $result->{Id} = $object->{Id} if exists $object->{Id};

    return $result;
}

sub _convert_to_record {
    my ( $self, $type, $object ) = @_;

    my $result;

    if ( $type eq 'Calls' ) {
        $result = $self->_task_to_call_record($object);
    }
    elsif ( $type eq 'Contacts' ) {
        $result = $self->_contact_to_contacts_record($object);
    }
    elsif ( $type eq 'Leads' ) {
        $result = $self->_leads_to_lead_record($object);
    }
    elsif ( $type eq 'Accounts' ) {
        $result = $self->_accounts_to_account_record($object);
    }
    else {
        $result = $object;
    }
}

sub upsert {
    my ( $self, $args ) = @_;

    my $want_object = delete $args->{want_object};

    $args->{object} = 'Calls'    if $args->{object} eq 'Task';
    $args->{object} = 'Contacts' if $args->{object} eq 'Contact';
    $args->{object} = 'Leads'    if $args->{object} eq 'Lead';
    $args->{object} = 'Accounts' if $args->{object} eq 'Account';

    $args->{record} =
      $self->_convert_to_record( $args->{object}, $args->{record} );

    my $result = {};

    if ( exists $args->{record}->{Id} ) {
        $result = $self->update($args);
    }
    else {
        $result = $self->insert($args);
    }

    if (   $want_object
        && ( defined $result )
        && ( ref($result) ne "HASH" )
        && ( ref($result) ne "ARRAY" ) )
    {
        $result = $self->get_record_by_id( $args->{object}, $result );
    }

    if ( ( ref($result) eq "ARRAY" ) && ( scalar @{$result} == 1 ) ) {
        $result = $result->[0];
    }

    return $result;
}

sub find {
    my ( $self, $object, $args ) = @_;

#   ( ( ( Last Name:Alex ) AND ( Company:Thirdlane ) ) OR ( Lead Status:Contacted ) )

    $self->pdump($args) if $DEBUG;

    my $uri = sprintf( '%s/searchRecords', $object );
    my $response = $self->_get_json( $uri, { criteria => '(' . $args . ')' } );

    $self->pdump($response) if $DEBUG;

    return undef unless $response;
    return undef if exists $response->{response}->{error};
    return [] if exists $response->{response}->{nodata};

    my $rows = $response->{response}->{result}->{$object}->{row};
    $rows = [$rows] if ref $rows eq "HASH";

    my $result;

    foreach my $row ( @{$rows} ) {
        my $item;

        foreach my $record ( @{ $row->{FL} } ) {
            $item->{ $record->{val} } = $record->{content};
        }

        $item = $self->_convert_record( $object, $item );
        push @{$result}, $item;
    }

    return $result;
}

sub query {
    my ( $self, $object, $args ) = @_;

    return $self->find( $object, $args );
}

sub query_by_field {
    my ( $self, $query, $fields ) = @_;

    my $lead_fields = {
        "firstname" => '(First Name:%s)',
        "lastname"  => '(Last Name:%s)',
        "comapny"   => '(Company:%s)',
        "phone"     => '(Phone:%s)',
        "email"     => '(Email:%s)'
    };

    my $contact_fields = {
        "firstname" => '(First Name:%s)',
        "lastname"  => '(Last Name:%s)',
        "comapny"   => '(Account Name:%s)',
        "phone"     => '(Phone:%s)',
        "email"     => '(Email:%s)'
    };

    if ( grep { lc $_ eq "fullname" } @{$fields} ) {
        push @{$fields}, ( 'firstname', 'lastname' );
    }

    my $leads_query =
      '(Last Name:%s)OR(First Name:%s)OR(Company:%s)OR(Phone:%s)OR(Email:%s)';
    my @leads_query_fields;

    my $contacts_query =
'(Last Name:%s)OR(First Name:%s)OR(Account Name:%s)OR(Phone:%s)OR(Email:%s)';
    my @contacts_query_fields;

    my $leads_query_fields_array    = [];
    my $contacts_query_fields_array = [];

    foreach my $field ( @{$fields} ) {
        push @{$leads_query_fields_array},
          sprintf( $lead_fields->{ lc $field }, $query, $query )
          if exists $lead_fields->{ lc $field };
        push @{$contacts_query_fields_array},
          sprintf( $contact_fields->{ lc $field }, $query, $query )
          if exists $contact_fields->{ lc $field };
    }

    $leads_query    = join " OR ", @{$leads_query_fields_array};
    $contacts_query = join " OR ", @{$contacts_query_fields_array};

    printf "leads query: %s\n",    $leads_query    if $DEBUG;
    printf "contacts query: %s\n", $contacts_query if $DEBUG;

    my $leads    = $self->find( 'Leads',    $leads_query );
    my $contacts = $self->find( 'Contacts', $contacts_query );

    my @result = ( @{$leads}, @{$contacts} );

    return \@result;
}

sub query_tasks_by_rec_id {
    my ( $self, $rec_id ) = @_;

    return $self->find( 'Calls',
        sprintf( qq{Call Result:"%s"}, md5_hex($rec_id) ) );
}

sub query_tasks_by_id {
    my ( $self, $id ) = @_;

    return $self->get_record_by_id( 'Calls', $id );
}

sub query_leads_by_id {
    my ( $self, $id ) = @_;

    return $self->get_record_by_id( 'Leads', $id );
}

sub query_contacts_by_id {
    my ( $self, $id ) = @_;

    return $self->get_record_by_id( 'Contacts', $id );
}

sub query_accounts_by_id {
    my ( $self, $id ) = @_;

    return $self->get_record_by_id( 'Accounts', $id );
}

1;
