package Thirdlane::Auth::OAuth2;

use utf8;
use strict;
use warnings;

#use only::latest;

use DBI;

use List::Util;
use File::Slurp;
use JSON::XS qw/decode_json encode_json/;

#use Cache::Memcached::Fast;

use Data::Dumper qw(Dumper);

use Data::UUID;
use Crypt::JWT;
use Class::Load ();
use SQL::Abstract;
use HTTP::Response;
use HTTP::Request;

our $VERSION = '1.00001';

use constant {
    AUTHENTICATION_TTL     => ( 60 * 60 * 24 ),
    AUTHENTICATION_MAX_TTL => ( 60 * 60 * 24 * 31 * 12 )
};

sub new {
    my $self = shift;
    my ($opts) = @_;

    my $class = ref($self) || $self;

    $opts->{initialized} = 0;

    my $obj = bless { %{$opts} }, $class;

    $obj->init();

    return $obj;
}

sub init {
    my $self = shift;

    $self->{debug} = 0
      unless exists $self->{debug}
      && defined $self->{debug}
      && $self->{debug} =~ m/^\d+$/;

    #    $self->{debug} = 2;
    #	namespace => 'oauth::',
    #    $self->{cache}->{memory} = new Cache::Memcached::Fast::Safe({
    #	servers => [ { address => '127.0.0.1:11211' } ],
    #	serialize_methods => [ \&Storable::freeze, \&Storable::thaw ],
    #	compress_threshold => 10_000,
    #	utf8 => 1
    #    });

#    if ( my $oauth_debug = $self->{cache}->{memory}->get("thirdlane::debug::oauth") ) { $self->{debug} = 3 if $oauth_debug eq "on"; }

#Thirdlane::Cache->new( namespace => 'Storage::Memory::Memcached::Encrypted', l1_cache => { storage => 'memory_storage_local' } );

#    $self->{cache}->{disk} = $self->{cache}->{memory}; #Thirdlane::Cache->new( namespace => 'Storage::Config::Encrypted' );

    $self->{sql} = SQL::Abstract->new( quote_char => '`' );
    $self->{instance}->{config} = $self->get_config();
    $self->{instance}->{webmin}->{port} = $self->get_webmin_port();
    $self->{dbh} = sub { $self->db_connect() };
    $self->{json} = $self->get_json_codec();

    $self->{providers_avail} = $self->get_providers();
    $self->{provider_class}  = $self->load_provider( $self->{provider} );
    $self->{provider_config} =
      $self->get_provider_config( { provider => $self->{provider} } );
    $self->{provider_handle} =
      $self->{provider_class}->new( $self->{provider_config} );

}

sub build_provider {
    my $self = shift;
    my $provider = shift || 'generic';

    my $provider_class = $self->load_provider($provider);
    my $provider_config =
      $self->get_provider_config( { provider => $provider } );

    return $provider_class->new($provider_config);
}

#sub cache {
#    my $self = shift;
#    my $type = shift || 'memory';
#    $type = 'memory' unless exists $self->{cache}->{$type};
#    return $self->{cache}->{$type};
#}

sub dbh {
    my $self = shift;

    return $self->{dbh}->();
}

sub get_subname {
    my $self = shift;
    my ($info) = @_;

    if ( defined( $info->{evaltext} ) ) {
        my $eval = $info->{evaltext};
        if ( $info->{is_require} ) {
            return "require $eval";
        }
        else {
            $eval =~ s/([\\\'])/\\$1/g;
            return "eval '" . $eval . "'";
        }
    }

    return ( $info->{sub} eq '(eval)' ) ? 'eval {...}' : $info->{sub};
}

sub caller_info {
    my $self = shift;
    my $i    = shift(@_) + 1;

    my %call_info;
    {

        package DB;
        our @args = \$i;
        @call_info{
            qw(pack file line sub has_args wantarray evaltext is_require)} =
          defined &{"CORE::GLOBAL::caller"}
          ? &{"CORE::GLOBAL::caller"}($i)
          : caller($i);
    }

    unless ( defined $call_info{pack} ) {
        return ();
    }

    my $sub_name = $self->get_subname( \%call_info );

    if ( $call_info{has_args} ) {
        if (   @DB::args == 1
            && ref $DB::args[0] eq ref \$i
            && $DB::args[0] == \$i )
        {
            @DB::args = ();
            $sub_name .= '(void)';
        }
        else {
            $sub_name .= '(' . $self->_args_dumper(@DB::args) . ')';
        }
    }

    $call_info{sub_name} = $sub_name;

    return wantarray() ? %call_info : \%call_info;
}

sub back_trace_stack {
    my $self = shift;
    my $i    = shift;

    $i++;

    my %i = $self->caller_info($i);

    my $call_stack_info = ["Endpoint at $i{file} line $i{line}"];

    while ( my %i = $self->caller_info( ++$i ) ) {
        push @{$call_stack_info},
          "\t$i{sub_name} called at $i{file} line $i{line}";
    }

    return $call_stack_info;
}

sub debug {
    my $self = shift;
    my (@text) = @_;

    return 1 unless $self->{debug};
    return 1 unless @text;

    return $self->_log(@_);
}

sub _log {
    my $self = shift;
    my (@text) = @_;

    return 1 unless $self->{debug};
    return 1 unless @text;

    my ( $package, $path, $line ) = caller(1);

    my $log_buffer = "\n" . "=" x 80 . "\n";

    $log_buffer .= "["
      . DateTime->now->strftime("%F %T.%3N") . "]: "
      . $package . " at "
      . $line . ":\n";

    $log_buffer .= join "\n",
      map { ref $_ ? $self->_dumper($_) : defined $_ ? $_ : "" } @text;

    my $call_stack_info = $self->back_trace_stack(3);
    $log_buffer .= "\n" . "=" x 80 . "\n";
    $log_buffer .= join "\n", @{$call_stack_info}
      if defined $call_stack_info && ref $call_stack_info eq "ARRAY";
    $log_buffer .= "\n" . "=" x 80 . "\n";

    print $log_buffer if $self->{debug} > 10;

    write_file( '/var/log/thirdlane/oauth2.log', { append => 1 }, $log_buffer );

    return 1;
}

sub _dumper {
    my $self = shift;

    eval { require Data::Dumper; };

    if ($@) {
        return eval { $self->{json}->encode(@_) };
    }

    return Data::Dumper->new( [@_] )->Purity(0)->Useqq(0)->Terse(1)
      ->Deepcopy(1)->Quotekeys(0)->Maxdepth(32)->Dump;
}

sub _args_dumper {
    my $self = shift;

    my $args = [];

    foreach my $arg (@_) {
        if ( ref $arg eq "" ) {
            push @{$args}, ( defined $arg ? sprintf( "'%s'", $arg ) : 'undef' );
        }
        elsif ( ref $arg eq "ARRAY" ) {
            push @{$args}, eval { $self->{json}->encode($arg) };
        }
        elsif ( ref $arg eq "HASH" ) {
            push @{$args}, eval { $self->{json}->encode($arg) };
        }
        elsif ( ref $arg eq "CODE" ) {
            push @{$args}, sprintf( "[%s]", $arg );
        }
        elsif ( ref $arg eq "SCALAR" ) {
            push @{$args}, sprintf( "'%s'", $arg );
        }
        elsif ( !defined( ref($arg) ) ) {
            push @{$args}, sprintf( "'%s'", $arg );
        }
        else {
            push @{$args},
              ( defined $arg ? sprintf( "[%s]", ref $arg ) : 'undef' );
        }
    }

    return join ", ", @{$args};
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub providers {
    my $self = shift;

    return $self->{providers_avail};
}

sub provider {
    my $self = shift;

    return $self->{provider_handle};
}

sub provider_config {
    my $self = shift;

    return $self->{provider_config}->{provider};
}

sub merge {
    my $self = shift;

    my $result = {};

    foreach my $arg (@_) {
        if ( defined $arg && ref $arg eq "HASH" ) {
            foreach my $key ( keys %{$arg} ) {
                $result->{$key} = $arg->{$key};
            }
        }
    }

    return $result;
}

sub get_json_codec {
    my $self = shift;

    my $json = JSON::XS->new();

    $json->allow_nonref(1);
    $json->allow_blessed(1);
    $json->utf8(1);

    return $json;
}

sub get_config {
    my $self = shift;

    my $config = {};

    #    my $config = $self->cache->get('config::instances::hash');

    #    unless ( defined $config ) {

    my @config_file = read_file("/etc/webmin/asterisk/instances.txt");
    for (@config_file) {
        $config->{$1} = $2 if m/^(.*?)\s*\=\s*(.*?)$/;
    }

    #	$self->cache->set('config::instances::hash', $config, 60*60*12);

    #    }

    return $config;
}

sub get_webmin_port {
    my $self = shift;
    if ( -f "/etc/nginx/conf.d/https/upstream/webmin.conf" ) {
        my $webmin_conf =
          eval { read_file("/etc/nginx/conf.d/https/upstream/webmin.conf") };
        my ($port) = $webmin_conf =~ m/server\s+.*?\:(\d+)\;/sg if $webmin_conf;

        return $port if $port;
    }

    return '22000';
}

sub db_connect {
    my $self = shift;

    my $dbh = eval { $self->db_connect_cached() };
    $dbh = eval { $self->db_connect_cached(1) } if $@;

    return undef unless defined $dbh;
    return $dbh;
}

sub db_connect_cached {
    my $self  = shift;
    my $force = shift;

    use feature 'state';

    state $dbh;

    if ( $force || !defined $dbh ) {

        my $db_attr = {
            mysql_auto_reconnect => 1,
            mysql_enable_utf8    => 1,
            RaiseError           => 0,
            AutoCommit           => 1,
            InactiveDestroy      => 1,
            AutoInactiveDestroy  => 1
        };

        my $data_source = sprintf(
            "dbi:mysql:database=%s;host=%s",
            $self->{instance}->{config}->{dbname},
            $self->{instance}->{config}->{dbhost}
        );

        eval { $dbh->disconnect() } if $force && defined $dbh;

        $dbh = DBI->connect_cached(
            $data_source,
            $self->{instance}->{config}->{dbuser},
            $self->{instance}->{config}->{dbpass}, $db_attr
        ) or die $DBI::errstr;
        $dbh->{HandleError} =
          sub { $self->debug( [ "SQL exception: ", \@_ ] ); return 0; };
    }

    return $dbh;
}

sub get_user_data {
    my $self = shift;
    my ($condition) = @_;

    my $combined_data = {};

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->select( 'user_extensions', '*', $condition );

    my $sth = $dbh->prepare($stmt) || return {};
    my $rv  = $sth->execute(@bind) || return {};

    my $hash_ref = $sth->fetchrow_hashref();

    $sth->finish();

    return $hash_ref;
}

sub get_providers {
    my $self      = shift;
    my $providers = {};

    my $condition = { section => "oauth" };

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->select( 'global_defaults', '*', $condition );

    my $sth = $dbh->prepare($stmt) || return [];
    my $rv  = $sth->execute(@bind) || return [];

    while ( my $hash_ref = $sth->fetchrow_hashref ) {
        my ( $provider_name, $key ) = $hash_ref->{key} =~ m/^(.*?)\.(.*?)$/sg;

        if ( defined $provider_name && length $provider_name ) {
            $providers->{$provider_name} = ucfirst lc $provider_name;
        }
    }

    $sth->finish();

    return [ keys %{$providers} ];
}

sub get_provider_config {
    my $self = shift;
    my ($args) = @_;

    return {}
      unless exists $args->{provider}
      && defined $args->{provider}
      && length $args->{provider};

    my $provider_config = {};

#    my $provider_config = $self->cache->get('oauth::config::providers::global::' . $args->{provider} . '::hash');

    #    unless ( defined $provider_config ) {

    my $condition = { section => "oauth" };

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->select( 'global_defaults', '*', $condition );

    my $sth = $dbh->prepare($stmt) || return [];
    my $rv  = $sth->execute(@bind) || return [];

    while ( my $hash_ref = $sth->fetchrow_hashref ) {
        my ( $provider_name, $key ) = $hash_ref->{key} =~ m/^(.*?)\.(.*?)$/sg;

        if ( defined $provider_name
            && lc $provider_name eq lc $args->{provider} )
        {
            $provider_config->{ lc $key } = $hash_ref->{value};
        }
    }

    $sth->finish();

#    $self->cache->set('oauth::config::providers::global::' . $args->{provider} . '::hash', $provider_config, 60*60)
#	if defined $provider_config && ref $provider_config eq "HASH" && scalar keys %{ $provider_config };

    #    }

    return {
        provider => $provider_config,
        db_handle  => sub { $self->{dbh}->() },
        debug      => $self->{debug},
        debug_code => sub { $self->debug(@_); },

        #	cache => $self->{cache}
    };
}

sub get_request_ttl {
    my $self = shift;
    return time() + AUTHENTICATION_TTL;
}

sub get_request_max_ttl {
    my $self = shift;
    return time() + AUTHENTICATION_MAX_TTL;
}

sub gen_uuid {
    my $self = shift;

    my @chars = ( "A" .. "Z", "a" .. "z", "0" .. "9" );
    my $string;
    $string .= $chars[ rand @chars ] for 1 .. 32;

    return Data::UUID->new->create_from_name_str( NameSpace_URL,
        $string . ".thirdlane.com" );
}

sub load_provider {
    my $self          = shift;
    my $provider_name = shift;

    my $provider = ucfirst lc $provider_name;

    my $provider_class = __PACKAGE__ . '::' . $provider;

    if ( !Class::Load::load_optional_class($provider_class) ) {
        $provider_class = __PACKAGE__ . '::Generic';
        $self->{provider} = 'generic';
    }

    return $provider_class;
}

sub issue_signin_url {
    my $self = shift;
    my ($args) = @_;

    return $self->provider->signin_url( { state => 'signin' } );
}

sub issue_authentication_url {
    my $self = shift;
    my ($args) = @_;

    my $uuid = $self->provider->gen_uuid();

    my $context = { context => $uuid };

    $args->{data} = {}
      unless exists $args->{data}
      && defined $args->{data}
      && ref $args->{data} eq "HASH";

    $args->{data}->{provider} = $self->{provider};
    $args->{data}->{user_id}  = $args->{user_id}
      if exists $args->{user_id}
      && defined $args->{user_id}
      && length $args->{user_id};

    $context->{user_id} = $args->{user_id}
      if exists $args->{user_id}
      && defined $args->{user_id}
      && length $args->{user_id};
    $context->{user_name} = $args->{user_name}
      if exists $args->{user_name}
      && defined $args->{user_name}
      && length $args->{user_name};

    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "type",
            value => "authentication",
            type  => "string",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    );
    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "data",
            value => $args->{data},
            type  => "JSON",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    ) if exists $args->{data};
    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "provider",
            value => $args->{data}->{provider},
            type  => "string",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    ) if exists $args->{data}->{provider};

    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "user_id",
            value => $args->{user_id},
            type  => "string",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    );

    return $self->provider->authentication_url( { state => $uuid } );
}

sub issue_authorization_url {
    my $self = shift;
    my ($args) = @_;

    my $uuid = $self->provider->gen_uuid();

    my $context = { context => $uuid };

    $args->{data} = {}
      unless exists $args->{data}
      && defined $args->{data}
      && ref $args->{data} eq "HASH";

    $args->{data}->{provider} = $self->{provider};
    $args->{data}->{user_id}  = $args->{user_id}
      if exists $args->{user_id}
      && defined $args->{user_id}
      && length $args->{user_id};

    $context->{user_id} = $args->{user_id}
      if exists $args->{user_id}
      && defined $args->{user_id}
      && length $args->{user_id};
    $context->{user_name} = $args->{user_name}
      if exists $args->{user_name}
      && defined $args->{user_name}
      && length $args->{user_name};

    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "type",
            value => "authorization",
            type  => "string",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    );
    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "data",
            value => $args->{data},
            type  => "JSON",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    ) if exists $args->{data};
    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "provider",
            value => $args->{data}->{provider},
            type  => "string",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    ) if exists $args->{data}->{provider};

    $self->provider->set_auth_data(
        {
            store => "requests",
            key   => "user_id",
            value => $args->{user_id},
            type  => "string",
            ttl   => $self->get_request_ttl(),
            %{$context}
        }
    );

    return $self->provider->authorization_url( { state => $uuid } );
}

sub update_auth_data_access {
    my $self = shift;
    my ($params) = @_;

    my $args        = $params->{args};
    my $access_data = $params->{access_data};

    if (   exists $args->{context}
        && defined $args->{context}
        && length $args->{context} )
    {
        $args->{state} = $args->{context};
    }

    if (   defined $access_data
        && exists $args->{user_id}
        && defined $args->{user_id}
        && length $args->{user_id}
        && exists $args->{state}
        && defined $args->{state}
        && length $args->{state} )
    {
        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "id_token",
                value   => $access_data->{id_token},
                type    => "JSON",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
        ) if exists $access_data->{id_token};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "external_id",
                value   => $access_data->{id_token}->{'sub'},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
        ) if exists $access_data->{id_token}->{'sub'};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "ext_id.name",
                value   => $access_data->{id_token}->{name},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
        ) if exists $access_data->{id_token}->{name};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "ext_id.email",
                value   => $access_data->{id_token}->{email},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
        ) if exists $access_data->{id_token}->{email};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "ext_id.exp",
                value   => $access_data->{id_token}->{'exp'},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
        ) if exists $access_data->{id_token}->{'exp'};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "public_keys",
                value   => $access_data->{public_keys},
                type    => "JSON",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
        ) if exists $access_data->{public_keys};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "access_token",
                value   => $access_data->{access_token},
                type    => "JSON",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
        ) if exists $access_data->{access_token};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "access_token",
                value   => $access_data->{access_token}->{access_token},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
          )
          if exists $access_data->{access_token}
          && exists $access_data->{access_token}->{access_token};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "access_token.expires_in",
                value   => $access_data->{access_token}->{expires_in},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
          )
          if exists $access_data->{access_token}
          && exists $access_data->{access_token}->{expires_in};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "access_token.token_type",
                value   => $access_data->{access_token}->{token_type},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
          )
          if exists $access_data->{access_token}
          && exists $access_data->{access_token}->{token_type};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "refresh_token",
                value   => $access_data->{access_token}->{refresh_token},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $args->{state}
            }
          )
          if exists $access_data->{access_token}
          && exists $access_data->{access_token}->{refresh_token};

        return 1;
    }

    return undef;
}

sub revoke_context {
    my $self = shift;
    my ($args) = @_;

    return [] unless defined $args && exists $args->{context};

    my $context = { context => $args->{context} };

    $context->{user_id} = $args->{user_id}
      if exists $args->{user_id}
      && defined $args->{user_id}
      && length $args->{user_id};

    my $auth_data = $self->provider->get_auth_data($context);

    return []
      unless defined $auth_data
      && exists $auth_data->{ $args->{context} }->{authentications}->{provider}
      ->{string};

    my $provider =
      $self->build_provider(
        $auth_data->{ $args->{context} }->{authentications}->{provider}
          ->{string} );

    return [] unless defined $provider;

    $provider->remove_user_context($context);

    return $self->clean_user_data($context);
}

sub clean_user_data {
    my $self = shift;
    my ($args) = @_;

    my $avail_providers = {};

    my $context_auth_data = $self->provider->get_auth_data(
        {
            store => "authentications",
            key   => "user_id",
            value => $args->{user_id}
        }
    );

    if ( defined $context_auth_data ) {
        foreach my $context_auth ( keys %{$context_auth_data} ) {
            my $posible_auth_data =
              $self->provider->get_auth_data( { context => $context_auth } );

            if ( defined $posible_auth_data
                && exists $posible_auth_data->{$context_auth}
                ->{authentications}->{provider}->{string}
                && defined $posible_auth_data->{$context_auth}
                ->{authentications}->{provider}->{string}
                && length $posible_auth_data->{$context_auth}
                ->{authentications}->{provider}->{string} )
            {
                $avail_providers->{ $posible_auth_data->{$context_auth}
                      ->{authentications}->{provider}->{string} } =
                  $args->{user_id};
            }
        }
    }

    foreach my $provider (qw(google generic microsoft)) {
        eval {
            $self->remove_contact_source(
                { user_id => $args->{user_id}, provider => $provider } );
        } unless exists $avail_providers->{$provider};
    }

}

sub process_received_code {
    my $self = shift;
    my ($args) = @_;

    return undef unless defined $args;
    return undef
      if exists $args->{error}
      && defined $args->{error}
      && length $args->{error};

    return undef
      unless exists $args->{code}
      && defined $args->{code}
      && length $args->{code};

    my $access_data =
      $self->provider->get_access_token( { code => $args->{code} } );

    my $is_login = 0;

    if ( $args->{state} eq "signin" ) {
        if (   exists $access_data->{id_token}->{'sub'}
            && defined $access_data->{id_token}->{'sub'}
            && length $access_data->{id_token}->{'sub'} )
        {
            my $context_auth_data = $self->provider->get_auth_data(
                {
                    store => "authentications",
                    key   => "external_id",
                    value => $access_data->{id_token}->{'sub'},
                    type  => "string"
                }
            );

            if ( defined $context_auth_data ) {
                foreach my $context_auth ( keys %{$context_auth_data} ) {
                    my $posible_auth_data = $self->provider->get_auth_data(
                        { context => $context_auth } );

                    if (
                        exists $posible_auth_data->{$context_auth}
                        ->{authentications}->{access}->{string}
                        && defined $posible_auth_data->{$context_auth}
                        ->{authentications}->{access}->{string}
                        && $posible_auth_data->{$context_auth}
                        ->{authentications}->{access}->{string} eq "approved" )
                    {
                        $args->{state} = $context_auth;
                        $is_login = 1;
                        last;
                    }
                }
            }
        }

        return undef unless $is_login;
    }

    my $is_requested = 0;

    my $uuid = $self->provider->gen_uuid();
    my $context = { context => $args->{state} };

    my $auth_data = $self->provider->get_auth_data($context);

    if (   defined $auth_data
        && exists $args->{state}
        && exists $auth_data->{ $args->{state} }->{requests}->{type}->{string}
        && $auth_data->{ $args->{state} }->{requests}->{type}->{string} eq
        "authorization" )
    {
        $is_requested = 1;
    }
    elsif (defined $auth_data
        && exists $args->{state}
        && exists $auth_data->{ $args->{state} }->{requests}->{type}->{string}
        && $auth_data->{ $args->{state} }->{requests}->{type}->{string} eq
        "authentication" )
    {
        $is_requested = 1;
    }

    if ($is_requested) {
        $args->{user_id} =
          $auth_data->{ $args->{state} }->{requests}->{user_id}->{string}
          unless exists $args->{user_id};

        return undef
          unless exists $args->{user_id}
          && defined $args->{user_id}
          && length $args->{user_id};

        $self->create_contact_source(
            {
                user_id => $args->{user_id},
                provider =>
                  $auth_data->{ $args->{state} }->{requests}->{provider}
                  ->{string}
            }
        );

        $self->provider->clear_auth_data(
            { context => $args->{state}, user_id => $args->{user_id} } );
        $self->provider->clear_auth_requests(
            {
                provider =>
                  $auth_data->{ $args->{state} }->{requests}->{provider}
                  ->{string},
                user_id => $args->{user_id}
            }
        );

        my $user_contexts = $self->provider->find_user_providers(
            { user_id => $args->{user_id} } );

        if ( defined $user_contexts ) {
            foreach my $avail_user_context ( keys %{$user_contexts} ) {
                if (
                    lc $user_contexts->{$avail_user_context} eq
                    lc $auth_data->{ $args->{state} }->{requests}->{data}
                    ->{JSON}->{provider} )
                {
                    $self->provider->remove_user_context(
                        {
                            user_id => $args->{user_id},
                            context => $avail_user_context
                        }
                    );
                }
            }
        }

        $self->provider->set_auth_data(
            {
                store => "authentications",
                key   => "requested_as",
                value =>
                  $auth_data->{ $args->{state} }->{requests}->{type}->{string},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $uuid
            }
          )
          if exists $auth_data->{ $args->{state} }->{requests}->{type}->{string}
          && length $auth_data->{ $args->{state} }->{requests}->{type}
          ->{string};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "user_id",
                value   => $args->{user_id},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $uuid
            }
        );

        $self->provider->set_auth_data(
            {
                store => "authentications",
                key   => "provider",
                value =>
                  $auth_data->{ $args->{state} }->{requests}->{data}->{JSON}
                  ->{provider},
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $uuid
            }
          )
          if exists $auth_data->{ $args->{state} }->{requests}->{data}->{JSON}
          ->{provider};

        $self->provider->set_auth_data(
            {
                store   => "authentications",
                key     => "access",
                value   => "approved",
                type    => "string",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $uuid
            }
        );

        $self->provider->set_auth_data(
            {
                store => "authentications",
                key   => "data",
                value =>
                  $auth_data->{ $args->{state} }->{requests}->{data}->{JSON},
                type    => "JSON",
                ttl     => $self->get_request_max_ttl(),
                user_id => $args->{user_id},
                context => $uuid
            }
          )
          if exists $auth_data->{ $args->{state} }->{requests}->{data}->{JSON};

        if (
            exists $auth_data->{ $args->{state} }->{requests}->{data}->{JSON}
            ->{'keys'} )
        {
            foreach my $add_key (
                @{
                    $auth_data->{ $args->{state} }->{requests}->{data}->{JSON}
                      ->{'keys'}
                }
              )
            {
                $self->provider->set_auth_data(
                    {
                        store   => "authentications",
                        key     => $add_key->{key},
                        value   => $add_key->{value},
                        type    => "string",
                        ttl     => $self->get_request_max_ttl(),
                        user_id => $args->{user_id},
                        context => $uuid
                    }
                );
            }
        }

        $args->{state} = $uuid;

        $auth_data = $self->provider->get_auth_data( { context => $uuid } );
    }
    else {
        return undef
          unless $auth_data->{ $args->{state} }->{authentications}->{access}
          ->{string} eq "approved";

        $args->{user_id} =
          $auth_data->{ $args->{state} }->{authentications}->{user_id}->{string}
          unless exists $args->{user_id};

        return undef
          unless exists $args->{user_id}
          && defined $args->{user_id}
          && length $args->{user_id};
    }

    if (   exists $args->{state}
        && defined $args->{state}
        && length $args->{state}
        && exists $args->{user_id}
        && defined $args->{user_id}
        && length $args->{user_id} )
    {
        $self->update_auth_data_access(
            { args => $args, access_data => $access_data } );

        $auth_data =
          $self->provider->get_auth_data( { context => $args->{state} } );
    }
    else {
        $auth_data = undef;
    }

    return $auth_data;
}

sub process_refresh_token {
    my $self = shift;
    my ($args) = @_;

    my $context = { context => $args->{context} };

    my $auth_data = $self->provider->get_auth_data($context);

    if (
        exists $auth_data->{ $args->{context} }->{authentications}
        ->{refresh_token}->{string} )
    {
        my $access_data = $self->provider->refresh_access_token(
            {
                refresh_token =>
                  $auth_data->{ $args->{context} }->{authentications}
                  ->{refresh_token}->{string}
            }
        );

        $self->update_auth_data_access(
            {
                args => {
                    context => $args->{context},
                    user_id =>
                      $auth_data->{ $args->{context} }->{authentications}
                      ->{user_id}->{string}
                },
                access_data => $access_data
            }
        );
    }

    my $saved_auth_data =
      $self->provider->get_auth_data( { context => $args->{context} } );

    return $saved_auth_data;
}

sub build_user_agent {
    my $self = shift;
    my ($args) = @_;

    my $context = { context => $args->{context} };

    $args->{provider} = lc $self->{provider} unless exists $args->{provider};

    if ( $args->{user_id} ) {
        my $params = $self->get_contact_source_params(
            { user_id => $args->{user_id}, provider => lc $args->{provider} } );

        if (   defined $params
            && exists $params->{access_token}
            && defined $params
            && length $params )
        {
            $args->{context} = $params->{context};
            $context = { context => $args->{context} };
        }
    }

    $self->debug( [ "build ua args: ", $args ] );

    my $provider_ua =
      $self->provider->build_ua( { context => $args->{context} } );

    return undef unless defined $provider_ua;

    $provider_ua->add_handler(
        response_done => sub {

            #	my($response, $ua, $h) = @_;
            if ( defined $_[0] ) {
                if ( my $internal_status = $_[0]->header('Tlintstatus') ) {
                    if ( $internal_status eq "stale_access_token" ) {
                        my $status_context =
                          $_[0]->header('Tlintstatuscontext');
                        $_[0]->header( Tlprovider => $self->provider->name );

                        $self->debug( ["Stale access_token. Request retry."] );

                        $self->process_refresh_token(
                            { context => $status_context } );

                        my $retry_provider_ua = $self->provider->build_ua(
                            { context => $status_context } );

                        my $retry_result =
                          $retry_provider_ua->request( $_[0]->request() );

                        if ( defined $retry_result ) {
                            $_[0]->content( $retry_result->content() );
                            $_[0]->code( $retry_result->code() );
                            $_[0]->message( $retry_result->message() );
                            $_[0]->headers()->clear();
                            map {
                                $_[0]->headers()
                                  ->header(
                                    $_ => $retry_result->headers()->header($_) )
                            } $retry_result->headers()->header_field_names();
                            map {
                                $_[0]->request()->headers()
                                  ->header(
                                    $_ => $retry_result->request()->headers()
                                      ->header($_) )
                              } $retry_result->request()->headers()
                              ->header_field_names();
                        }

                        $self->debug( [ "Fresh response: ", $_[0] ] );
                    }
                }
            }
        }
    );

    $self->debug( [ "build ua final: ", $provider_ua ] );

    return $provider_ua;
}

sub create_contact_source {
    my $self = shift;
    my ($args) = @_;

    my $condition =
      { user_id => $args->{user_id}, type => lc $args->{provider} };

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->select( 'contact_source', '*', $condition );

    my $sth = $dbh->prepare($stmt);
    my $rv  = $sth->execute(@bind);

    my $hash_ref = $sth->fetchrow_hashref();

    $sth->finish();

    if ( defined $hash_ref && exists $hash_ref->{id} ) {
        return $hash_ref->{id};
    }

    my $values = {
        user_id     => $args->{user_id},
        type        => lc $args->{provider},
        description => sprintf( "%s", ucfirst lc $args->{provider} ),
        weight      => 1,
        status      => 1
    };

    ( $stmt, @bind ) = $sql->insert( 'contact_source', $values );

    $sth = $dbh->prepare($stmt);
    $rv  = $sth->execute(@bind);

#    my $contact_source_id = $dbh->last_insert_id(undef, $self->{instance}->{config}->{dbname}, 'contact_source', 'id');
    my $contact_source_id = $sth->{mysql_insertid};

    $sth->finish();

    if ( defined $contact_source_id && $contact_source_id ) {
        $values = {
            contact_source_id => $contact_source_id,
            key               => "status",
            value             => "1"
        };

        ( $stmt, @bind ) = $sql->insert( 'contact_source_attr', $values );
        $sth = $dbh->prepare($stmt);
        $rv  = $sth->execute(@bind);
        $sth->finish();

        return $contact_source_id;
    }
    else {
        return undef;
    }
}

sub remove_contact_source {
    my $self = shift;
    my ($args) = @_;

    my $condition =
      { user_id => $args->{user_id}, type => lc $args->{provider} };

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->select( 'contact_source', '*', $condition );

    my $sth = $dbh->prepare($stmt);
    my $rv  = $sth->execute(@bind);

    my $hash_ref = $sth->fetchrow_hashref();

    $sth->finish();

    if ( defined $hash_ref && exists $hash_ref->{id} ) {
        ( $stmt, @bind ) = $sql->delete( 'contact_source', $condition );
        $sth = $dbh->prepare($stmt);
        $rv  = $sth->execute(@bind);
        $sth->finish();

        $condition = { contact_source_id => $hash_ref->{id} };

        ( $stmt, @bind ) = $sql->delete( 'contact_source_attr', $condition );
        $sth = $dbh->prepare($stmt);
        $rv  = $sth->execute(@bind);
        $sth->finish();

        return $hash_ref->{id};
    }
    else {
        return undef;
    }
}

sub get_contact_source_params {
    my $self = shift;
    my ($args) = @_;

    return undef
      unless exists $args->{user_id}
      && defined $args->{user_id}
      && length $args->{user_id};

    my $user_contexts =
      $self->provider->find_user_providers( { user_id => $args->{user_id} } );

    if ( defined $user_contexts ) {
        foreach my $avail_user_context ( keys %{$user_contexts} ) {
            if (
                lc $user_contexts->{$avail_user_context} eq
                lc $args->{provider} )
            {
                my $auth_data = $self->provider->get_auth_data(
                    {
                        store   => "authentications",
                        key     => "access_token",
                        type    => "string",
                        user_id => $args->{user_id},
                        context => $avail_user_context
                    }
                );

                if (
                    exists $auth_data->{$avail_user_context}->{authentications}
                    ->{access_token}->{string}
                    && defined $auth_data->{$avail_user_context}
                    ->{authentications}->{access_token}->{string}
                    && length $auth_data->{$avail_user_context}
                    ->{authentications}->{access_token}->{string} )
                {
                    return {
                        access_token =>
                          $auth_data->{$avail_user_context}->{authentications}
                          ->{access_token}->{string},
                        context => $avail_user_context,
                        user_id => $args->{user_id}
                    };
                }
            }
        }
    }

    return undef;
}

1;
