package Thirdlane::Auth::OAuth2::Generic;

use utf8;
use strict;
use warnings;

use DBI;

use MIME::Base64;
use File::Slurp;
use IO::Socket::SSL;
use LWP::UserAgent;

use Cwd 'abs_path';
use POSIX;

use Digest::MD5;

use URI::Encode;
use URI;

use JSON::XS qw/decode_json encode_json/;
use MIME::Base64 qw/encode_base64url decode_base64url/;
use Digest::SHA qw/hmac_sha256 hmac_sha384 hmac_sha512/;

use Data::Dumper qw(Dumper);

use SQL::Abstract;
use DateTime::Format::Strptime;

use Data::UUID;

use HTTP::Request;
use HTTP::Headers;

our $VERSION = '1.00001';

use constant { USER_AGENT => "ThirdlanePBX/1.01 (Linux x86_64)" };

our $MATCH_JSON = qr{
(?&VALUE) (?{ $_ = $^R->[1] if 0 })

(?(DEFINE)

(?<OBJECT>
  #(?{ [$^R, {}] })
  \{\s*
    (?: (?&KV) # [[$^R, {}], $k, $v]
    #  (?{ # warn Dumper { obj1 => $^R };
    #      die "Duplicate key '$^R->[1]'" if exists $^R->[0][1]->{$^R->[1]};
    #      [$^R->[0][0], {$^R->[1] => $^R->[2]}] })
      (?: \s*,\s* (?&KV) # [[$^R, {...}], $k, $v]
    #    (?{ # warn Dumper { obj2 => $^R };
    #        die "Duplicate key '$^R->[1]'" if exists $^R->[0][1]->{$^R->[1]};
    #        [$^R->[0][0], {%{$^R->[0][1]}, $^R->[1] => $^R->[2]}] })
      )*
    )?
  \s*\}
)

(?<KV>
  (?&STRING) # [$^R, "string"]
  \s*:\s* (?&VALUE) # [[$^R, "string"], $value]
  #(?{ # warn Dumper { kv => $^R };
  #   [$^R->[0][0], $^R->[0][1], $^R->[1]] })
)

(?<ARRAY>
  #(?{ [$^R, []] })
  \[\s*
    (?: (?&VALUE) #(?{ [$^R->[0][0], [$^R->[1]]] })
      (?: \s*,\s* (?&VALUE) #(?{ # warn Dumper { atwo => $^R };
			 #[$^R->[0][0], [@{$^R->[0][1]}, $^R->[1]]] })
      )*
    )?
  \s*\]
)

(?<VALUE>
  \s*
  (
      (?&STRING)
    |
      (?&NUMBER)
    |
      (?&OBJECT)
    |
      (?&ARRAY)
    |
    true #(?{ [$^R, 1] })
  |
    false #(?{ [$^R, 0] })
  |
    null #(?{ [$^R, undef] })
  )
  \s*
)

(?<STRING>
  (
    "
    (?:
      [^\\"]+
    |
      \\ ["\\/bfnrt]
#    |
#      \\ u [0-9a-fA-f]{4}
    )*
    "
  )

  #(?{ [$^R, eval $^N] })
)

(?<NUMBER>
  (
    -?
    (?: 0 | [1-9]\d* )
    (?: \. \d+ )?
    (?: [eE] [-+]? \d+ )?
  )

  #(?{ [$^R, eval $^N] })
)

) }xms;

sub new {
    my $self = shift;
    my ($opts) = @_;

    my $class = ref($self) || $self;

    $opts->{initialized} = 0;

    my $obj = bless { %{$opts} }, $class;

    $obj->init();

    return $obj;
}

sub init {
    my $self = shift;

    $self->{debug} = "" unless exists $self->{debug} && defined $self->{debug};
    $self->{sql} = SQL::Abstract->new( quote_char => '`' );
    $self->{instance}->{config} = $self->get_config();
    $self->{instance}->{webmin}->{port} = $self->get_webmin_port();
    if ( exists $self->{db_handle} ) {
        if ( defined $self->{db_handle} && ref $self->{db_handle} eq "CODE" ) {
            $self->{dbh} = $self->{db_handle};
        }
        else {
            $self->{dbh} = sub { $self->db_connect() };
        }

        delete $self->{db_handle};
    }
    $self->{ua}              = $self->get_ua();
    $self->{json}            = $self->get_json_codec();
    $self->{mysql_dt_parder} = $self->build_mysql_datetime_parser();
}

#sub cache {
#    my $self = shift;
#    my $type = shift || 'memory';
#    $type = 'memory' unless exists $self->{cache}->{$type};
#    return $self->{cache}->{$type};
#}

sub build_mysql_datetime_parser {
    my $self = shift;

    return DateTime::Format::Strptime->new(
        pattern   => '%F %T',
        locale    => 'C',
        time_zone => 'local',
        on_error  => sub { $self->debug( [ 'Parse datetime error: ', @_ ] ); }
    );
}

sub mysql_parse_datetime {
    my $self = shift;
    my ($date_time) = @_;

    if ( $date_time =~ m{^0000\-00\-00} ) {
        return DateTime->now()->epoch();
    }

    return
      eval { $self->{mysql_dt_parder}->parse_datetime($date_time)->epoch() };
}

sub debug {
    my $self = shift;

    return $self->{debug_code}->( \@_ )
      if exists $self->{debug_code}
      && defined $self->{debug_code}
      && ref $self->{debug_code} eq "CODE"
      && $self->{debug};
}

sub is_json {
    my $self   = shift;
    my $string = shift;

    my $re = qr/\A$MATCH_JSON\z/;

    return $string =~ $re ? 1 : 0;
}

sub ua {
    my $self = shift;

    return $self->{ua};
}

sub json {
    my $self = shift;

    return $self->{json};
}

sub _ua_get {
    my $self = shift;
    my ( $url, $ua_conf ) = @_;

    return undef unless defined $url && length $url;

    my $ua = $self->ua();

    my $headers = HTTP::Headers->new( @{ $ua_conf->{headers} } );

    my $request = HTTP::Request->new( "GET" => $url, $headers );

    if ( defined $ua_conf ) {
        $ua = $self->get_ua();

        $ua->ssl_opts( $ua_conf->{ssl_opts} ) if exists $ua_conf->{ssl_opts};
    }

    $self->debug( [ "GET: ", $url, $ua_conf, $request ] );

    my $response = eval { $ua->request($request) };

    $self->debug( [ "GET response: ", $response ] );

    if ( defined $response ) {
        $self->debug(
            [ "Get response content: ", $response->decoded_content() ] );

        if ( $response->is_success() ) {
            return $response->decoded_content();
        }
    }

    return undef;
}

sub _ua_post {
    my $self = shift;
    my ( $url, $data, $ua_conf ) = @_;

    return undef unless defined $url && length $url;

    my $ua = $self->ua();

    my $headers = HTTP::Headers->new( @{ $ua_conf->{headers} } );

    my $request = HTTP::Request->new( "POST" => $url, $headers );

    my $content_type = $request->header('Content-Type');
    unless ($content_type) {
        $content_type = 'application/x-www-form-urlencoded';
    }

    my $body = $data;

    if ( ref $data ) {
        my $uri = URI->new('http:');
        $uri->query_form( ref $data eq "HASH" ? %{$data} : @{$data} );
        $body = $uri->query();
        $body =~ s/(?<!%0D)%0A/%0D%0A/g if defined($body);
    }

    if ( defined $body ) {
        $request->header( 'Content-Length' => length($body) ) unless ref $body;
        $request->content($body);
    }
    else {
        $request->header( 'Content-Length' => 0 );
    }

    $request->header( 'Content-Type' => $content_type );

    if ( defined $ua_conf ) {
        $ua = $self->get_ua();

        $ua->ssl_opts( $ua_conf->{ssl_opts} ) if exists $ua_conf->{ssl_opts};
    }

    $self->debug( [ "POST url: ",     $url ] );
    $self->debug( [ "POST data: ",    $data ] );
    $self->debug( [ "POST ua_conf: ", $ua_conf ] );
    $self->debug( [ "POST body: ",    $body ] );
    $self->debug( [ "POST request: ", $request ] );

    my $response = eval { $ua->request($request) };

    $self->debug( [ "POST response: ", $response ] );

    if ( defined $response ) {
        $self->debug(
            [ "POST response content: ", $response->decoded_content() ] );

        if ( $response->is_success() ) {
            return $response->decoded_content();
        }
    }

    return undef;
}

sub web_req {
    my $self = shift;
    my ($args) = @_;

    $self->debug( [ $args->{method} . " request args: ", $args ] );

    $args = { url => $args, method => "GET" }
      if ( defined $args && ref $args eq "" && length $args );

    if ( exists $args->{query} && defined $args->{query} ) {
        my $uri = URI->new( $args->{url} );

        if ( ref $args->{query} eq "HASH" ) {
            $uri->query_form( $args->{query} );
        }
        elsif ( ref $args->{query} eq "ARRAY" ) {
            $uri->query_form( { @{ $args->{query} } } );
        }

        $args->{url} = $uri->as_string;
    }

    if ( exists $args->{ua}->{headers} && defined $args->{ua}->{headers} ) {
        if ( ref $args->{ua}->{headers} eq "ARRAY" ) {
            push @{ $args->{ua}->{headers} }, ( User_Agent => USER_AGENT );
        }
        elsif ( ref $args->{ua}->{headers} eq "HASH" ) {
            $args->{ua}->{headers} =
              [ %{ $args->{ua}->{headers} }, ( User_Agent => USER_AGENT ) ];
        }
    }
    else {
        $args->{ua}->{headers} = [ User_Agent => USER_AGENT ];
    }

    if (   exists $args->{body}
        && defined $args->{body}
        && ref $args->{body} eq "HASH" )
    {
        push @{ $args->{ua}->{headers} },
          ( Content_Type => 'application/json; charset=UTF-8' );
        $args->{body} = eval { $self->json->encode( $args->{body} ) };
        return undef if $@;
    }
    elsif (exists $args->{body}
        && defined $args->{body}
        && ref $args->{body} eq "ARRAY" )
    {
        push @{ $args->{ua}->{headers} },
          ( Content_Type =>
              'application/x-www-form-urlencoded; charset=UTF-8' );
    }
    elsif (exists $args->{body}
        && defined $args->{body}
        && ref $args->{body} eq "" )
    {
        if ( $self->is_json( $args->{body} ) ) {
            push @{ $args->{ua}->{headers} },
              ( Content_Type => 'application/json; charset=UTF-8' );
        }
        else {
            push @{ $args->{ua}->{headers} },
              ( Content_Type => 'application/binary' );
        }
    }

    $self->debug( [ $args->{method} . " final request args: ", $args ] );

    if ( uc $args->{method} eq "POST" ) {
        my $response =
          $self->_ua_post( $args->{url}, $args->{body}, $args->{ua} );

        if ( defined $response ) {
            $self->debug(
                [ $args->{method} . " final response: ", $response ] );

            if ( exists $args->{json} && $args->{json} ) {
                $response = eval { $self->json->decode($response) };
                $self->debug(
                    [ $args->{method} . " final json response: ", $response ] );
                $self->debug(
                    [ $args->{method} . " final json response error: ", $@ ] )
                  if $@;
            }

            return $response;
        }
    }
    else {
        my $response = $self->_ua_get( $args->{url}, $args->{ua} );

        if ( defined $response ) {
            $self->debug(
                [ $args->{method} . " final response: ", $response ] );

            if ( exists $args->{json} && $args->{json} ) {
                $response = eval { $self->json->decode($response) };
                $self->debug(
                    [ $args->{method} . " final json response: ", $response ] );
                $self->debug(
                    [ $args->{method} . " final json response error: ", $@ ] )
                  if $@;
            }

            return $response;
        }
    }
}

sub merge {
    my $self = shift;

    my $result = {};

    foreach my $arg (@_) {
        if ( defined $arg && ref $arg eq "HASH" ) {
            foreach my $key ( keys %{$arg} ) {
                $result->{$key} = $arg->{$key};
            }
        }
    }

    return $result;
}

sub get_json_codec {
    my $self = shift;

    my $json = JSON::XS->new();

    $json->allow_blessed(1);
    $json->allow_nonref(1);
    $json->utf8(1);

    return $json;
}

sub get_ua {
    my $self = shift;
    my ($config) = @_;

    my $ua_config = $self->merge( $self->{internal}->{config}->{ua}, $config );

    my $ua = LWP::UserAgent->new( %{ $self->{internal}->{config}->{ua} } );

    $ua->timeout( $self->{internal}->{config}->{ua}->{timeout} )
      if exists $self->{internal}->{config}->{ua}->{timeout};
    $ua->env_proxy();

    return $ua;
}

sub get_config {
    my $self = shift;

    my $config = {};

    #    my $config = $self->cache->get('config::instances::hash');
    #    unless ( defined $config ) {

    my @config_file = read_file("/etc/webmin/asterisk/instances.txt");
    for (@config_file) {
        $config->{$1} = $2 if m/^(.*?)\s*\=\s*(.*?)$/;
    }

    #    $self->cache->set('config::instances::hash', $config, 60*60*24);
    #    }

    return $config;
}

sub get_webmin_port {
    my $self = shift;

    if ( -f "/etc/nginx/conf.d/https/upstream/webmin.conf" ) {
        my $webmin_conf =
          eval { read_file("/etc/nginx/conf.d/https/upstream/webmin.conf") };
        my ($port) = $webmin_conf =~ m/server\s+.*?\:(\d+)\;/sg if $webmin_conf;

        return $port if $port;
    }

    return '22000';
}

sub dbh {
    my $self = shift;
    return $self->{dbh}->();
}

sub db_connect {
    my $self = shift;

    use feature 'state';

    state $dbh;

    unless ( defined $dbh ) {
        my $data_source = sprintf(
            "dbi:mysql:database=%s;host=%s",
            $self->{instance}->{config}->{dbname},
            $self->{instance}->{config}->{dbhost}
        );
        $dbh = DBI->connect_cached(
            $data_source,
            $self->{instance}->{config}->{dbuser},
            $self->{instance}->{config}->{dbpass}
        );
        $dbh->{HandleError} =
          sub { $self->debug( [ "SQL exception: ", \@_ ] ); return 0; };
    }

    return $dbh;
}

sub get_user_data {
    my $self = shift;
    my ($condition) = @_;

    my $combined_data = {};

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->select( 'user_extensions', '*', $condition );

    $self->debug(
        [
            "get_user_data args: ", $condition,
            "statement: ",          $stmt,
            "bind params: ",        @bind
        ]
    ) if $self->{debug} > 3;

    my $sth = $dbh->prepare($stmt) || $self->debug(
        [
            "sql error: ",   $dbh->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return {};
    my $rv = $sth->execute(@bind) || $self->debug(
        [
            "sql error: ",   $sth->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return {};

    my $hash_ref = $sth->fetchrow_hashref();
    $sth->finish();

    $self->debug( [ "get_user_data result: ", $hash_ref ] )
      if $self->{debug} > 3;

    return $hash_ref;
}

sub db_get {
    my $self = shift;
    my ($args) = @_;

    my $condition = { %{$args} };

    $condition->{user_id} =
      get_user_data( { name => delete $condition->{user_name} } )->{id}
      if ( exists $condition->{user_name}
        && defined $condition->{user_name}
        && length $condition->{user_name} );

#    return [] unless exists $condition->{user_id} && defined $condition->{user_id} && length $condition->{user_id};

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->select( 'user_oauth', '*', $condition );

    $self->debug(
        [
            "db_get: ",    $args, "condition: ",   $condition,
            "statement: ", $stmt, "bind params: ", @bind
        ]
    ) if $self->{debug} > 3;

    my $sth = $dbh->prepare($stmt) || $self->debug(
        [
            "sql error: ",   $dbh->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return [];
    my $rv = $sth->execute(@bind) || $self->debug(
        [
            "sql error: ",   $sth->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return [];

    my @result;

    while ( my $hash_ref = $sth->fetchrow_hashref ) {
        foreach my $date_time_field (qw(created updated)) {
            $hash_ref->{$date_time_field} =
              $self->mysql_parse_datetime( $hash_ref->{$date_time_field} );
        }

        push @result, $hash_ref;
    }

    $sth->finish();

    $self->debug( [ "db_get result: ", @result ] ) if $self->{debug} > 3;

    return \@result;
}

sub db_update {
    my $self = shift;
    my ( $val_args, $cond_args ) = @_;

    my $values    = { %{$val_args} };
    my $condition = { %{$cond_args} };

    return 0 unless exists $condition->{context};

    $condition->{user_id} =
      get_user_data( { name => delete $condition->{user_name} } )->{id}
      if ( exists $condition->{user_name}
        && defined $condition->{user_name}
        && length $condition->{user_name} );

#    return [] unless exists $condition->{user_id} && defined $condition->{user_id} && length $condition->{user_id};

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->update( 'user_oauth', $values, $condition );

    $self->debug(
        [
            "db_update value args: ", $val_args,
            "condition args: ",       $cond_args,
            "statement: ",            $stmt,
            "bind params: ",          @bind
        ]
    ) if $self->{debug} > 3;

    my $sth = $dbh->prepare($stmt) || $self->debug(
        [
            "sql error: ",   $dbh->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return 0;
    my $rv = $sth->execute(@bind) || $self->debug(
        [
            "sql error: ",   $sth->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return 0;

    $sth->finish();

    return $rv;
}

sub db_insert {
    my $self = shift;
    my ($args) = @_;

    my $values = { %{$args} };

    return 0 unless exists $values->{context};

    $values->{user_id} =
      get_user_data( { name => delete $values->{user_name} } )->{id}
      if ( exists $values->{user_name}
        && defined $values->{user_name}
        && length $values->{user_name} );

#    return [] unless exists $values->{user_id} && defined $values->{user_id} && length $values->{user_id};

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->insert( 'user_oauth', $values );

    $self->debug(
        [
            "db_insert args: ", $args, "statement: ", $stmt,
            "bind params: ",    @bind
        ]
    ) if $self->{debug} > 3;

    my $sth = $dbh->prepare($stmt) || $self->debug(
        [
            "sql error: ",   $dbh->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return 0;
    my $rv = $sth->execute(@bind) || $self->debug(
        [
            "sql error: ",   $sth->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return 0;

    $sth->finish();

    return $rv;
}

sub db_delete {
    my $self = shift;
    my ($args) = @_;

    my $condition = { %{$args} };

    return 0 unless exists $condition->{context};

    $condition->{user_id} =
      get_user_data( { name => delete $condition->{user_name} } )->{id}
      if ( exists $condition->{user_name}
        && defined $condition->{user_name}
        && length $condition->{user_name} );

#    return [] unless exists $condition->{user_id} && defined $condition->{user_id} && length $condition->{user_id};

    my $dbh = $self->dbh;

    my $sql = $self->{sql};
    my ( $stmt, @bind ) = $sql->delete( 'user_oauth', $condition );

    $self->debug(
        [
            "db_delete args: ", $args, "statement: ", $stmt,
            "bind params: ",    @bind
        ]
    ) if $self->{debug} > 3;

    my $sth = $dbh->prepare($stmt) || $self->debug(
        [
            "sql error: ",   $dbh->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return 0;
    my $rv = $sth->execute(@bind) || $self->debug(
        [
            "sql error: ",   $sth->errstr, "statement: ", $stmt,
            "bind params: ", @bind
        ]
    ) && return 0;

    $sth->finish();

    return $rv;
}

sub get_auth_data {
    my $self = shift;
    my ($args) = @_;

    my $condition = { %{$args} };

    delete $condition->{user_id}
      unless defined $condition->{user_id} && length $condition->{user_id};
    delete $condition->{user_name}
      unless defined $condition->{user_name} && length $condition->{user_name};

    my $combined_data = {};

    foreach my $record ( @{ $self->db_get($condition) } ) {

        $record->{value} =
          eval { JSON::XS->new->utf8->decode( $record->{value} ) }
          if uc $record->{type} eq "JSON";

        foreach
          my $meta_field (qw(user_id created updated created_by updated_by))
        {
            $combined_data->{ $record->{context} }->{ $record->{store} }
              ->{ $record->{key} }->{meta}->{ $record->{type} }->{$meta_field}
              = $record->{$meta_field}
              if exists $record->{$meta_field}
              && defined $record->{$meta_field}
              && length $record->{$meta_field};
        }

        if (
            exists $combined_data->{ $record->{context} }->{ $record->{store} }
            ->{ $record->{key} }->{ $record->{type} }
            && defined $combined_data->{ $record->{context} }
            ->{ $record->{store} }->{ $record->{key} }->{ $record->{type} } )
        {
            if (
                ref $combined_data->{ $record->{context} }
                ->{ $record->{store} }->{ $record->{key} }->{ $record->{type} }
                eq "ARRAY" )
            {
                push @{ $combined_data->{ $record->{context} }
                      ->{ $record->{store} }->{ $record->{key} }
                      ->{ $record->{type} } }, $record->{value};
            }
            else {
                $combined_data->{ $record->{context} }->{ $record->{store} }
                  ->{ $record->{key} }->{ $record->{type} } =
                  [ $record->{value} ];
            }
        }
        else {
            $combined_data->{ $record->{context} }->{ $record->{store} }
              ->{ $record->{key} }->{ $record->{type} } = $record->{value};
        }
    }

    return $combined_data;
}

sub find_user_contexts {
    my $self = shift;
    my ($args) = @_;

    my $condition = {
        user_id => $args->{user_id},
        key     => "user_id",
        type    => "string",
        store   => "authentications"
    };

    return {}
      unless exists $condition->{user_id}
      && defined $condition->{user_id}
      && length $condition->{user_id};

    my $combined_data = {};

    foreach my $record ( @{ $self->db_get($condition) } ) {
        $combined_data->{ $record->{context} } = $record->{store};
    }

    return $combined_data;
}

sub find_user_requests {
    my $self = shift;
    my ($args) = @_;

    my $condition = {
        user_id => $args->{user_id},
        key     => "provider",
        type    => "string",
        store   => "requests"
    };

    return {}
      unless exists $condition->{user_id}
      && defined $condition->{user_id}
      && length $condition->{user_id};

    my $combined_data = {};

    foreach my $record ( @{ $self->db_get($condition) } ) {
        $combined_data->{ $record->{context} } = $record->{value};
    }

    return $combined_data;
}

sub find_user_providers {
    my $self = shift;
    my ($args) = @_;

    my $condition = {
        user_id => $args->{user_id},
        key     => "provider",
        type    => "string",
        store   => "authentications"
    };

    return {}
      unless exists $condition->{user_id}
      && defined $condition->{user_id}
      && length $condition->{user_id};

    my $combined_data = {};

    foreach my $record ( @{ $self->db_get($condition) } ) {
        $combined_data->{ $record->{context} } = $record->{value};
    }

    return $combined_data;
}

sub find_user_providers_details {
    my $self = shift;
    my ($args) = @_;

    my $condition = {
        user_id => $args->{user_id},
        key     => "provider",
        type    => "string",
        store   => "authentications"
    };

    return {}
      unless exists $condition->{user_id}
      && defined $condition->{user_id}
      && length $condition->{user_id};

    my $combined_data = {};

    foreach my $record ( @{ $self->db_get($condition) } ) {
        $combined_data->{ $record->{value} } = $record->{context};
    }

    my $providers = {};

    foreach my $provider ( keys %{$combined_data} ) {
        foreach my $record (
            @{
                $self->db_get(
                    {
                        user_id => $args->{user_id},
                        key     => "ext_id.email",
                        type    => "string",
                        context => $combined_data->{$provider}
                    }
                )
            }
          )
        {
            $self->debug( [ $combined_data, $provider, $record ] );
            $providers->{$provider}->{ $record->{value} }->{id} =
              $record->{context};
            $providers->{$provider}->{ $record->{value} }->{status} =
              "AUTHENTICATED";
        }
    }

    return $providers;
}

sub remove_user_context {
    my $self = shift;
    my ($args) = @_;

    my $condition =
      { user_id => $args->{user_id}, context => $args->{context} };

    return $self->clear_auth_data($condition);
}

sub set_auth_data {
    my $self = shift;
    my ($args) = @_;

    my $condition = { %{$args} };

    my $data_signature = {
        context => $condition->{context},
        store   => $condition->{store},
        key     => $condition->{key},
        type    => $condition->{type} || "string"
    };

    $data_signature->{user_id} = $condition->{user_id}
      if defined $condition->{user_id} && length $condition->{user_id};
    $data_signature->{user_name} = $condition->{user_name}
      if defined $condition->{user_name} && length $condition->{user_name};

    if ( exists $condition->{value} && defined $condition->{value} ) {
        my $auth_data = $self->get_auth_data($data_signature);

        $condition->{value} =
          eval { JSON::XS->new->utf8->encode( $condition->{value} ) }
          if uc $condition->{type} eq "JSON";

        if (
            exists $auth_data->{ $condition->{context} }
            ->{ $condition->{store} }->{ $condition->{key} }
            ->{ $condition->{type} } )
        {
            $self->db_update( $condition, $data_signature );
        }
        else {
            $self->db_insert($condition);
        }
    }
    else {
        $self->db_delete($data_signature);
    }

    return $self->get_auth_data($data_signature);
}

sub clear_auth_data {
    my $self = shift;
    my ($args) = @_;

    my $condition = { %{$args} };

    my $data_signature = { context => $condition->{context} };

    $data_signature->{user_id} = $condition->{user_id}
      if defined $condition->{user_id} && length $condition->{user_id};
    $data_signature->{user_name} = $condition->{user_name}
      if defined $condition->{user_name} && length $condition->{user_name};

    my $auth_tokens = {};

    foreach my $record (
        @{
            $self->db_get(
                {
                    context => $condition->{context},
                    store   => "authentications",
                    key     => "access_token",
                    type    => "JSON"
                }
            )
        }
      )
    {
        my $token = eval { JSON::XS->new->utf8->decode( $record->{value} ) };
        if ( defined $token ) {
            $auth_tokens->{ $token->{access_token} } = $condition->{context}
              if exists $token->{access_token};
            $auth_tokens->{ $token->{refresh_token} } = $condition->{context}
              if exists $token->{refresh_token};
        }
    }

    foreach my $record (
        @{
            $self->db_get(
                {
                    context => $condition->{context},
                    store   => "authentications",
                    key     => "access_token",
                    type    => "string"
                }
            )
        }
      )
    {
        my $token = eval { JSON::XS->new->utf8->decode( $record->{value} ) };
        if ( defined $record->{value} ) {
            $auth_tokens->{ $record->{value} } = $condition->{context};
        }
    }

    foreach my $record (
        @{
            $self->db_get(
                {
                    context => $condition->{context},
                    store   => "authentications",
                    key     => "refresh_token",
                    type    => "string"
                }
            )
        }
      )
    {
        my $token = eval { JSON::XS->new->utf8->decode( $record->{value} ) };
        if ( defined $record->{value} ) {
            $auth_tokens->{ $record->{value} } = $condition->{context};
        }
    }

    foreach my $access_token ( keys %{$auth_tokens} ) {
        eval { $self->revoke_token( { token => $access_token } ) };
    }

    return $self->db_delete($data_signature);

    return [ keys %{$auth_tokens} ];
}

sub clear_auth_requests {
    my $self = shift;
    my ($args) = @_;

    my $condition = { %{$args} };

    return undef
      unless defined $condition->{user_id} && length $condition->{user_id};
    return undef
      unless exists $condition->{provider}
      && defined $condition->{provider}
      && length $condition->{provider};

    my $user_providers =
      $self->find_user_requests( { user_id => $condition->{user_id} } );

    foreach my $user_provider_context ( keys %{$user_providers} ) {
        if (
            lc $user_providers->{$user_provider_context} eq
            lc $condition->{provider} )
        {
            $self->db_delete(
                {
                    context => $user_provider_context,
                    store   => "requests",
                    user_id => $condition->{user_id}
                }
            );
        }
    }

    return 1;
}

sub gen_uuid {
    my $self = shift;

    my @chars = ( "A" .. "Z", "a" .. "z", "0" .. "9" );
    my $string;
    $string .= $chars[ rand @chars ] for 1 .. 32;

    return Data::UUID->new->create_from_name_str( NameSpace_URL,
        $string . ".thirdlane.com" );
}

sub try_decode_jwt {
    my $self = shift;
    my ( $jwt, $key ) = @_;

    require Crypt::JWT;
    Crypt::JWT->import(qw(decode_jwt));

    return decode_jwt( token => $jwt, kid_keys => $key );
}

sub get_provider_data {
    my $self = shift;
    my ($data) = @_;

    my $discovery = $self->get_discovery_data();

    if ( defined $data && defined $discovery && ref $discovery eq 'HASH' ) {
        return $discovery->{$data};
    }

    return $discovery;
}

sub get_discovery_data {
    my $self = shift;

    my $config = {};

#    my $config = $self->cache->get('oauth::config::provider::' .$self->name . '::discovery::hash');

    #    unless ( defined $config ) {
    $config = $self->web_req(
        { url => $self->discovery_url(), method => "GET", json => 1 } );

#	$self->cache->set('oauth::config::provider::' .$self->name . '::discovery::hash', $config, 60*60*24)
#	    if defined $config && ref $config eq 'HASH' && scalar keys %{ $config };
#    }

    return $config;

    #    return defined $config ? $config : { };
}

sub get_access_token {
    my $self = shift;
    return undef;
}

sub authentication_url {
    my $self = shift;
    return undef;
}

sub authorization_url {
    my $self = shift;
    return undef;
}

sub signin_url {
    my $self = shift;
    return undef;
}

sub discovery_url {
    my $self = shift;
    return undef;
}

sub is_provider {
    my $self = shift;
    return 0;
}

sub build_ua {
    my $self = shift;
    my $ua   = $self->get_ua();
    return $ua;
}

sub revoke_token {
    my $self = shift;
    my ($args) = @_;

    my $revoke_token_query = {
        token => $args->{token}    # access_token or refresh_token
    };

    return 1;
}

sub name {
    my $self         = shift;
    my $package_name = ref($self);
    my ($name)       = $package_name =~ m{.+\:{2}(.*?)$}sg;
    return lc $name;
}

1;
